/*
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	Implementation of task thread pool.
 *
 */
#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <complib/cl_mpthreadpool.h>

#define CL_MP_THREAD_POOL_MODE_BIT	0x80000000L

unsigned cl_mp_thread_pool_thread_id()
{
	return cl_thread_pool_thread_id();
}

static void cl_mp_thread_pool_run_task(void *context)
{
	cl_mp_thread_pool_t *p_mp_thread_pool = (cl_mp_thread_pool_t *)context;
	void **task;
	unsigned num_tasks;
	int result;

	pthread_mutex_lock(&p_mp_thread_pool->mutex);

	/* calculate number of tasks to run */

	num_tasks = p_mp_thread_pool->num_tasks - p_mp_thread_pool->next_task;
	if (!num_tasks)
		goto end;

	if (num_tasks > p_mp_thread_pool->batch_size)
		num_tasks = p_mp_thread_pool->batch_size;

	task = &p_mp_thread_pool->tasks[p_mp_thread_pool->next_task];
	p_mp_thread_pool->next_task += num_tasks;

	/* invoke callback function on selected tasks */

	pthread_mutex_unlock(&p_mp_thread_pool->mutex);

	result = p_mp_thread_pool->pfn_callback(task, num_tasks,
						p_mp_thread_pool->context,
						cl_mp_thread_pool_thread_id());

	pthread_mutex_lock(&p_mp_thread_pool->mutex);

	p_mp_thread_pool->result += !!result;

	/*
	 * Update number of pending tasks, and if all tasks are completed
	 * signal the thread pool object.
	 */

	p_mp_thread_pool->num_pending_tasks -= num_tasks;
	if (!p_mp_thread_pool->num_pending_tasks)
		pthread_cond_signal(&p_mp_thread_pool->cond);

end:
	pthread_mutex_unlock(&p_mp_thread_pool->mutex);
}

cl_status_t cl_mp_thread_pool_init(IN cl_mp_thread_pool_t * const p_mp_thread_pool,
				   IN unsigned num_threads,
				   IN uint16_t max_threads_per_core,
				   IN const char *const name)
{
	cl_status_t status = CL_INSUFFICIENT_MEMORY;
	cpu_set_t cpu_set, mask;
	pthread_t pthread;
	unsigned num_cores, max_threads;
	unsigned i, c;

	CL_ASSERT(p_mp_thread_pool);
	p_mp_thread_pool->state = CL_UNINITIALIZED;
	p_mp_thread_pool->num_threads = 0;

	memset(p_mp_thread_pool, 0, sizeof(*p_mp_thread_pool));
	if (!num_threads)
		num_threads = cl_proc_count();

	if (sched_getaffinity(0, sizeof(cpu_set), &cpu_set)) {
		status = CL_ERROR;
		goto Exit;
	}
	num_cores = CPU_COUNT(&cpu_set);

	if (!num_threads || !num_cores) {
		status = CL_ERROR;
		goto Exit;
	}

	/* if max_threads_per_core specified, limit number of threads accordingly */
	if (max_threads_per_core) {
		max_threads = num_cores * max_threads_per_core;
		if (num_threads	> max_threads)
			num_threads = max_threads;
	}

	pthread_mutex_init(&p_mp_thread_pool->mutex, NULL);
	pthread_cond_init(&p_mp_thread_pool->cond, NULL);

	status = cl_thread_pool_init(&p_mp_thread_pool->thread_pool, num_threads,
				     cl_mp_thread_pool_run_task, p_mp_thread_pool, name);
	if (status != CL_SUCCESS) {
		pthread_cond_destroy(&p_mp_thread_pool->cond);
		pthread_mutex_destroy(&p_mp_thread_pool->mutex);
		goto Exit;
	}

	CL_ASSERT(p_mp_thread_pool->thread_pool.running_count <= num_threads);
	CL_ASSERT(p_mp_thread_pool->thread_pool.running_count);

	num_threads = p_mp_thread_pool->thread_pool.running_count;

	/* set threads affinity */
	if (max_threads_per_core) {
		c = 0;
		CPU_ZERO(&mask);
		for (i = 0; i < num_threads; i++) {
			pthread = p_mp_thread_pool->thread_pool.threads[i].thread;

			/* find next CPU */
			for (c = (c + 1) % CPU_SETSIZE; !CPU_ISSET(c, &cpu_set);
			     c = (c + 1) % CPU_SETSIZE);

			/* set affinity */
			CPU_SET(c, &mask);
			if (pthread_setaffinity_np(pthread,
						   sizeof(mask),
						   &mask)) {
				cl_mp_thread_pool_destroy(p_mp_thread_pool);
				status = CL_ERROR;
				goto Exit;
			}
			CPU_CLR(c, &mask);
		}
	}

	p_mp_thread_pool->num_threads = p_mp_thread_pool->thread_pool.running_count;

	status = CL_SUCCESS;
	p_mp_thread_pool->state = CL_INITIALIZED;

Exit:
	return status;
}

void cl_mp_thread_pool_destroy(IN cl_mp_thread_pool_t * const p_mp_thread_pool)
{
	CL_ASSERT(p_mp_thread_pool);

	if (p_mp_thread_pool->state == CL_INITIALIZED) {

		p_mp_thread_pool->state = CL_DESTROYING;

		cl_thread_pool_destroy(&p_mp_thread_pool->thread_pool);
		pthread_cond_destroy(&p_mp_thread_pool->cond);
		pthread_mutex_destroy(&p_mp_thread_pool->mutex);

		p_mp_thread_pool->state = CL_DESTROYED;
	}
}

cl_status_t cl_mp_thread_pool_state(IN cl_mp_thread_pool_t * const p_mp_thread_pool)
{
	return p_mp_thread_pool->state;
}

unsigned cl_mp_thread_pool_num_threads(IN cl_mp_thread_pool_t * const p_mp_thread_pool)
{
	return p_mp_thread_pool->num_threads;
}

unsigned cl_mp_thread_pool_num_thrads(IN cl_mp_thread_pool_t * const p_mp_thread_pool)
{
	unsigned num_threads;

	num_threads = p_mp_thread_pool->state == CL_INITIALIZED ?
		      p_mp_thread_pool->num_threads :
		      1;

	CL_ASSERT(num_threads);

	return num_threads;
}

unsigned cl_mp_thread_pool_array_apply(IN cl_mp_thread_pool_t * const p_mp_thread_pool,
				       IN void * tasks[],
				       IN unsigned num_tasks,
				       IN cl_mp_thread_pool_mode_t mode,
				       IN cl_pfn_mp_thread_pool_apply_t pfn_callback,
				       IN void *context)
{
	unsigned num_batches;
	unsigned batch_size;
	unsigned num_threads;
	unsigned i;

	CL_ASSERT(p_mp_thread_pool);
	CL_ASSERT(p_mp_thread_pool->state == CL_INITIALIZED);
	CL_ASSERT(pfn_callback);

	if (!num_tasks)
		return 0;

	num_threads = p_mp_thread_pool->num_threads;
	if (num_threads == 0)
		num_threads = 1;

	/* calculate number of batches according to input */
	if (mode == CL_MP_THREAD_POOL_BATCH_PER_THREAD) {
		/* number of batchs equal to number of tasks */
		num_batches = num_threads;
		batch_size = (num_tasks + num_batches - 1) / num_batches;
	} else if (mode & CL_MP_THREAD_POOL_MODE_BIT) {
		/* number of batches */
		num_batches = mode & ~CL_MP_THREAD_POOL_MODE_BIT;
		batch_size = (num_tasks + num_batches - 1) / num_batches;
	} else {
		/* tasks per batch */
		batch_size = mode;
		num_batches = (num_tasks + batch_size - 1) / batch_size;
	}

	if (num_batches > num_tasks)
		num_batches = num_tasks;

	CL_ASSERT(num_batches * batch_size >= num_tasks);

	/* if not initialized or running with single thread, dont use thread pool */
	if (num_threads == 1) {
		unsigned result = 0;
		unsigned n;

		for (i = 0; i < num_tasks; i += batch_size) {
			n = num_tasks - i;
			if (n > batch_size)
				n = batch_size;
			result += !!pfn_callback(tasks + i, n, context, 0);
		}
		p_mp_thread_pool->result = result;

		goto exit;
	}

	pthread_mutex_lock(&p_mp_thread_pool->mutex);

	p_mp_thread_pool->pfn_callback = pfn_callback;
	p_mp_thread_pool->context = context;
	p_mp_thread_pool->num_tasks = num_tasks;
	p_mp_thread_pool->num_pending_tasks = num_tasks;
	p_mp_thread_pool->next_task = 0;
	p_mp_thread_pool->num_batches = num_batches;
	p_mp_thread_pool->batch_size = batch_size;
	p_mp_thread_pool->tasks = tasks;
	p_mp_thread_pool->result = 0;

	for (i = 0; i < num_batches; i++)
		cl_thread_pool_signal(&p_mp_thread_pool->thread_pool);

	pthread_cond_wait(&p_mp_thread_pool->cond, &p_mp_thread_pool->mutex);

	p_mp_thread_pool->tasks = NULL;

	pthread_mutex_unlock(&p_mp_thread_pool->mutex);

exit:
	return p_mp_thread_pool->result;
}

unsigned cl_mp_thread_pool_xarray_apply(IN cl_mp_thread_pool_t * const p_mp_thread_pool,
					IN void *tasks,
					IN unsigned num_tasks,
					IN unsigned size,
					IN cl_mp_thread_pool_mode_t mode,
					IN cl_pfn_mp_thread_pool_apply_t pfn_callback,
					IN void *context)
{
	void **p_tasks;
	unsigned i = 0;
	unsigned result = 0;

	if (!num_tasks)
		return result;

	p_tasks = malloc(num_tasks * sizeof(void *));
	if (!p_tasks)
		return num_tasks;

	for (i = 0; i < num_tasks; i++) {
		p_tasks[i] = tasks + (i * size);
	}

	result = cl_mp_thread_pool_array_apply(p_mp_thread_pool, p_tasks, num_tasks,
					       mode, pfn_callback, context);

	free(p_tasks);

	return result;
}

unsigned cl_mp_thread_pool_qmap_apply(IN cl_mp_thread_pool_t * const p_mp_thread_pool,
				      IN cl_qmap_t * const p_map,
				      IN cl_mp_thread_pool_mode_t mode,
				      IN cl_pfn_mp_thread_pool_apply_t pfn_callback,
				      IN void *context)
{
	cl_map_item_t *item;
	void **tasks;
	unsigned num_tasks;
	unsigned i = 0;
	unsigned result = 0;

	num_tasks = cl_qmap_count(p_map);
	if (!num_tasks)
		return result;

	tasks = malloc(num_tasks * sizeof(void *));
	if (!tasks)
		return num_tasks;

	for (item = cl_qmap_head(p_map); item != cl_qmap_end(p_map);
             item = cl_qmap_next(item)) {
		tasks[i++] = item;
	}

	result = cl_mp_thread_pool_array_apply(p_mp_thread_pool, tasks, num_tasks,
					       mode, pfn_callback, context);

	free(tasks);

	return result;
}
