%{
/*
 * Copyright (c) 2021 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Grammar of OSM device configuration parser.
 *
 * Author:
 *    Or Nechemia, Mellanox
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_DEVICE_CONF_PARSER_Y_Y
#include <opensm/osm_opensm.h>

#define CHECK_PARAMS_PHYSP 	(uint8_t)1
#define CHECK_PARAMS_ALL 	(uint8_t)0xF

static void parser_tmp_struct_reset();
#define RESET_BUFFER  parser_tmp_struct_reset()

typedef struct tmp_parser_struct_t_ {
	cl_list_t str_list;
	uint8_t num;
	boolean_t boolean;
} tmp_parser_struct_t;

static struct osm_opensm *p_osm;
char *__device_conf_parser_yylval = NULL;
extern FILE * __device_conf_parser_in;
tmp_parser_struct_t cap_tmp_parser_struct;
cl_list_t port_group_list;
osm_conf_params_physp_t cur_conf_params_physp;
osm_conf_params_node_t cur_conf_params_node;

int cap_column_num;
int cap_line_num;

static void parser_tmp_struct_init()
{
	cl_list_construct(&cap_tmp_parser_struct.str_list);
	cl_list_init(&cap_tmp_parser_struct.str_list, 10);
	cl_list_construct(&port_group_list);
	cl_list_init(&port_group_list, 10);
}

static void parser_tmp_struct_reset()
{
	cl_list_remove_all(&cap_tmp_parser_struct.str_list);
	cap_tmp_parser_struct.num = 0;
	cap_tmp_parser_struct.boolean = 0;
}

static void parser_tmp_struct_destroy()
{
	parser_tmp_struct_reset();
	cl_list_destroy(&cap_tmp_parser_struct.str_list);
	cl_list_remove_all(&port_group_list);
	cl_list_destroy(&port_group_list);
}

static void set_non_default_conf_params_physp(osm_physp_t * p_physp)
{
	if (cur_conf_params_physp.max_op_vl)
		p_physp->conf_params.max_op_vl = cur_conf_params_physp.max_op_vl;
}

static void set_non_default_conf_params_node(osm_node_t * p_node)
{
	if (cur_conf_params_node.sharp_enabled)
		p_node->conf_params.sharp_enabled =
		    cur_conf_params_node.sharp_enabled;
}

static void configure_physps_conf_params(osm_port_group_t * p_pgrp)
{
	cl_map_item_t *p_item;
	osm_port_group_member_t *p_grp_m;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	int i;

	for (p_item = cl_qmap_head(&p_pgrp->ports);
	     p_item != cl_qmap_end(&p_pgrp->ports);
	     p_item = cl_qmap_next(p_item)) {
		p_grp_m = (osm_port_group_member_t *)p_item;
		p_port = p_grp_m->port;

		if (p_port->p_node->sw) {
			for (i = 1; i < p_port->p_node->physp_tbl_size; i++) {
				p_physp = osm_node_get_physp_ptr(p_port->p_node, i);

				if (p_grp_m->physp_members[i])
					set_non_default_conf_params_physp(p_physp);
			}
			p_physp = &p_port->p_node->physp_table[0];
		} else {
			p_physp = p_port->p_physp;
			set_non_default_conf_params_physp(p_physp);
		}

		set_non_default_conf_params_node(p_physp->p_node);
	}
}

static void parser_conf_group_end()
{
	cl_list_iterator_t list_item;
	osm_port_group_t *p_cur_pgrp;

	for (list_item = cl_list_head(&port_group_list);
	     list_item != cl_list_end(&port_group_list);
	     list_item = cl_list_next(list_item)) {
		p_cur_pgrp = (osm_port_group_t *)cl_list_obj(list_item);
		if (p_cur_pgrp)
			configure_physps_conf_params(p_cur_pgrp);
	}

	memset((void *)&cur_conf_params_physp, 0, sizeof(cur_conf_params_physp));
	memset((void *)&cur_conf_params_node, 0, sizeof(cur_conf_params_node));
	cl_list_remove_all(&port_group_list);
}

static void clear_val(char **c)
{
	free(*c);
	*c = NULL;
}

static void yyerror(const char *format, ...)
{
	char s[256];
	va_list pvar;

	OSM_LOG_ENTER(&p_osm->log);

	va_start(pvar, format);
	vsnprintf(s, sizeof(s), format, pvar);
	va_end(pvar);

	OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9C07: "
		"Syntax error (line %d:%d): %s\n",
		cap_line_num, cap_column_num, s);
	OSM_LOG_EXIT(&p_osm->log);
}

int __device_conf_parser_wrap()
{
        return 1;
}

int __device_conf_parser_parse();
int __device_conf_parser_lex();

static void init_devices()
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	uint8_t port_num;
	cl_map_item_t *p_next;

	p_next = cl_qmap_head(&p_osm->subn.node_guid_tbl);
	while (p_next != cl_qmap_end(&p_osm->subn.node_guid_tbl)) {
		p_node = (osm_node_t *)p_next;
		p_next = cl_qmap_next(p_next);

		/* Initialize node configuration parameters values */
		p_node->conf_params.sharp_enabled = 0;

		for (port_num = 1; port_num < osm_node_get_num_physp(p_node);
		     port_num++) {
			p_physp = osm_node_get_physp_ptr(p_node, port_num);
			if (!p_physp)
				continue;

			/* Initialize port configuration parameters values */
			p_physp->conf_params.max_op_vl = 0;
		}
	}
}

int osm_parse_device_configuration_file(struct osm_opensm * osm_ptr)
{
	int res = 0;
	p_osm = osm_ptr;
	__device_conf_parser_in = NULL;
	parser_tmp_struct_init();
	cap_line_num = 0;
	cap_column_num = 0;

	init_devices();

	if (!p_osm->subn.opt.device_configuration_file) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 9C01: No device configuration file provided - "
			"Quit cap parser\n");
		res = 1;
		goto Exit;
	}

	__device_conf_parser_in = fopen(p_osm->subn.opt.device_configuration_file, "r");
	if (!__device_conf_parser_in) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 9C02: Device configuration file specified but not found\n");
		res = 1;
		goto Exit;
	}

        res = __device_conf_parser_parse();
	if (res)
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 9C03: Device configuration file parsing failed\n");
	else
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Device configuration "
			"file parsed successfully\n");

Exit:
	if (__device_conf_parser_in) {
		fclose(__device_conf_parser_in);
		__device_conf_parser_in = NULL;
	}
	clear_val(&__device_conf_parser_yylval);
	parser_tmp_struct_destroy();
	return res;
}
%}

%token COMMENT
%token COMMA
%token NUMBER
%token TEXT
%token TK_TRUE
%token TK_FALSE

%token PORT_CONF_START
%token PORT_CONF_END
%token PORT_GROUP_NAME
%token GROUP_MAX_OP_VL
%token GROUP_SHARP_ENABLED

%%

cap_policy: configurations;

configurations: /* empty */
		| configurations configuration
		;

comment: COMMENT {};

configuration:    conf_groups
		| comment
		;

/*
 * Parsing port configuration groups:
 * -------------------
 * port-conf
 *	port-group-name: NAME1
 *	...
 *	max-op-vl: 2
 * 	...
 * end-port-conf
 * port-conf
 *	...
 * end-port-conf
 */

conf_groups:	  conf_group
		| conf_groups conf_group
		;

conf_group: conf_group_start conf_group_entries conf_group_end;

conf_group_start: PORT_CONF_START {};

conf_group_end: PORT_CONF_END {
		parser_conf_group_end();
	};

conf_group_entries: /* empty */
		    | conf_group_entries conf_group_entry
		    ;

conf_group_entry:    port_group_name
		   | max_op_vl
		   | sharp_enabled
		   ;

/*
 * port_group_entry values:
 *	port_group_name
 *	max_op_vl
 */

port_group_name: port_group_name_start string_list {
		cl_list_iterator_t list_item;
		char *tmp_str;
		osm_port_group_t *p_port_grp;

		for (list_item = cl_list_head(&cap_tmp_parser_struct.str_list);
		     list_item != cl_list_end(&cap_tmp_parser_struct.str_list);
		     list_item = cl_list_next(list_item)) {
			tmp_str = (char *)cl_list_obj(list_item);
			if (tmp_str) {
				p_port_grp = osm_get_port_grp(p_osm, tmp_str);
				if (p_port_grp) {
					cl_list_insert_tail(&port_group_list, p_port_grp);
				} else {
					OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9C04: "
						"Port group with name (%s) does not exist\n",
						tmp_str);
				}
			}
			free(tmp_str);
		}
		cl_list_remove_all(&cap_tmp_parser_struct.str_list);
	};

port_group_name_start: PORT_GROUP_NAME {
		RESET_BUFFER;
	};

string_list:	single_string
		| string_list COMMA single_string
		;

single_string:	TEXT {
		char *tmp_str = strdup(__device_conf_parser_yylval);
		if (!tmp_str) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 9C05: Couldn't allocate memory for "
				"port group name, Aborting parsing\n");
			return 1;
		}
		cl_list_insert_tail(&cap_tmp_parser_struct.str_list, tmp_str);
	};

max_op_vl: max_op_vl_start number {
		cur_conf_params_physp.max_op_vl = cap_tmp_parser_struct.num;
	};

max_op_vl_start: GROUP_MAX_OP_VL {
		RESET_BUFFER;
	};

number: NUMBER {
		cap_tmp_parser_struct.num = strtoull(__device_conf_parser_yylval, NULL, 0);
		cap_tmp_parser_struct.boolean = (cap_tmp_parser_struct.num > 0);
	};

sharp_enabled: sharp_enabled_start bool {
		cur_conf_params_node.sharp_enabled = cap_tmp_parser_struct.boolean ? 2 : 1;
	};

sharp_enabled_start: GROUP_SHARP_ENABLED {
		RESET_BUFFER;
	};

bool: number | true | false {};

true: TK_TRUE {
		cap_tmp_parser_struct.boolean = 1;
	};

false: TK_FALSE {
		cap_tmp_parser_struct.boolean = 0;
	};

%%
