/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2019 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_eni_mad_rcv_process.
 * This object represents the ExtendedNodeInfo MADs Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <stdlib.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_MLNX_EXT_NODE_INFO_RCV_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_gpu.h>

static void mlnx_eni_rcv_process_new_gpu(osm_sm_t * sm,
					 osm_node_t * p_node)
{
	osm_gpu_t *p_gpu;
	osm_gpu_t *p_gpu_check;
	ib_net64_t node_guid = osm_node_get_node_guid(p_node);

	/* Create new GPU */
	p_gpu = osm_gpu_new(p_node);
	if (p_gpu == NULL) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 8903: "
			"Unable to create new GPU object\n");
		return;
	}

	p_gpu_check = (osm_gpu_t *) cl_qmap_insert(&sm->p_subn->gpu_guid_tbl,
						   node_guid, &p_gpu->map_item);
	if (p_gpu_check != p_gpu) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 8904: "
			"Unable to add node GUID:0x%016" PRIx64
			" to GPU table\n",
			cl_ntoh64(node_guid));

		free(p_gpu);
		return;
	}

	p_node->p_gpu = p_gpu;
}

static void mlnx_eni_rcv_process_gpu(osm_sm_t * sm,
				     osm_node_t * p_node,
				     osm_madw_t * p_madw)
{
	/*
	 * GPUs are created after discovery is finished. Meaning, all nodes
	 * exist at this point.
	 * Hence, if p_node->p_gpu exists, it means GPU was already discovered
	 * and no need to recreate it.
	 */
	if (!p_node->p_gpu)
		mlnx_eni_rcv_process_new_gpu(sm, p_node);

	/* For both exist or new node */
	if (sm->p_subn->sm_state == IB_SMINFO_STATE_MASTER)
		p_node->p_gpu->is_new = 1;
}

void osm_mlnx_eni_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	ib_mlnx_ext_node_info_t *p_payload;
	osm_eni_context_t *p_context;

	osm_node_t *p_node;
	ib_net64_t node_guid;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_eni_context_ptr(p_madw);
	p_payload = ib_smp_get_payload_ptr(p_smp);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_MLNX_EXTENDED_NODE_INFO);

	node_guid = p_context->node_guid;
	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_node = osm_get_node_by_guid(sm->p_subn, node_guid);
	if (!p_node || (p_node->node_info.node_type == IB_NODE_TYPE_SWITCH && !p_node->sw)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 8901: "
			"Received ExtendedNodeInfo for nonexistent node "
			"0x%" PRIx64 "\n", cl_ntoh64(node_guid));
		CL_PLOCK_RELEASE(sm->p_lock);
		goto Exit;
	}

	/* mark as successful */
	if (p_node->sw)
		p_node->sw->ar_data.pending_mads--;

	p_node->ext_node_info = *p_payload;

	if (osm_subn_is_nvlink_enabled(sm->p_subn) && osm_node_is_gpu(p_node))
		mlnx_eni_rcv_process_gpu(sm, p_node, p_madw);

	p_node->ext_node_info_updated = TRUE;

	CL_PLOCK_RELEASE(sm->p_lock);
Exit:
	OSM_LOG_EXIT(sm->p_log);
}
