/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	Module for NVLink discovery
 *	Set planes and rails
 *
 * Author:
 *	Or Nechemia, NVIDIA
 */
#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_NVLINK_C
#include <opensm/osm_lid_mgr.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_gpu.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_db_pack.h>

extern ib_api_status_t osm_ar_set_extended_node_info(osm_sm_t * sm, osm_node_t * p_node);

static void nvlink_get_alid_info(IN cl_map_item_t * p_map_item, IN void *context)
{
	osm_gpu_t *p_gpu = (osm_gpu_t *)p_map_item;
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_physp_t *p_physp;
	osm_nvlink_mad_context_t mad_context;
	ib_api_status_t status;
	uint8_t port_num;

	port_num = ib_node_info_get_local_port_num(&p_gpu->p_node->node_info);
	p_physp = osm_node_get_physp_ptr(p_gpu->p_node, port_num);
	if (!p_physp) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0908: "
			"Could not find local port for GPU 0x%"PRIx64"\n",
			cl_ntoh64(p_gpu->p_node->node_info.node_guid));
		return;
	}

	memset(&mad_context, 0, sizeof(mad_context));
	mad_context.node_guid = p_gpu->p_node->node_info.node_guid;
	mad_context.mad_method = OSM_AR_REQ_TYPE_GET;

	/* if we send GET for ALIDInfo, then our data is not updated */
	p_gpu->alid_info_updated = FALSE;

	status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp),
			     IB_MAD_ATTR_ALID_INFO, 0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0909: "
			"Getting ALIDInfo from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(p_physp->port_guid),	ib_get_err_str(status));

}

ib_api_status_t osm_nvlink_send_set_alid_info(osm_sm_t * sm, osm_gpu_t * p_gpu)
{
	osm_physp_t *p_physp;
	osm_nvlink_mad_context_t mad_context;
	uint8_t payload[IB_SMP_DATA_SIZE];
	ib_alid_info_t *p_alid_info_data = (ib_alid_info_t *)payload;
	ib_api_status_t status;
	uint8_t port_num;
	uint8_t num_alids = p_gpu->p_node->new_ext_node_info.anycast_lid_table_top;

	/* No need for an update */
	if (p_gpu->alid_info_updated &&
	    !memcmp(p_gpu->calc_alid_info.alid_block, p_gpu->subn_alid_info.alid_block,
		    sizeof(ib_alid_entry_t) * num_alids))
		return IB_SUCCESS;

	memset(&mad_context, 0, sizeof(mad_context));
	mad_context.node_guid = p_gpu->p_node->node_info.node_guid;
	mad_context.mad_method = OSM_AR_REQ_TYPE_SET;

	/* Prepare ALIDInfo SET request */
	memcpy(p_alid_info_data, &p_gpu->calc_alid_info, sizeof(ib_alid_info_t));

	port_num = ib_node_info_get_local_port_num(&p_gpu->p_node->node_info);
	p_physp = osm_node_get_physp_ptr(p_gpu->p_node, port_num);
	OSM_ASSERT(p_physp);

	status = osm_req_set(sm, osm_physp_get_dr_path_ptr(sm, p_physp),
			     p_physp, payload, sizeof(ib_alid_info_t),
			     IB_MAD_ATTR_ALID_INFO, 0, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 090B: "
			"Setting ALIDInfo on 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_gpu_get_guid(p_gpu)), ib_get_err_str(status));
	return status;
}

static void nvlink_validate_sw_plane_db(IN osm_lid_mgr_t * p_mgr)
{
	cl_qlist_t guids;
	osm_db_guid_elem_t *p_item;
	uint8_t plane;
	boolean_t valid_entry = TRUE;

	OSM_LOG_ENTER(p_mgr->p_log);

	cl_qlist_init(&guids);


	if (osm_db_guids(p_mgr->nvlink_data.p_guid2planes, &guids)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR A80F: "
			"could not get guid list\n");
		goto Exit;
	}

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_guid2plane_get(p_mgr->nvlink_data.p_guid2planes, p_item->guid, &plane)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR A810: "
				"could not get plane for guid:0x%016" PRIx64 "\n",
				p_item->guid);
		} else {
			/* Plane should be non zero */
			if (plane < OSM_NVLINK_MIN_PLANES || plane > OSM_NVLINK_MAX_PLANES) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR A811: "
					"Plane %u of guid:0x%016" PRIx64 " is illegal. "
					"Valid range is [%u,%u].\n",
					plane, p_item->guid, OSM_NVLINK_MIN_PLANES,
					OSM_NVLINK_MAX_PLANES);

				valid_entry = FALSE;
			}

			if (!valid_entry &&
			    osm_db_guid2lid_delete(p_mgr->nvlink_data.p_guid2planes,
						   p_item->guid))
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR A812: "
					"failed to delete plane entry for "
					"guid:0x%016" PRIx64 "\n",
					p_item->guid);
		}
		free(p_item);
	}

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
}

static void nvlink_init_sw_plane_db(IN osm_lid_mgr_t * p_mgr)
{
	osm_sm_t *p_sm = p_mgr->sm;
	osm_opensm_t *p_osm = p_mgr->p_subn->p_osm;

	OSM_LOG_ENTER(p_sm->p_log);


	/* Initialize the given database */
	p_mgr->nvlink_data.p_guid2planes = osm_db_domain_init(&p_osm->db, "guid2planes");
	if (!p_mgr->nvlink_data.p_guid2planes) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR A808: "
			"Error initializing switch GUID to plane persistent database\n");
		return;
	}

	if (osm_db_restore(p_mgr->nvlink_data.p_guid2planes)) {
#ifndef __WIN__
		/*
		 * When Windows is BSODing, it might corrupt files that
		 * were previously opened for writing, even if the files
		 * are closed, so we might see corrupted guid2mkey file.
		 */
		if (p_osm->subn.opt.exit_on_fatal) {
			osm_log(&p_osm->log, OSM_LOG_SYS,
				"FATAL: Error restoring switch GUID to plane "
				"persistent database\n");
			return;
		} else
#endif
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR A809: Error restoring switch GUID to plane "
				"persistent database\n");
	}

	nvlink_validate_sw_plane_db(p_mgr);

	OSM_LOG_EXIT(p_sm->p_log);
}

static ib_api_status_t nvlink_iterate_within_plane_validate_port(osm_subn_t * p_subn,
								 osm_switch_t * p_cur_sw,
								 uint8_t port_num,
								 osm_switch_t ** queue_plane,
								 uint32_t * queue_plane_in,
								 uint32_t * queue_plane_out)
{
	osm_physp_t *p_physp, *p_remote_physp;
	osm_node_t *p_remote_node;
	ib_api_status_t status = IB_ERROR;
	uint8_t remote_plane, cur_plane;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	p_physp = osm_node_get_physp_ptr(p_cur_sw->p_node, port_num);
	if (!p_physp)
		goto SkipPort;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp)
		goto SkipPort;

	/*
	 * Check port:
	 * 1. For service port, remote is of another plane, no need to set it to this plane.
	 * 2. If remote port is a switch port, add remote switch to this plane queue,
	 *    as it belongs to current plane and should keep iterating through.
	 */
	p_remote_node = p_remote_physp->p_node;

	/*
	 * Planes are set to switches.
	 * GPU port plane is as the plane of its remote switch.
	 */
	if (osm_node_get_type(p_remote_node) != IB_NODE_TYPE_SWITCH)
		goto SkipPort;

	if ((ib_mlnx_epi_get_is_fnm_port(&p_physp->ext_port_info) &&
	     !ib_mlnx_epi_get_is_fnm_port(&p_physp->p_remote_physp->ext_port_info)) ||
	    (!ib_mlnx_epi_get_is_fnm_port(&p_physp->ext_port_info) &&
	     ib_mlnx_epi_get_is_fnm_port(&p_physp->p_remote_physp->ext_port_info))) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Found switch 0x%"PRIx64" port %u is %s port while remote "
			"is %s port. Fail NVLink planes configuration.\n",
			cl_ntoh64(osm_node_get_node_guid(p_cur_sw->p_node)), port_num,
			ib_mlnx_epi_get_is_fnm_port(&p_physp->ext_port_info) ? "FNM" : "Mission",
			ib_mlnx_epi_get_is_fnm_port(&p_physp->p_remote_physp->ext_port_info) ?
			    "FNM" : "Mission");
		status = IB_INVALID_PORT;
		goto Exit;
	}

	remote_plane = p_remote_node->sw->nvlink_sw_data.plane;
	cur_plane = p_cur_sw->nvlink_sw_data.plane;
	if (ib_mlnx_epi_get_is_fnm_port(&p_physp->ext_port_info)) {
		/*
		 * Planes are indexed from 1 to N and in preprocess step
		 * all to 0. Hence plane indicates if remote was visited.
		 */
		if (remote_plane) {
			/* Remote was visited and has different plane, as it should */
			if (cur_plane != remote_plane)
				goto SkipPort;

			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Found switch 0x%"PRIx64" and remote switch "
				"0x%"PRIx64" are part of the same plane %u, "
				"although connected through FNM port. "
				"Reassign all planes.\n",
				cl_ntoh64(osm_node_get_node_guid(p_cur_sw->p_node)),
				cl_ntoh64(osm_node_get_node_guid(p_remote_node)),
				cur_plane);

			goto Exit;
		}
	} else {
		/*
		 * Planes are indexed from 1 to N and in preprocess step
		 * all to 0. Hence plane indicates if remote was visited.
		 */
		if (remote_plane) {
			/* Remote was visited and has the same plane, as it should */
			if (cur_plane == remote_plane)
				goto SkipPort;

			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Found switch 0x%"PRIx64" plane %u is"
				" different than remote switch 0x%"PRIx64
				" plane %u. Reassign all planes.\n",
				cl_ntoh64(osm_node_get_node_guid(p_cur_sw->p_node)),
				cur_plane,
				cl_ntoh64(osm_node_get_node_guid(p_remote_node)),
				remote_plane);

			goto Exit;
		}

		/*
		 * For ports that are not service ports, if remote is a switch,
		 * add it to this plane queue, as it is a part of current plane.
		 */
		queue_plane[(*queue_plane_in)++] = p_remote_node->sw;
	}

SkipPort:
	status = IB_SUCCESS;
Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return status;
}

static ib_api_status_t nvlink_iterate_within_plane(osm_subn_t * p_subn,
						   osm_switch_t * p_start_sw,
						   uint8_t plane)
{
	osm_lid_mgr_t *p_mgr = &p_subn->p_osm->sm.lid_mgr;
	osm_log_t *p_log = &p_subn->p_osm->log;
	osm_switch_t **queue_plane = NULL;
	osm_switch_t *p_cur_sw;
	uint32_t queue_plane_out = 0, queue_plane_in = 0;
	uint8_t port_num, num_ports, db_plane;
	uint64_t cur_guid;
	ib_api_status_t status = IB_ERROR;

	/* use for the internal iteration until discovers all port within the plane */
	queue_plane = (osm_switch_t **)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
					      sizeof(osm_switch_t *));
	if (!queue_plane) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A801: "
			"Cannot allocate memory for sw Queue\n");
		goto Exit;
	}

	/* Start iteration from input switch */
	queue_plane[queue_plane_in++] = p_start_sw;

	/*
	 * Starting BFS from input p_sw, iterate over the subnet in order to find
	 * all switches belong to the same plane as the input switch.
	 *
	 * Service port (FNM port) connects between planes (used for IB communication
	 * and avoided by NVLink communication)
	 * Hence, when gets to one, should not move forward with that switch.
	 */
	while (queue_plane_out < queue_plane_in) {
		p_cur_sw = queue_plane[queue_plane_out++];
		cur_guid = cl_ntoh64(osm_node_get_node_guid(p_cur_sw->p_node));

		/* Validate DB plane is identical to current iterated plane */
		if (!osm_db_guid2plane_get(p_mgr->nvlink_data.p_guid2planes, cur_guid, &db_plane) &&
		    (plane != db_plane)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A80E: "
				"DB plane for switch 0x%" PRIx64 " is %u, "
				"while discovered through switch of plane %u. "
				"Invalidate NVLink planes cache and reassign planes.\n",
				cur_guid, db_plane, plane);

			status = IB_ERROR;
			goto Exit;
		} else if (!db_plane) {
			osm_db_guid2plane_set(p_mgr->nvlink_data.p_guid2planes, cur_guid, plane);
		}

		/* Set current plane to current switch */
		p_cur_sw->nvlink_sw_data.plane = plane;

		/* Continue discovering through switch ports */
		num_ports = osm_node_get_num_physp(p_cur_sw->p_node);
		for (port_num = 1; port_num < num_ports; port_num++) {
			status = nvlink_iterate_within_plane_validate_port(
				    p_subn, p_cur_sw, port_num,
				    queue_plane, &queue_plane_in, &queue_plane_out);
			if (status)
				goto Exit;
		}
	}

	status = IB_SUCCESS;
Exit:
	free(queue_plane);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return status;
}

static void nvlink_init_sw_plane(cl_map_item_t * item, void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *)item;
	p_sw->nvlink_sw_data.plane = OSM_NVLINK_NO_PLANE;
}

static void nvlink_plane_discover_preprocess(osm_subn_t * p_subn)
{
	/* Zero all Switch previous planes */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, nvlink_init_sw_plane, NULL);
}

static void nvlink_invalidate_plane_cache(osm_subn_t * p_subn,
					  osm_switch_t ** queue,
					  uint32_t *queue_in)
{
	osm_lid_mgr_t *p_mgr = &p_subn->p_osm->sm.lid_mgr;
	osm_switch_t *p_sw;
	cl_map_item_t *item;

	osm_db_clear(p_mgr->nvlink_data.p_guid2planes);

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *) item;

		p_sw->nvlink_sw_data.plane = OSM_NVLINK_NO_PLANE;
	
		queue[(*queue_in)++] = p_sw;
	}
}

static inline uint8_t nvlink_update_next_free_plane(boolean_t * is_used,
						    uint8_t next_free_plane)
{
	while (next_free_plane <= OSM_NVLINK_MAX_PLANES && is_used[next_free_plane])
		next_free_plane++;

	if (next_free_plane <= OSM_NVLINK_MAX_PLANES)
		is_used[next_free_plane] = TRUE;

	return next_free_plane;
}

static ib_api_status_t nvlink_plane_discover(osm_subn_t * p_subn, boolean_t is_cache_valid)
{
	osm_lid_mgr_t *p_mgr = &p_subn->p_osm->sm.lid_mgr;
	osm_log_t *p_log = &p_subn->p_osm->log;
	osm_switch_t **queue;
	osm_switch_t *p_cur_sw;
	uint32_t queue_in = 0, queue_out = 0;
	uint8_t cur_plane = 1;
	uint64_t cur_guid;
	boolean_t is_used[OSM_NVLINK_MAX_PLANES + 1];
	uint8_t next_free_plane = 1;
	cl_map_item_t *item;
	ib_api_status_t status = IB_ERROR;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	queue = (osm_switch_t **)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
					sizeof(osm_switch_t *));
	if (!queue) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A802: "
			"Cannot allocate memory for sw Queue\n");
		goto Exit;
	}

	nvlink_plane_discover_preprocess(p_subn);
	memset(is_used, 0, sizeof(is_used));

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_cur_sw = (osm_switch_t *) item;

		/*
		 * If switch does not have an entry on switch plane cache, skip it.
		 *	It will get a plane in one of the following ways:
		 *	1. While iterating from another switch that exists on DB.
		 *	2. While iterating on the queue holding the switches that
		 *	   could not be discovered from the DB switches.
		 */
		cur_guid = cl_ntoh64(osm_node_get_node_guid(p_cur_sw->p_node));
		if (osm_db_guid2plane_get(p_mgr->nvlink_data.p_guid2planes, cur_guid, &cur_plane)) {
			queue[queue_in++] = p_cur_sw;
			continue;
		}

		is_used[cur_plane] = TRUE;

		/*
		 * Start BFS from current switch.
		 * Any switch visited while iterating should have the same plane
		 * as current switch.
		 * If switch with cached plane different than current plane is found
		 * on this iteration, invalidate the cache and reassign all planes.
		 */
		is_cache_valid = !nvlink_iterate_within_plane(p_subn, p_cur_sw, cur_plane);
	}

	if (!is_cache_valid) {
		/* This function also inserts all of the switches to the queue */
		nvlink_invalidate_plane_cache(p_subn, queue, &queue_in);
		memset(is_used, 0, sizeof(is_used));
		next_free_plane = 1;
	}

	/* 
	 * At this point, queue holds all the switches that are missing from plane DB.
	 * Some may have gotten a plane while iterating by the BFS from switches that
	 * exist on DB.
	 * On this loop we set planes to those switches that are yet to be reached.
	 */
	while (queue_out < queue_in) {
		p_cur_sw = queue[queue_out++];

		/* Skip switch that was already set to a plane */
		if (p_cur_sw->nvlink_sw_data.plane)
			continue;

		next_free_plane = nvlink_update_next_free_plane(is_used, next_free_plane);
		if (next_free_plane > OSM_NVLINK_MAX_PLANES) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR A803: "
				"Found more than maximal allowed planes (%u). "
				"Fail NVLink plane discovery.\n",
				OSM_NVLINK_MAX_PLANES);
			goto Exit;
		}

		/* Set current plane to current switch */
		p_cur_sw->nvlink_sw_data.plane = next_free_plane;
		osm_db_guid2plane_set(p_mgr->nvlink_data.p_guid2planes,
				      cl_ntoh64(osm_node_get_node_guid(p_cur_sw->p_node)),
				      next_free_plane);

		/* Start iterating on a new plane, from next unvisited switch */
		nvlink_iterate_within_plane(p_subn, p_cur_sw, next_free_plane);
	}

	status = IB_SUCCESS;
Exit:
	free(queue);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return status;
}

static void nvlink_rail_filter_config_mark_port_mask(ib_rail_filter_config_t * p_rail_filter_config,
						     osm_nvlink_port_type_t port_type,
						     uint8_t port_num,
						     uint8_t mask_block_num)
{
	uint8_t *p_port_mask = NULL;
	uint8_t mask_index, mask_offset_bit;
	uint16_t mask_len, port_index;

	CL_ASSERT(port_type < IB_PORT_TYPE_MAX);

	switch (port_type) {
	case IB_PORT_TYPE_INGRESS:
		mask_len = IB_RAIL_FILTER_CFG_INGRESS_MASK_LEN;
		p_port_mask = p_rail_filter_config->ingress_port_mask;
		break;
	case IB_PORT_TYPE_EGRESS:
		mask_len = IB_RAIL_FILTER_CFG_EGRESS_MASK_LEN;
		p_port_mask = p_rail_filter_config->egress_port_mask;
		break;
	default:
		return;
	}

	/* Assert that input mask block num matches the relevant block for input port number */
	OSM_ASSERT(mask_block_num == port_num / mask_len);

	port_index = port_num % mask_len;

	/*
	 * Mask index is calculated as port mask is a 128 or 256 bitmask field.
	 * Meaning port indices start from the MSB of the field. Hence, convert
	 * port_index with (mask_len - 1 - port_index).
	 */
	mask_index = port_index / (uint8_t)IB_RAIL_FILTER_CFG_PORT_MASK_BLOCK_SIZE;

	mask_offset_bit = port_index % IB_RAIL_FILTER_CFG_PORT_MASK_BLOCK_SIZE;
	/* Convert to start from byte's MSB */
	mask_offset_bit = IB_RAIL_FILTER_CFG_PORT_MASK_BLOCK_SIZE - 1 - mask_offset_bit;

	p_port_mask[mask_index] |= ((uint8_t)1 << mask_offset_bit);
}

static void nvlink_rail_filter_config_send_block(osm_subn_t * p_subn,
						 osm_switch_t * p_sw,
						 uint8_t rail,
						 uint8_t ingress_block,
						 uint8_t egress_block)
{
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	osm_rail_filter_cfg_context_t context;
	osm_physp_t *p_physp0;
	ib_rail_filter_config_t *p_calc_rail_filter_cfg =
	    osm_nvlink_get_rail_filter_mad(p_sw, rail, ingress_block, egress_block, OSM_RAIL_FILTER_CALC);
	ib_rail_filter_config_t *p_subn_rail_filter_cfg =
	    osm_nvlink_get_rail_filter_mad(p_sw, rail, ingress_block, egress_block, OSM_RAIL_FILTER_SUBN);
	uint32_t attr_mod;
	ib_api_status_t status;

	if (!p_subn->need_update && !p_sw->need_update &&
	    !memcmp(p_calc_rail_filter_cfg, p_subn_rail_filter_cfg,
		    sizeof(ib_rail_filter_config_t)))
		return;

	p_physp0 = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp0) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A805: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	/* Using ingress port mask, no need for port number on attr_mod */
	attr_mod = (ingress_block << IB_RAIL_FILTER_ATTR_MOD_INGRESS_BLOCK_SHIFT) |
	    (egress_block << IB_RAIL_FILTER_ATTR_MOD_EGRESS_BLOCK_SHIFT);

	memset(&context, 0, sizeof(context));
	context.node_guid = p_sw->p_node->node_info.node_guid;
	context.mad_method = IB_MAD_METHOD_SET;
	context.ingress_block = ingress_block;
	context.egress_block = egress_block;
	context.rail = rail;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp0),
			     p_physp0, (const uint8_t *)p_calc_rail_filter_cfg,
			     sizeof(ib_rail_filter_config_t), IB_MAD_ATTR_RAIL_FILTER_CONFIG,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp0->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS)
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A806: "
			"Setting RailFilterConfig of rail %u (ingress block %u, egress block %u) "
			"to 0x%" PRIx64 " failed (%s)\n",
			rail, ingress_block, egress_block,
			cl_ntoh64(p_physp0->port_guid), ib_get_err_str(status));
}

static void nvlink_rail_filter_preprocess(cl_map_item_t * item, void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *)item;
	ib_rail_filter_config_t *p_calc_rail_filter_cfg;
	uint32_t i, num_rail_filter_mads;

	num_rail_filter_mads = OSM_NVLINK_SN_NUM_RAILS *
			       p_sw->nvlink_sw_data.max_ingress_block *
			       p_sw->nvlink_sw_data.max_egress_block;

	for (i = 0; i < num_rail_filter_mads; i++) {
		p_calc_rail_filter_cfg = &p_sw->nvlink_sw_data.calc_rail_filter_cfg[i];

		memset(p_calc_rail_filter_cfg, 0, sizeof(ib_rail_filter_config_t));

		p_calc_rail_filter_cfg->vl_mask |= IB_RAIL_FILTER_CFG_NVLINK_VL_MASK;
		p_calc_rail_filter_cfg->mc_uc_enable |= (IB_RAIL_FILTER_CFG_MC_ENABLE_BIT |
						     IB_RAIL_FILTER_CFG_UC_ENABLE_BIT);
	}
}

static void nvlink_set_rail_filter_config(cl_map_item_t * item, void *context)
{
	osm_subn_t *p_subn = (osm_subn_t *)context;
	osm_switch_t *p_sw = (osm_switch_t *)item;
	uint8_t max_ingress_block, max_egress_block;
	uint8_t ingress_block, egress_block, rail;

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_RAIL_FIILTER_SUPPORTED_BIT))
		return;

	max_ingress_block = p_sw->nvlink_sw_data.max_ingress_block;
	max_egress_block = p_sw->nvlink_sw_data.max_egress_block;

	for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++)
		for (ingress_block = 0; ingress_block < max_ingress_block; ingress_block++) {
			for (egress_block = 0; egress_block < max_egress_block; egress_block++) {
				nvlink_rail_filter_config_send_block(p_subn, p_sw, rail,
								     ingress_block, egress_block);
		}
	}
}

static void nvlink_rail_filter_config_rail_add_port_num(osm_switch_t * p_sw,
							uint8_t rail,
							uint8_t port_num)
{
	ib_rail_filter_config_t *p_remote_rail_filter_cfg;
	uint8_t max_block, port_in_block, port_e_block;
	uint8_t ingress_block, egress_block;

	OSM_ASSERT(rail != OSM_NVLINK_NO_RAIL);

	port_in_block = port_num / IB_RAIL_FILTER_CFG_INGRESS_MASK_LEN;
	port_e_block = port_num / IB_RAIL_FILTER_CFG_EGRESS_MASK_LEN;

	/* Set rail egress port masks across all MADs with relevant egress mask block */
	max_block = p_sw->nvlink_sw_data.max_ingress_block;
	for (ingress_block = 0; ingress_block < max_block; ingress_block++) {
		p_remote_rail_filter_cfg = osm_nvlink_get_rail_filter_mad(
		    p_sw, rail, ingress_block, port_e_block, OSM_RAIL_FILTER_CALC);

		nvlink_rail_filter_config_mark_port_mask(
		    p_remote_rail_filter_cfg, IB_PORT_TYPE_EGRESS, port_num, port_e_block);
	}

	/* Set rail ingress port masks across all MADs with relevant ingress mask block */
	max_block = p_sw->nvlink_sw_data.max_egress_block;
	for (egress_block = 0; egress_block < max_block; egress_block++) {
		p_remote_rail_filter_cfg = osm_nvlink_get_rail_filter_mad(
		    p_sw, rail, port_in_block, egress_block, OSM_RAIL_FILTER_CALC);

		nvlink_rail_filter_config_mark_port_mask(
		    p_remote_rail_filter_cfg, IB_PORT_TYPE_INGRESS, port_num, port_in_block);
	}
}

static void nvlink_rail_configuration(osm_subn_t * p_subn)
{
	cl_map_item_t *p_item;
	osm_gpu_t *p_gpu;
	osm_physp_t *p_physp;
	uint8_t num_ports, plane, port_num, remote_port_num;
	uint8_t visited_plane_count[OSM_NVLINK_MAX_PLANES + 1];
	osm_node_t *p_remote_node;
	osm_switch_t *p_remote_sw;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* Zero previously calculated RailFilterConfig MADs */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, nvlink_rail_filter_preprocess, NULL);

	/* Single Node with 2 switches (low priority topology) */
	for (p_item = cl_qmap_head(&p_subn->gpu_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->gpu_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_gpu = (osm_gpu_t *)p_item;

		/*
		 * Counts how many time plane was visited for this GPU, so each GPU port
		 * visited that is part of this plane will get a new rail sequentially
		 */
		memset(visited_plane_count, 0, sizeof(visited_plane_count));

		num_ports = osm_node_get_num_physp(p_gpu->p_node);
		for (port_num = 1; port_num < num_ports; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_gpu->p_node, port_num);
			if (!p_physp)
				continue;

			p_remote_node = p_physp->p_remote_physp->p_node;
			if (osm_node_get_type(p_remote_node) != IB_NODE_TYPE_SWITCH)
				continue;

			/*
			 * At this point, each port is a part of some plane.
			 * Iterate over the ports and set rails for each plane,
			 * increasing on each plane revisit.
			 */
			plane = osm_gpu_get_port_plane(p_gpu, port_num);
			/* Plane 0 is invalid plane */
			OSM_ASSERT(plane);

			p_physp->rail = visited_plane_count[plane];
			visited_plane_count[plane]++;

			/* When getting to maximal rail, start from 0 again */
			if (visited_plane_count[plane] >= OSM_NVLINK_SN_NUM_RAILS)
				visited_plane_count[plane] = 1;

			/* On nvlink topologies, remote port of GPU is a switch port */
			p_remote_sw = p_remote_node->sw;

			remote_port_num = p_physp->p_remote_physp->port_num;
			nvlink_rail_filter_config_rail_add_port_num(p_remote_sw, p_physp->rail,
								    remote_port_num);
		}
	}

	/* Iterate over switches and send switch's req RailFilterConfig for each rail */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, nvlink_set_rail_filter_config, p_subn);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

void osm_nvlink_sw_data_init(osm_switch_t * p_sw)
{
	uint8_t num_ports, max_ingress_block, max_egress_block;
	uint32_t num_mads;

	num_ports = p_sw->num_ports;

	/* Initialize nvlink MAD configuration structures */
	max_ingress_block = (num_ports / IB_RAIL_FILTER_CFG_INGRESS_MASK_LEN) + 1;
	max_egress_block = (num_ports / IB_RAIL_FILTER_CFG_EGRESS_MASK_LEN) + 1;
	p_sw->nvlink_sw_data.max_ingress_block = max_ingress_block;
	p_sw->nvlink_sw_data.max_egress_block = max_egress_block;

	num_mads = OSM_NVLINK_SN_NUM_RAILS * max_ingress_block * max_egress_block;
	p_sw->nvlink_sw_data.subn_rail_filter_cfg =
	    (ib_rail_filter_config_t *)calloc(num_mads, sizeof(ib_rail_filter_config_t));
	p_sw->nvlink_sw_data.calc_rail_filter_cfg =
	    (ib_rail_filter_config_t *)calloc(num_mads, sizeof(ib_rail_filter_config_t));
}

void osm_nvlink_sw_data_destroy(osm_nvlink_sw_data_t * nvlink_sw_data)
{
	free(nvlink_sw_data->subn_rail_filter_cfg);
	free(nvlink_sw_data->calc_rail_filter_cfg);
}

ib_rail_filter_config_t * osm_nvlink_get_rail_filter_mad(osm_switch_t * p_sw,
							 uint8_t rail,
							 uint8_t ingress_block,
							 uint8_t egress_block,
							 boolean_t is_calc)
{
	uint16_t rail_offset, in_offset;

	CL_ASSERT((rail < OSM_NVLINK_SN_NUM_RAILS) &&
		   (ingress_block < p_sw->nvlink_sw_data.max_ingress_block) &&
		   (egress_block < p_sw->nvlink_sw_data.max_egress_block));

	rail_offset = rail *
	    p_sw->nvlink_sw_data.max_ingress_block *
	    p_sw->nvlink_sw_data.max_egress_block;
	in_offset = ingress_block * p_sw->nvlink_sw_data.max_egress_block;

	if (is_calc)
		return &p_sw->nvlink_sw_data.calc_rail_filter_cfg[rail_offset +
								  in_offset +
								  egress_block];

	return &p_sw->nvlink_sw_data.subn_rail_filter_cfg[rail_offset + in_offset + egress_block];
}

/* Set NVLINK parameters */
ib_api_status_t osm_nvlink_init(osm_subn_t * p_subn)
{

	osm_lid_mgr_t *p_mgr = &p_subn->p_osm->sm.lid_mgr;
	uint8_t plane;
	uint16_t next_range_start;

	/* As global lids take range [1, OSM_NVLINK_GLOBAL_LID_RANGE_LEN] */
	p_mgr->nvlink_data.alid_range_start = OSM_NVLINK_GLOBAL_LID_RANGE_LEN + 1;
	p_mgr->nvlink_data.num_alids_per_gpu = OSM_NVLINK_NUM_ALIDS_PER_GPU;

	next_range_start = p_mgr->nvlink_data.alid_range_start + OSM_NVLINK_NVLID_RANGE_LEN;
	for (plane = 1; plane <= OSM_NVLINK_MAX_PLANES; plane++) {
		p_mgr->nvlink_data.p_start_ranges_per_plane[plane] = next_range_start;
		next_range_start += OSM_NVLINK_NVLID_RANGE_LEN;
	}

	nvlink_init_sw_plane_db(p_mgr);

	return IB_SUCCESS;
}

ib_api_status_t osm_nvlink_set_gpu_extended_node_info(osm_sm_t * sm, osm_gpu_t * p_gpu)
{
	osm_node_t *p_node = p_gpu->p_node;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(sm->p_log);

	/* Set ENI for GPUs should be before setting alid_tbl_top */
	p_node->new_ext_node_info.anycast_lid_table_top =
	    sm->lid_mgr.nvlink_data.num_alids_per_gpu;

	if ((p_node->ext_node_info.anycast_lid_table_top ==
	    p_node->new_ext_node_info.anycast_lid_table_top) &&
	    p_node->ext_node_info_updated && !sm->p_subn->need_update)
		goto Exit;

	status = osm_ar_set_extended_node_info(sm, p_node);

Exit:
	OSM_LOG_EXIT(sm->p_log);

	return status;
}

/* Retrieves information used later by lid manager */
ib_api_status_t osm_nvlink_discovery(osm_subn_t * p_subn)
{
	osm_sm_t *sm = &p_subn->p_osm->sm;
	ib_api_status_t status = IB_SUCCESS;

	OSM_ASSERT(sm->lid_mgr.nvlink_data.p_guid2planes);

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	cl_qmap_apply_func(&p_subn->gpu_guid_tbl, nvlink_get_alid_info, sm);

	/* Attempt to set planes according to plane DB */
	status = nvlink_plane_discover(p_subn, TRUE);

	/*
	 * Failed to set planes according to DB.
	 * IB_INVALID_PORT error means:
	 * 	a link was found in which one side is FNM port and the other is not.
	 * 	TODO: For NVLink multinode, do this validation where SM configures the FNM ports.
	 *
	 * Any other error, DB is invalid, SM should rediscover the planes.
	 */
	if (status == IB_INVALID_PORT) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A813: "
			"Failed to reassign NVLink planes\n");
		status = IB_ERROR;
		goto Exit;

	} else if (status) {
		status = nvlink_plane_discover(p_subn, FALSE);
		if (status) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A807: "
				"Failed to reassign NVLink planes\n");
			status = IB_ERROR;
			goto Exit;
		}
	}

	/* store the guid to plane table in persistent db */
	osm_db_store(sm->lid_mgr.nvlink_data.p_guid2planes, p_subn->opt.fsync_high_avail_files);

	nvlink_rail_configuration(p_subn);

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);

	return status;
}
