/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2015 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 * This file content handles receiving of adjacent router table block
 * This object represents the RouterTable Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTER_TABLE_RCV_C
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_router.h>

/*
 * WE ONLY RECEIVE GET only responses
 */
void osm_adj_rtr_table_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	ib_net64_t port_guid;
	osm_ri_context_t *p_context;
	osm_router_t *p_router;
	uint32_t block_num, top_block_num;
	cl_disp_msgid_t msg_id = CL_DISP_MSGID_NONE;
	uint16_t top = 0;
	void *p_payload;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);
	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_ri_context_ptr(p_madw);
	port_guid = p_context->port_guid;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	/* the router should exist in router table from NI rcv flow */
	p_router = (osm_router_t *)cl_qmap_get(&sm->p_subn->rtr_guid_tbl,
					       port_guid);

	if (PF(p_router ==
		(osm_router_t *)cl_qmap_end(&sm->p_subn->rtr_guid_tbl))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7900: "
			"Failed to find router object for port 0x%016"
			PRIx64 " in router adjacent table\n",
                        cl_ntoh64(port_guid));
		goto Exit;
	}

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	switch (p_smp->attr_id) {
		case IB_MAD_ATTR_RTR_ADJ_SITE_LOCAL_TBL:
			top = p_router->router_info.adjacent_site_local_subnets_tbl_top;
			top_block_num = RTR_TBL_NUM_BLOCKS(top,
							   IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK);
			msg_id = OSM_MSG_MAD_RTR_ADJ_SITE_LOCAL_TBL;
			break;
		case IB_MAD_ATTR_RTR_ADJ_ROUTER_LID_INFO:
			top = p_router->router_info.adjacent_subnets_router_lid_info_table_top;
			top_block_num = RTR_TBL_NUM_BLOCKS(top,
							   IB_ROUTER_ADJ_SUBNETS_LID_INFO_ELEMENTS);
			msg_id = OSM_MSG_MAD_RTR_ADJ_ROUTER_LID_INFO;
			break;
		case IB_MAD_ATTR_RTR_ROUTER_LID_TBL:
			top = p_router->global_flid_end;
			top_block_num = RTR_TBL_NUM_BLOCKS(top, IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM);
			msg_id = OSM_MSG_MAD_RTR_ROUTER_LID_TBL;
			break;
		default:
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7902: "
				"Received unsupported MAD Attribute 0x%x(%s) "
				" for switch node 0x%" PRIx64 "\n",
				cl_ntoh16(p_smp->attr_id),
				ib_get_sm_attr_str(p_smp->attr_id),
				cl_ntoh64(osm_node_get_node_guid(
						osm_router_get_node_ptr(p_router))));
			goto Exit;
	}

	block_num = cl_ntoh32(p_smp->attr_mod);
	if (PF(block_num > top_block_num)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7901: "
			"In MAD %s: got invalid block_num %u > top block %u (top %u) "
			"for port 0x%" PRIx64 "\n",
			osm_get_disp_msg_str(msg_id), block_num, top_block_num, top,
			cl_ntoh64(port_guid));
		goto Exit;
	}

	p_payload = ib_smp_get_payload_ptr(p_smp);

	switch (p_smp->attr_id) {
		case IB_MAD_ATTR_RTR_ADJ_SITE_LOCAL_TBL:
			osm_dump_adjacent_router_block(sm->p_log, port_guid, block_num,
						       (ib_rtr_adj_table_block_t *)p_payload,
						       FILE_ID, OSM_LOG_DEBUG);

			osm_adj_router_tbl_set(&p_router->adj_router_tbl, block_num,
					       (ib_rtr_adj_table_block_t *)p_payload);
			break;
		case IB_MAD_ATTR_RTR_ADJ_ROUTER_LID_INFO:
			osm_adj_router_lid_info_set(p_router, block_num,
				 (ib_rtr_adj_subnets_lid_info_block_t *)p_payload);
			break;
		case IB_MAD_ATTR_RTR_ROUTER_LID_TBL:
			/*
			 * Router lid table SET is performed for local router LIDs,
			 * router lid table GET is performed for global router LIDs.
			 * The response is stored in different tables.
			 */
			osm_router_lid_table_set(p_router, block_num,
						 (ib_rtr_router_lid_tbl_block_t *)p_payload);
			break;
		default:
			/* already verified before */
			break;
	}

Exit:
	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}
