/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of TENANT virtualization.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <iba/ib_types.h>
#include <opensm/osm_base.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_TENANT_MGR_C
#include <opensm/osm_tenant_mgr.h>

/* -------------------------------------------------------------------------------------------------
 *
 * Macros for manipulating underlying data structs.
 */

#define TENANT_GUIDLIST_INIT(P_LIST)				cl_hashmap_init_uint64(P_LIST)
#define TENANT_GUIDLIST_DESTROY(P_LIST)				cl_hashmap_destroy(P_LIST)
#define TENANT_GUIDLIST_INSERT_GUID(P_LIST, GUID, TENANT)	cl_hashmap_insert(P_LIST, (uintptr_t)GUID, (uintptr_t)TENANT)
#define TENANT_GUIDLIST_REMOVE_GUID(P_LIST, GUID)		cl_hashmap_remove(P_LIST, GUID)
#define TENANT_GUIDLIST_REMOVE_ALL(P_LIST)			cl_hashmap_remove_all(P_LIST)
#define TENANT_GUIDLIST_FIND(P_LIST, GUID)			cl_hashmap_find(P_LIST, GUID)

/* -------------------------------------------------------------------------------------------------
 *
 * Internal guid list functions.
 */

int tenant_guidlist_init(IN osm_tenant_guid_list_t *p_list)
{
	return TENANT_GUIDLIST_INIT(p_list);
}

void tenant_guidlist_destroy(IN osm_tenant_guid_list_t *p_list)
{
	TENANT_GUIDLIST_DESTROY(p_list);
}

uint64_t tenant_guidlist_find(IN osm_tenant_guid_list_t *p_list, IN uint64_t guid)
{
	return TENANT_GUIDLIST_FIND(p_list, guid);
}

/* -------------------------------------------------------------------------------------------------
 *
 * Internal tenant functions
 */

static void tenant_fatal_error(IN osm_tenant_mgr_t *p_tenant_mgr)
{
	OSM_LOG(p_tenant_mgr->p_log, OSM_LOG_ERROR, "ERR AE04: fatal error - exiting\n");
	exit(1);
}

static osm_tenant_t *tenant_alloc(IN osm_subn_t *p_subn)
{
	osm_tenant_t		*p_tenant = NULL;

	if ((p_tenant = calloc(1, sizeof(osm_tenant_t))) != NULL) {
		tenant_guidlist_init(&p_tenant->virtual_guids);
		tenant_guidlist_init(&p_tenant->physical_guids);
	} else {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR AE05: allocation error\n");
	}

	return p_tenant;
}

static void tenant_free(IN osm_tenant_t *p_tenant)
{
	if (p_tenant != NULL) {
		tenant_guidlist_destroy(&p_tenant->virtual_guids);
		tenant_guidlist_destroy(&p_tenant->physical_guids);
		free(p_tenant);
	}
}

/* -------------------------------------------------------------------------------------------------
 *
 * Parser interface.
 */

extern void	__tenant_parser_set_in(FILE *);
extern void	__tenant_parser_restart(FILE *);
extern void	__tenant_parser_set_lineno(int);
extern void	__tenant_parser__flush_buffer(FILE *);
extern int	__tenant_parser_parse(osm_tenant_parser_t *);

static int tenant_parse_file(IN osm_tenant_mgr_t *p_tenant_mgr, IN osm_tenant_t *p_tenant, FILE *std_tenant)
{
	int			status = -1;
	osm_tenant_parser_t	parser;
	struct timeval		start, stop;
	float			elapsed_time;

	memset((void *)&parser, 0, sizeof(parser));
	parser.p_log = p_tenant_mgr->p_log;
	parser.p_tenant = p_tenant;

	/*
	 * Parse the file.
	 */
	__tenant_parser_restart(std_tenant);
	__tenant_parser_set_in(std_tenant);
	__tenant_parser__flush_buffer(NULL);
	__tenant_parser_set_lineno(1);

	gettimeofday(&start, NULL);
	status = __tenant_parser_parse(&parser);
	gettimeofday(&stop, NULL);

	if (status == 0) {
		elapsed_time = (double)(stop.tv_usec - start.tv_usec) / 1000000 + (double)(stop.tv_sec - start.tv_sec);
		OSM_LOG(p_tenant_mgr->p_log, OSM_LOG_DEBUG, "tenants_policy_file parsing complete.  Time = %11.8f secs\n", elapsed_time);
	}

	return status;
}

/* -------------------------------------------------------------------------------------------------
 *
 * API interface.
 */

int osm_tenant_guidlist_insert_guid(IN osm_tenant_guid_list_t *p_list, IN uint64_t guid, IN uint64_t map_id)
{
	return TENANT_GUIDLIST_INSERT_GUID(p_list, guid, map_id);
}

int osm_tenant_mgr_validate(IN osm_sm_t *p_sm, IN uint64_t virtual_guid, IN uint64_t physical_guid, uint16_t vport_index)
{
	int		status;
	uint64_t	virtual_group_id = 0;
	uint64_t	physical_group_id = 0;
	osm_tenant_t	*p_tenant = p_sm->tenant_mgr.p_tenant;

	if (!p_sm->p_subn->opt.tenants_policy_enabled) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "tenant mgr is not enabled for validation\n");
		return 1;
	}

	/*
	 * If the physical port is managed and the virtual port is managed, then the GUIDS must
	 * be in the same group.  There are 5 possibilities:
	 *
	 *           is managed?      | same
	 *	virtual  |  physical  | group  |  status     |  reason
	 *	---------+------------+--------+-------------+------------------------------
	 *	  yes    |    yes     |  yes   |  good       |  both ports in same group
	 *	  yes    |    yes     |  no    |  violation  |  ports in different groups
	 *	  yes    |    no      |  n/a   |  violation  |  only one port managed
	 *	  no     |    yes     |  n/a   |  violation  |  only one port managed
	 *	  no     |    no      |  n/a   |  good       |  neither port managed
	 */
	virtual_group_id = tenant_guidlist_find(&p_tenant->virtual_guids, virtual_guid);
	physical_group_id = tenant_guidlist_find(&p_tenant->physical_guids, physical_guid);

	if ((status = (virtual_group_id == physical_group_id)) == 0) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR,
			"ERR AE01: tenant violation: virt=0x%016lx (%ld) phys=0x%016lx (%ld)  vport=%d\n",
			virtual_guid, virtual_group_id, physical_guid, physical_group_id, vport_index);

		osm_report_notice_1601(p_sm->p_subn,
			virtual_guid, virtual_group_id,
			physical_guid, physical_group_id,
			vport_index);
	} else {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"tenant validation successful: virt=0x%016lx (%ld) phys=0x%016lx (%ld)  vport=%d\n",
			virtual_guid, virtual_group_id, physical_guid, physical_group_id, vport_index);
	}

	return status;
}

int osm_tenant_mgr_rescan(IN osm_subn_t *p_subn)
{
	int			status = 0;
	uint32_t		crc = 0;
	osm_sm_t		*p_sm = NULL;
	osm_tenant_t		*p_tenant = NULL;
	osm_tenant_t		*p_tenant_old = NULL;
	osm_tenant_mgr_t	*p_tenant_mgr = NULL;
	FILE			*std_tenant = NULL;

	/*
	 * If the tenant mgr is not enabled, just return.
	 */
	if (!p_subn->opt.tenants_policy_enabled) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG, "tenant mgr is not enabled for rescanning\n");
		return 0;
	}

	/*
	 * It is a fatal error if we can't read the file.
	 */
	p_sm = &p_subn->p_osm->sm;
	p_tenant_mgr = &p_sm->tenant_mgr;

	if ((std_tenant = fopen(p_tenant_mgr->filename, "r")) == NULL) {
		OSM_LOG(p_tenant_mgr->p_log, OSM_LOG_ERROR, "ERR AE03: tenants policy file - %s\n", strerror(errno));
		tenant_fatal_error(p_tenant_mgr);
	}

	/*
	 * Check if the file has changed.  If it has, parse it.
	 */
	crc = osm_get_crc_value(p_sm, OSM_CRC_TENANTS_POLICY_CONF_FILE);
	if (crc != p_tenant_mgr->crc) {
		OSM_LOG(p_tenant_mgr->p_log, OSM_LOG_DEBUG, "rescanning tenants policy file\n");
		if ((p_tenant = tenant_alloc(p_subn)) == NULL) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR AE06: allocation error\n");
			tenant_fatal_error(p_tenant_mgr);
		}

		if (tenant_parse_file(p_tenant_mgr, p_tenant, std_tenant) != 0) {
			fclose(std_tenant);
			tenant_fatal_error(p_tenant_mgr);
		}

		/*
		 * The file was parsed correctly.  Update state and signal the virt_mgr to revalidate.
		 */
		p_tenant_old = p_tenant_mgr->p_tenant;
		p_tenant_mgr->p_tenant = p_tenant;
		p_tenant_mgr->crc = crc;
		tenant_free(p_tenant_old);
		status = 1;
	}

	fclose(std_tenant);
	return status;
}

void osm_tenant_mgr_config_change(osm_subn_t *p_subn, void *p_value)
{
	boolean_t		value = *(boolean_t *)p_value;
	osm_tenant_mgr_t	*p_tenant_mgr = NULL;

	p_tenant_mgr = &p_subn->p_osm->sm.tenant_mgr;

	if (value) {
		/*
		 * We were disabled, but now we are enabled.  We need to setup for rescanning.
		 */
		p_tenant_mgr->crc = (uint32_t)-1;
	} else {
		/*
		 * We were enabled, but now we are disabled.  We need to clean up our resources.
		 */
		tenant_free(p_tenant_mgr->p_tenant);
		p_tenant_mgr->p_tenant = NULL;
	}

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO, "tenant mgr is now %s\n", value ? "enabled" : "disabled");
}

ib_api_status_t osm_tenant_mgr_init(IN osm_tenant_mgr_t *p_tenant_mgr, osm_sm_t *p_sm)
{
	ib_api_status_t		status = IB_SUCCESS;
	osm_subn_t		*p_subn = p_sm->p_subn;

	/*
	 * Initialize the tenant mgr.  We set the crc to -1 so that the first time through
	 * we rescan the conf file correctly.
	 */
	memset((void *)p_tenant_mgr, 0, sizeof(osm_tenant_mgr_t));
	p_tenant_mgr->p_log = p_sm->p_log;
	p_tenant_mgr->crc = (uint32_t)-1;

	/*
	 * Get configuration filename.
	 */
	if (p_subn->opt.tenants_policy_file != NULL) {
		p_tenant_mgr->filename = p_subn->opt.tenants_policy_file;
	} else {
		p_tenant_mgr->filename = OSM_DEFAULT_TENANTS_POLICY_CONF_FILE;
	}

	OSM_LOG(p_sm->p_log, OSM_LOG_INFO, "tenant mgr is %s\n", p_subn->opt.tenants_policy_enabled ? "enabled" : "disabled");
	return status;
}

void osm_tenant_mgr_destroy(IN osm_tenant_mgr_t *p_tenant_mgr, osm_sm_t *p_sm)
{
	tenant_free(p_tenant_mgr->p_tenant);
	memset((void *)p_tenant_mgr, 0, sizeof(osm_tenant_mgr_t));

	OSM_LOG(p_sm->p_log, OSM_LOG_INFO, "osm_tenant_mgr_destroy complete\n");
}
