/*
 * Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2008 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2013 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of Up Down Algorithm using ranking & Min Hop
 *      Calculation functions
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <ctype.h>
#include <complib/cl_debug.h>
#include <complib/cl_qmap.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UCAST_UPDN_C
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_mcast_mgr.h>
#include <opensm/osm_root_detection.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_helper.h>

extern int osm_apply_func_on_port_group_nodes(char *pgrp_name, int pfn_func(void *, uint64_t , char *),
                                              void *cxt, char *pi, void *p_osm, boolean_t goto_node);

/* //////////////////////////// */
/*  Local types                 */
/* //////////////////////////// */

/* direction */
typedef enum updn_switch_dir {
	UP = 0,
	DOWN
} updn_switch_dir_t;


typedef struct updn_minhop_data {
	cl_list_item_t list;
	struct osm_switch *sw;
	uint64_t id;
	updn_switch_dir_t dir;
	uint8_t visited;
	void *prev_priv;
} updn_minhop_data_t;

typedef uint64_t color_t;

#define NO_COLOR 0
#define MAX_LEAF_COLOR 0x8000000000000000
#define MAX_COLOR UINT64_MAX


typedef struct updn_node {
	cl_list_item_t list;
	osm_switch_t *sw;
	uint8_t num_hca;
	boolean_t is_leaf;
	uint32_t *dlid_routes;
	uint32_t *dlid_sw_routes;
	uint32_t total_routes;
	color_t color;
} updn_node_t;

typedef struct endport {
	osm_port_t *p_port;
	updn_node_t *adj_sw;
	uint8_t path_bit;
} endport_t;

typedef struct group_offset {
	int offset;
	color_t color;
} group_offset_t;


/* updn structure */
typedef struct updn {
	unsigned num_roots;
	osm_opensm_t *p_osm;
	osm_subn_t *p_subn;
	unsigned num_nodes;
	updn_node_t **ranked_nodes;
	group_offset_t *offset_arr;
	int offset_arr_len;
	endport_t *port_arr;
	int port_arr_len;
} updn_t;

typedef struct subn_sparce_graph {
	osm_subn_t *p_subn;
	flow_graph_t *p_flow_graph;

} subn_sparce_graph_t;

/*
 * This structure is a bundle of data structures
 * that are needed for FLID routing algorithm
 */
typedef struct flid_routing_block {
	osm_ar_subgroup_t *ar_groups;
	uint8_t *flid_routed;
	uint8_t *static_ports;
	uint16_t *flids_routed_ar;
} flid_routing_block_t;

static void flid_routing_block_free(flid_routing_block_t *p_route_block)
{
	free(p_route_block->ar_groups);
	free(p_route_block->flid_routed);
	free(p_route_block->static_ports);
	free(p_route_block->flids_routed_ar);
}

static boolean_t flid_routing_block_init(flid_routing_block_t *p_route_block, uint32_t global_flids)
{
	memset(p_route_block, 0, sizeof(*p_route_block));
	p_route_block->ar_groups = malloc(global_flids * sizeof(osm_ar_subgroup_t));
	p_route_block->flid_routed = malloc(global_flids * sizeof(uint8_t));
	p_route_block->static_ports = malloc(global_flids * sizeof(uint8_t));
	p_route_block->flids_routed_ar = malloc(global_flids * sizeof(uint16_t));

	if (!p_route_block->ar_groups ||
	    !p_route_block->flid_routed ||
	    !p_route_block->static_ports ||
	    !p_route_block->flids_routed_ar) {
		flid_routing_block_free(p_route_block);
		return FALSE;
	}

	return TRUE;
}

static void flid_routing_block_clear(flid_routing_block_t *p_route_block, uint32_t global_flids)
{
	memset(p_route_block->ar_groups, 0, global_flids * sizeof(osm_ar_subgroup_t));
	memset(p_route_block->flid_routed, 0, global_flids * sizeof(uint8_t));
	memset(p_route_block->static_ports, 0, global_flids * sizeof(uint8_t));
	memset(p_route_block->flids_routed_ar, 0, global_flids * sizeof(uint16_t));
}

static inline char *direction_str(updn_switch_dir_t direction)
{
	return direction == UP ? "UP" : "DOWN";
}

static inline osm_subn_t *updn_get_subnet(IN updn_t * p_updn)
{
	return p_updn->p_subn;
}

/* This function returns direction based on rank and guid info of current &
   remote ports */
static updn_switch_dir_t updn_get_dir(unsigned cur_rank, unsigned rem_rank,
				      uint64_t cur_id, uint64_t rem_id,
				      uint8_t hops)
{
	/* HACK: comes to solve root nodes connection, in a classic subnet root nodes do not connect
	   directly, but in case they are we assign to root node an UP direction to allow UPDN to discover
	   the subnet correctly (and not from the point of view of the last root node).
	 */
	if (!cur_rank && !rem_rank)
		return UP;

	if (cur_rank < rem_rank) {
		/* allow one hop between "siblings" on a way down */
		if (hops > (rem_rank - cur_rank + 1))
			return UP;
		return DOWN;
	} else if (cur_rank > rem_rank) {
		return UP;
	} else {
		/* Equal rank, decide by id number, bigger == UP direction */
		if (cur_id > rem_id || hops > 1)
			return UP;
		else
			return DOWN;
	}
}

/**********************************************************************
 * This function does the bfs of min hop table calculation by guid index
 * as a starting point.
 **********************************************************************/
static void updn_bfs_by_node(IN void * task, IN void * context, uint32_t id)
{
	osm_switch_t * p_sw = task;
	subn_sparce_graph_t * p_subn_sparce_graph = context;
	osm_subn_t * p_subn = p_subn_sparce_graph->p_subn;
	flow_graph_t *p_flow_graph = p_subn_sparce_graph->p_flow_graph;
	osm_log_t *p_log = &p_subn->p_osm->log;
	uint8_t pn_rem;
	cl_qlist_t list;
	uint16_t lid;
	updn_switch_dir_t next_dir, current_dir;
	updn_minhop_data_t *u;
	unsigned int vertex_idx;
	flow_vertex_t *p_vertex, *p_remote_vertex;
	flow_vertex_t *vertices = p_flow_graph->vertices;
	flow_edge_t *edges = p_flow_graph->edges;
	flow_edge_t *p_edge = NULL;
	unsigned int idx, start_edges_index, end_edges_index;

	OSM_LOG_ENTER(p_log);

	lid = osm_node_get_base_lid(p_sw->p_node, 0);
	lid = cl_ntoh16(lid);
	osm_switch_set_hops(p_sw, lid, 0, 0);

	OSM_LOG(p_log, OSM_LOG_DEBUG,
		"Starting from switch - port GUID 0x%" PRIx64 " lid %u\n",
		cl_ntoh64(p_sw->p_node->node_info.port_guid), lid);


	u = (updn_minhop_data_t *) (p_sw->priv);
	u = &u[id];
	u->dir = UP;

	/* Update list with the new element */
	cl_qlist_init(&list);
	cl_qlist_insert_tail(&list, &u->list);

	/* BFS the list till no next element */
	while (!cl_is_qlist_empty(&list)) {
		u = (updn_minhop_data_t *)cl_qlist_remove_head(&list);
		u->visited = 0;	/* cleanup */
		current_dir = u->dir;
		osm_physp_t *p_physp = &u->sw->p_node->physp_table[0];

		vertex_idx = p_flow_graph->sw_lid_to_vertex_mapping[cl_ntoh16(p_physp->port_info.base_lid)];
		p_vertex = &vertices[vertex_idx];
		start_edges_index = p_vertex->start_edges_index;
		end_edges_index = p_vertex->end_edges_index;

		/* Go over all ports of the switch and find unvisited remote nodes */
		for (idx = start_edges_index; idx < end_edges_index; idx++) {
			osm_node_t *p_remote_node;
			updn_minhop_data_t *rem_u;
			uint8_t current_min_hop, remote_min_hop,
			    set_hop_return_value;
			osm_switch_t *p_remote_sw;
			p_edge = &edges[idx];
			p_remote_vertex = &vertices[p_edge->adj_index];
			pn_rem = p_edge->adj_port_num;

			if (!p_remote_vertex || !p_remote_vertex->p_sw) {
				continue;
			}

			p_remote_sw = p_remote_vertex->p_sw;
			p_remote_node = p_remote_sw->p_node;
			rem_u = (updn_minhop_data_t *) p_remote_sw->priv;
			rem_u = &rem_u[id];
			/* Decide which direction to mark it (UP/DOWN) */
			next_dir = updn_get_dir(u->sw->rank, rem_u->sw->rank,
						u->sw->id, rem_u->sw->id, 1);

			/* Check if this is a legal step : the only illegal step is going
			   from DOWN to UP */
			if ((current_dir == DOWN) && (next_dir == UP)) {
				OSM_LOG(p_log, OSM_LOG_DEBUG,
					"Avoiding move from 0x%016" PRIx64
					" to 0x%016" PRIx64 "\n",
					cl_ntoh64(osm_node_get_node_guid(u->sw->p_node)),
					cl_ntoh64(osm_node_get_node_guid(p_remote_node)));
				/* Illegal step */
				continue;
			}
			/* Set MinHop value for the current lid */
			current_min_hop = osm_switch_get_least_hops(u->sw, lid);
			/* Check hop count if better insert into list && update
			   the remote node Min Hop Table */
			remote_min_hop =
			    osm_switch_get_hop_count(p_remote_sw, lid, pn_rem);
			if (current_min_hop + 1 < remote_min_hop) {
				set_hop_return_value =
				    osm_switch_set_hops(p_remote_sw, lid,
							pn_rem,
							current_min_hop + 1);
				if (set_hop_return_value) {
					OSM_LOG(p_log, OSM_LOG_ERROR, "ERR AA01: "
						"Invalid value returned from set min hop is: %d\n",
						set_hop_return_value);
				}
				/* Check if remote port has already been visited */
				if (!rem_u->visited) {
					/* Insert updn_switch item into the list */
					rem_u->dir = next_dir;
					rem_u->visited = 1;
					cl_qlist_insert_tail(&list,
							     &rem_u->list);
				}
			}
		}
	}

	OSM_LOG_EXIT(p_log);
}


static int updn_bfs_by_node_par(IN void ** task, uint32_t task_len, void * context,
				IN unsigned thread_id)
{
	osm_switch_t ** p_sw_arr = (osm_switch_t **)task;
	osm_switch_t * p_sw;
	int i;

	for (i = 0; i < task_len; i++) {
		p_sw = p_sw_arr[i];
		updn_bfs_by_node(p_sw, context, thread_id);
	}

	return 0;
}


static void updn_rank_adjustment(IN updn_t * p_updn, cl_qlist_t *p_list)
{
	osm_subn_t *p_subn;
	osm_log_t *p_log;
	osm_switch_t *p_sw;
	osm_physp_t *p_physp;
	osm_physp_t *p_remote_physp;
	struct updn_node *u;
	struct updn_node *remote_u;
	struct updn_node *first_adj;
	int num_ports;
	int port_num;
	int new_rank;
	int num_adj_leaves;

	p_subn = updn_get_subnet(p_updn);
	p_log = &p_updn->p_osm->log;

	/*
	 * The following algorithm iterate over list of switches that have
	 * local maximum rank (possibly leaf switches), and checks if their rank
	 * should be promoted (by decreasing their current rank by 2).
	 *
	 * In order for switch to be promoted, the following conditions should apply:
	 * 1. The switch does not have CAs connected to it.
	 *    otherwise it might introduce down/up turns.
	 * 2. Current rank has to be at least 3.
	 *    otherwise that switch will become root in contrast to root selecting
	 *    decision or result with negative rank.
	 * 3. Switch has to be connected to at least two other switches with hosts
	 */

	while (!cl_is_qlist_empty(p_list)) {
		u = (struct updn_node *)cl_qlist_remove_head(p_list);
		p_sw = u->sw;
		num_ports = p_sw->num_ports;

		if (p_sw->rank < 3 || u->num_hca) {
			osm_subn_add_leaf(p_subn, p_sw);
			u->is_leaf = TRUE;
			continue;
		}

		/*
		 * Current switch is at rank >= 3, and does not have hosts, and
		 * is local maximum rank.
		 * If it has at least two neighboring switches with hosts, we
		 * promote it so traffic between the two switches can go through
		 * it.
		 */
		new_rank = p_sw->rank;
		first_adj = NULL;
		num_adj_leaves = 0;
		for (port_num = 1; port_num < num_ports; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
			if (!p_physp)
				continue;

			p_remote_physp = p_physp->p_remote_physp;
			if (!p_remote_physp)
				continue;

			if (!p_remote_physp->p_node->sw)
				continue;

			remote_u = p_remote_physp->p_node->sw->priv;
			if (u == remote_u || !remote_u->num_hca)
				continue;

			/*
			 * If it is the first neighboring switch with hosts,
			 * store it, and continue to the next switch.
			 */
			if (num_adj_leaves == 0) {
				first_adj = remote_u;
				num_adj_leaves = 1;
				continue;
			}

			/*
			 * Avoid counting the first adjacent switch with CAs
			 * more than once due to parallel links.
			 */
			if (first_adj == remote_u)
				continue;

			/*
			 * Mark the first adjacent switch with hosts as leaf.
			 */
			if (num_adj_leaves == 1) {
				first_adj->is_leaf = TRUE;
				osm_subn_add_leaf(p_subn, first_adj->sw);
				new_rank = p_sw->rank - 2;
				num_adj_leaves = 2;
			}

			/*
			 * Switch denoted by remote_u is a leaf.
			 */
			if (!remote_u->is_leaf) {
				remote_u->is_leaf = TRUE;
				osm_subn_add_leaf(p_subn, remote_u->sw);
			}
		}

		if (p_sw->rank != new_rank) {
			/* Rank changed */
			OSM_LOG(p_log, OSM_LOG_VERBOSE,
				"Changing rank of switch 0x%" PRIx64 " from %d to %d\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)), p_sw->rank, new_rank);
			p_sw->rank = new_rank;
		} else {
			/*
			 * Keep the current switch as leaf as it was previously.
			 */
			u->is_leaf = TRUE;
			osm_subn_add_leaf(p_subn, p_sw);
		}
	}
}


/* NOTE : PLS check if we need to decide that the first */
/*        rank is a SWITCH for BFS purpose */
static int updn_subn_rank(IN updn_t * p_updn)
{
	osm_switch_t *p_sw;
	osm_physp_t *p_physp, *p_remote_physp;
	cl_qlist_t list_bfs;
	cl_qlist_t list_leaves;
	cl_map_item_t *item;
	struct updn_node *u, *remote_u;
	uint8_t num_ports, port_num;
	osm_log_t *p_log = &p_updn->p_osm->log;
	unsigned max_rank = 0;
	osm_subn_t *p_subn;
	boolean_t is_leaf;

	OSM_LOG_ENTER(p_log);
	cl_qlist_init(&list_bfs);
	cl_qlist_init(&list_leaves);

	p_subn = updn_get_subnet(p_updn);

	/* add all roots to the list */
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		u = p_sw->priv;
		if (!p_sw->rank)
			cl_qlist_insert_tail(&list_bfs, &u->list);
	}

	/* BFS the list till it's empty */
	while (!cl_is_qlist_empty(&list_bfs)) {
		u = (struct updn_node *)cl_qlist_remove_head(&list_bfs);
		is_leaf = TRUE;
		/* Go over all remote nodes and rank them (if not already visited) */
		p_sw = u->sw;
		num_ports = p_sw->num_ports;
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Handling switch GUID 0x%" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
		for (port_num = 1; port_num < num_ports; port_num++) {
			ib_net64_t port_guid;

			/* Current port fetched in order to get remote side */
			p_physp =
			    osm_node_get_physp_ptr(p_sw->p_node, port_num);

			if (!p_physp)
				continue;

			p_remote_physp = p_physp->p_remote_physp;

			/*
			   make sure that all the following occur on p_remote_physp:
			   1. The port isn't NULL
			   2. It is a switch
			 */
			if (!p_remote_physp)
				continue;

			if (p_remote_physp->p_node->sw) {
				remote_u = p_remote_physp->p_node->sw->priv;
				port_guid = p_remote_physp->port_guid;

				if (remote_u->sw->rank > u->sw->rank)
					is_leaf = FALSE;

				if (remote_u->sw->rank > u->sw->rank + 1) {
					/* Rank assigned to switch only once */
					OSM_ASSERT(remote_u->sw->rank ==
						   OSM_SW_NO_RANK);
					remote_u->sw->rank = u->sw->rank + 1;
					max_rank = remote_u->sw->rank;
					cl_qlist_insert_tail(&list_bfs,
							     &remote_u->list);
					OSM_LOG(p_log, OSM_LOG_DEBUG,
						"Rank of port GUID 0x%" PRIx64
						" = %u\n", cl_ntoh64(port_guid),
						remote_u->sw->rank);
				}
				continue;
			}

			if (osm_node_get_type(p_remote_physp->p_node) != IB_NODE_TYPE_CA)
				continue;

			if (osm_phsyp_get_special_port_type(p_remote_physp))
				continue;

			u->num_hca++;
		}

		if (!is_leaf)
			continue;

		/*
		 * This switch is possibly a leaf switch (local maximum rank)
		 */
		if (p_subn->opt.routing_flags & OSM_ROUTING_FLAGS_RANK_ADJUSTMENT) {
			cl_qlist_insert_tail(&list_leaves, &u->list);
		} else {
			osm_subn_add_leaf(p_subn, u->sw);
			u->is_leaf = TRUE;
		}
	}

	if (!cl_is_qlist_empty(&list_leaves)) {
		updn_rank_adjustment(p_updn, &list_leaves);
	}

	/* Print Summary of ranking */
	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Subnet ranking completed. Max Node Rank = %d\n", max_rank);
	OSM_LOG_EXIT(p_log);
	return 0;
}


/* hack: preserve min hops entries to any other root switches */
static void updn_clear_non_root_hops(IN osm_subn_t * p_subn, osm_switch_t * sw)
{
	osm_port_t *port;
	unsigned i;

	for (i = 0; i < sw->num_hops; i++)
		if (sw->hops[i]) {
			port = osm_get_port_by_lid_ho(p_subn, i);
			if (!port || !port->p_node->sw ||
			    port->p_node->sw->rank)
				memset(sw->hops[i], 0xff, sw->num_ports);
		}
}

static void osm_flow_restore_graph_privs(flow_graph_t **pp_graph)
{
	flow_graph_t *p_graph = *pp_graph;
	flow_vertex_t *vertices;
	void **prev_privs;
	osm_switch_t *p_sw;
	int i;

	if (!p_graph)
		return;

	if (p_graph->prev_privs) {
		vertices = p_graph->vertices;
		prev_privs = p_graph->prev_privs;
		for (i = 0; i < p_graph->vertices_num; i++) {
			p_sw = vertices[i].p_sw;
			p_sw->priv = prev_privs[i];
		}
	}
}

int updn_calculate_min_hop_table(IN osm_subn_t *p_subn)
{
	osm_log_t *p_log = &p_subn->p_osm->log;
	osm_switch_t *p_sw;
	cl_map_item_t *item;
	uint32_t i, number_of_threads;
	uint32_t sw_idx, successful_alloc_switch = 0;
	int res = 0;
	osm_ucast_mgr_t * p_mgr = &p_subn->p_osm->sm.ucast_mgr;
	updn_minhop_data_t *u = NULL;
	subn_sparce_graph_t subn_sparce_graph;

	subn_sparce_graph.p_subn = p_subn;
	subn_sparce_graph.p_flow_graph = osm_flow_build_graph(p_subn);

	osm_flow_restore_graph_privs(&subn_sparce_graph.p_flow_graph);

	OSM_LOG_ENTER(p_log);

	/* Go over all the switches in the subnet - for each init their Min Hop
	   Table */
	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Init Min Hop Table of all switches [\n");

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		/* Clear Min Hop Table */
		if (p_subn->opt.connect_roots)
			updn_clear_non_root_hops(p_subn, p_sw);
		else
			osm_switch_clear_hops(p_sw);
	}

	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Init Min Hop Table of all switches ]\n");

	/* Now do the BFS for each port  in the subnet */
	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"BFS through all port guids in the subnet [\n");

	if (cl_mp_thread_pool_state(&p_mgr->thread_pool) != CL_INITIALIZED)
		number_of_threads = 1;
	else
		number_of_threads = p_mgr->thread_pool.num_threads;

	/* create sw->priv array per thread */
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl), successful_alloc_switch = 0;
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item), successful_alloc_switch++) {

		p_sw = (osm_switch_t *)item;
		u = (updn_minhop_data_t *) calloc (1, sizeof(updn_minhop_data_t) * number_of_threads);
		if (u == NULL) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR AA11: "
				"Not enough memory for priv objects\n");
			res = -1;
			goto Exit;
		}
		for (i = 0; i < number_of_threads; i++) {
			u[i].sw = p_sw;
			u[i].prev_priv = p_sw->priv;
		}
		/* change the pointer */
		p_sw->priv = u;
	}

	if (number_of_threads == 1) {
		for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
		     item != cl_qmap_end(&p_subn->sw_guid_tbl);
		     item = cl_qmap_next(item)) {
			p_sw = (osm_switch_t *)item;
			updn_bfs_by_node(p_sw, &subn_sparce_graph, 0);
		}
		res = 0;
		goto Exit;
	}

	/* Multithreaded part of the algorithm */
	cl_mp_thread_pool_qmap_apply(&p_mgr->thread_pool, &p_subn->sw_guid_tbl,
				     CL_MP_THREAD_POOL_BATCH_PER_THREAD,
				     updn_bfs_by_node_par, &subn_sparce_graph);
	
Exit:
	/* restore previous value for switches that changed */
	for (sw_idx = 0, item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     sw_idx < successful_alloc_switch && item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     sw_idx++, item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		u = (updn_minhop_data_t *) p_sw->priv;
		/* restore previous data */
		p_sw->priv = u[0].prev_priv;
		/* free previously allocated data for minhop calculation */
		free(u);
	}

	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"BFS through all port guids in the subnet ]\n");
	osm_flow_destroy_graph(&subn_sparce_graph.p_flow_graph);
	/* Cleanup */
	OSM_LOG_EXIT(p_log);
	return res;
}

static int updn_build_lid_matrices(IN updn_t * p_updn)
{
	int status;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);

	OSM_LOG_ENTER(&p_updn->p_osm->log);

	OSM_LOG(&p_updn->p_osm->log, OSM_LOG_VERBOSE,
		"Ranking all port guids in the list\n");
	if (!p_updn->num_roots) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA0A: "
			"No guids were provided or number of guids is 0\n");
		status = -1;
		goto _exit;
	}

	/* Check if it's not a switched subnet */
	if (cl_is_qmap_empty(&p_subn->sw_guid_tbl)) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA0B: "
			"This is not a switched subnet, cannot perform UPDN algorithm\n");
		status = -1;
		goto _exit;
	}

	/* Rank the subnet switches */
	if (updn_subn_rank(p_updn)) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA0E: "
			"Failed to assign ranks\n");
		status = -1;
		goto _exit;
	}

	/* After multiple ranking need to set Min Hop Table by UpDn algorithm  */
	OSM_LOG(&p_updn->p_osm->log, OSM_LOG_VERBOSE,
		"Setting all switches' Min Hop Table\n");
	status = updn_calculate_min_hop_table(p_subn);

_exit:
	OSM_LOG_EXIT(&p_updn->p_osm->log);
	return status;
}

static struct updn_node *create_updn_node(osm_switch_t * sw)
{
	struct updn_node *u;

	u = malloc(sizeof(*u));
	if (!u)
		return NULL;
	memset(u, 0, sizeof(*u));
	u->sw = sw;
	u->is_leaf = FALSE;
	u->num_hca = 0;
	u->dlid_routes = NULL;
	u->dlid_sw_routes = NULL;
	u->total_routes = 0;
	sw->rank = OSM_SW_NO_RANK;
	return u;
}

static void delete_updn_node(struct updn_node *u)
{
	u->sw->priv = NULL;
	free(u->dlid_routes);
	free(u->dlid_sw_routes);
	free(u);
}

static void updn_delete_all(updn_t *p_updn)
{
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	cl_map_item_t *item;
	osm_switch_t *p_sw;

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *) item;
		delete_updn_node(p_sw->priv);
	}

	free(p_updn->ranked_nodes);
	p_updn->ranked_nodes = NULL;

	free(p_updn->port_arr);
	free(p_updn->offset_arr);
	p_updn->port_arr = NULL;
	p_updn->offset_arr = NULL;
	p_updn->port_arr_len = 0;
	p_updn->offset_arr_len = 0;
}

/* Find Root nodes automatically by Min Hop Table info */
static void updn_find_root_nodes_by_min_hop(OUT updn_t * p_updn)
{
	osm_opensm_t *p_osm = p_updn->p_osm;
	osm_switch_t *p_sw;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	cl_map_item_t *item;
	double thd1, thd2;
	unsigned i, cas_num = 0;
	unsigned *cas_per_sw;
	uint16_t lid_ho;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);

	OSM_LOG_ENTER(&p_osm->log);

	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"Current number of ports in the subnet is %d\n",
		cl_qmap_count(&p_subn->port_guid_tbl));

	lid_ho = p_subn->max_lid + 1;
	cas_per_sw = malloc(lid_ho * sizeof(*cas_per_sw));
	if (!cas_per_sw) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR AA14: "
			"cannot alloc mem for CAs per switch counter array\n");
		goto _exit;
	}
	memset(cas_per_sw, 0, lid_ho * sizeof(*cas_per_sw));

	/* Find the Maximum number of CAs (and routers) for histogram normalization */
	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"Finding the number of CAs and storing them in cl_map\n");
	for (item = cl_qmap_head(&p_subn->port_guid_tbl);
	     item != cl_qmap_end(&p_subn->port_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_port = (osm_port_t *)item;
		if (!p_port->p_node->sw) {
			p_physp = p_port->p_physp->p_remote_physp;
			if (!p_physp || !p_physp->p_node->sw ||
			    osm_phsyp_get_special_port_type(p_port->p_physp))
				continue;
			lid_ho = osm_node_get_base_lid(p_physp->p_node, 0);
			lid_ho = cl_ntoh16(lid_ho);
			cas_per_sw[lid_ho]++;
			cas_num++;
		}
	}

	thd1 = cas_num * 0.9;
	thd2 = cas_num * 0.05;
	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"Found %u CAs and RTRs, %u SWs in the subnet. "
		"Thresholds are thd1 = %f && thd2 = %f\n",
		cas_num, cl_qmap_count(&p_subn->sw_guid_tbl), thd1, thd2);

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"Passing through all switches to collect Min Hop info\n");
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		unsigned hop_hist[IB_SUBNET_PATH_HOPS_MAX];
		uint16_t max_lid_ho;
		uint8_t hop_val;
		uint16_t numHopBarsOverThd1 = 0;
		uint16_t numHopBarsOverThd2 = 0;

		p_sw = (osm_switch_t *) item;

		memset(hop_hist, 0, sizeof(hop_hist));

		max_lid_ho = p_sw->max_lid_ho;
		for (lid_ho = 1; lid_ho <= max_lid_ho; lid_ho++)
			if (cas_per_sw[lid_ho]) {
				hop_val =
				    osm_switch_get_least_hops(p_sw, lid_ho);
				if (hop_val >= IB_SUBNET_PATH_HOPS_MAX)
					continue;

				hop_hist[hop_val] += cas_per_sw[lid_ho];
			}

		/* Now recognize the spines by requiring one bar to be
		   above 90% of the number of CAs and RTRs */
		for (i = 0; i < IB_SUBNET_PATH_HOPS_MAX; i++) {
			if (hop_hist[i] > thd1)
				numHopBarsOverThd1++;
			if (hop_hist[i] > thd2)
				numHopBarsOverThd2++;
		}

		/* If thd conditions are valid - rank the root node */
		if (numHopBarsOverThd1 == 1 && numHopBarsOverThd2 == 1) {
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
				"Ranking GUID 0x%" PRIx64 " as root node\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			p_sw->rank = 0;
			p_updn->num_roots++;
		}
	}

	free(cas_per_sw);
_exit:
	OSM_LOG_EXIT(&p_osm->log);
	return;
}

static void dump_roots(cl_map_item_t *item, FILE *file, void *cxt)
{
	osm_switch_t *sw = (osm_switch_t *)item;
	if (!sw->rank)
		fprintf(file, "0x%" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(sw->p_node)));
}

static int rank_root_node(void *cxt, uint64_t guid, char *p)
{
	updn_t *updn = cxt;
	osm_switch_t *sw;

	sw = osm_get_switch_by_guid(updn_get_subnet(updn),
				    cl_hton64(guid));
	if (!sw) {
		OSM_LOG(&updn->p_osm->log, OSM_LOG_ERROR, "ERR AA04: "
			"Root switch with guid 0x%" PRIx64 " is not found\n", guid);
		return 0;
	}

	OSM_LOG(&updn->p_osm->log, OSM_LOG_DEBUG,
		"Ranking root port GUID 0x%" PRIx64 "\n", guid);

	sw->rank = 0;
	updn->num_roots++;

	return 0;
}

/* UPDN callback function */
static int updn_lid_matrices(void *ctx)
{
	updn_t *p_updn = ctx;
	cl_map_item_t *item;
	osm_switch_t *p_sw;
	int ret = 0;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	char root_file_path[1024];

	OSM_LOG_ENTER(&p_updn->p_osm->log);

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		p_sw->priv = create_updn_node(p_sw);
		if (!p_sw->priv) {
			OSM_LOG(&(p_updn->p_osm->log), OSM_LOG_ERROR, "ERR AA0C: "
				"cannot create updn node\n");
			OSM_LOG_EXIT(&p_updn->p_osm->log);
			return -1;
		}
	}

	/* First setup root nodes */
	p_updn->num_roots = 0;

	if (p_subn->opt.root_guid_file) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_DEBUG,
			"UPDN - Fetching root nodes from file \'%s\'\n",
			p_subn->opt.root_guid_file);

		ret = parse_node_map_str(p_subn->opt.root_guid_file,
					 osm_apply_func_on_port_group_nodes, rank_root_node,
					 p_updn, p_updn->p_osm, TRUE);
		if (ret) {
			OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA02: "
				"cannot parse root guids file \'%s\' num_roots %u\n",
				p_subn->opt.root_guid_file, p_updn->num_roots);
		} else if (!p_updn->num_roots) {
			OSM_LOG(&p_updn->p_osm->log, OSM_LOG_DEBUG,
				"No roots were found from parsing root guids file \'%s\'\n",
				p_subn->opt.root_guid_file);
		}

		if (ret || !p_updn->num_roots) {
			osm_ucast_mgr_build_lid_matrices(&p_updn->p_osm->sm.ucast_mgr);

			if (p_subn->opt.find_roots_color_algorithm)
				osm_root_detection_find_root_nodes_by_coloring(
				    p_updn->p_osm,
				    updn_get_subnet(p_updn),
				    OSM_TOPOLOGY_TYPE_TREE,
				    &p_updn->num_roots, NULL);
			else
				updn_find_root_nodes_by_min_hop(p_updn);

		} else if (p_subn->opt.connect_roots &&
			   p_updn->num_roots > 1)
			osm_ucast_mgr_build_lid_matrices(&p_updn->p_osm->sm.ucast_mgr);
	} else {
		osm_ucast_mgr_build_lid_matrices(&p_updn->p_osm->sm.ucast_mgr);

		if (p_subn->opt.find_roots_color_algorithm)
			osm_root_detection_find_root_nodes_by_coloring(
			    p_updn->p_osm,
			    updn_get_subnet(p_updn),
			    OSM_TOPOLOGY_TYPE_TREE,
			    &p_updn->num_roots, NULL);
		else
			updn_find_root_nodes_by_min_hop(p_updn);
	}

	/* Only if there are assigned root nodes do the algorithm, otherwise perform do nothing */
	if (p_updn->num_roots) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_DEBUG,
			"activating UPDN algorithm\n");
		ret = updn_build_lid_matrices(p_updn);
	} else {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_INFO,
			"disabling UPDN algorithm, no root nodes were found\n");
		ret = -1;
	}

	if (OSM_LOG_IS_ACTIVE_V2(&p_updn->p_osm->log, OSM_LOG_ROUTING)) {
		snprintf(root_file_path, sizeof(root_file_path), "%s/opensm-updn-roots.dump",
			 p_updn->p_osm->subn.opt.dump_files_dir);
		osm_dump_qmap_to_file(p_updn->p_osm, root_file_path,
				      &p_subn->sw_guid_tbl,
				      dump_roots, NULL);
	}

	if (ret)
		updn_delete_all(p_updn);

	OSM_LOG_EXIT(&p_updn->p_osm->log);
	return ret;
}
/*
 * Go through sorted ranked switches for DOWN to TOP and skip leafs
 * 1. For each non-leaf, check whether switches below it already have colors.
 *    Create color for the given switch as bitwise OR of colors below it
 * 2. Assign combined color to the leafs attached to the given switch
 * 3. If no leafs are colored yet, assign new color to the given switch
 *    and apply #2
*/
static void updn_assign_color(updn_t *p_updn)
{
	int i, pn;
	color_t tmp_color = NO_COLOR;
	color_t next_color = 1;
	updn_node_t *u;
	osm_log_t *p_log = &p_updn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	for (i = p_updn->num_nodes - 1; i >= 0; i--) {
		u = p_updn->ranked_nodes[i];
		/* skip leafs */
		if (u->is_leaf && u->sw->rank)
			continue;

		tmp_color = 0;
		/* find downwards switches with assigned color */
		for (pn = 1; pn < u->sw->num_ports; pn++) {
			osm_node_t *p_remote_node;
			struct updn_node *rem_u;
			osm_switch_t *p_remote_sw;
			osm_physp_t *p_physp;
			p_remote_node =
			    osm_node_get_remote_node(u->sw->p_node, pn,
						     NULL);

			p_physp = osm_node_get_physp_ptr(u->sw->p_node, pn);

			/* If no remote node OR remote node is not a SWITCH
			   continue to next pn */
			if (!p_remote_node || !p_remote_node->sw ||
			    !p_physp || !osm_link_is_healthy(p_physp))
				continue;

			p_remote_sw = p_remote_node->sw;
			rem_u = (struct updn_node *) p_remote_sw->priv;
			rem_u = &rem_u[0];

			if (DOWN != updn_get_dir(u->sw->rank, rem_u->sw->rank,
						u->sw->id, rem_u->sw->id, 1))
				continue;
			tmp_color |= rem_u->color;
		}
		/* no colors assigned to the children */
		if (tmp_color == NO_COLOR) {
			if (next_color == MAX_LEAF_COLOR) {
				OSM_LOG(p_log, OSM_LOG_ERROR,
					"ERR AA21: More than 64 colors required."
					 "Set zero color \n");
				next_color = NO_COLOR;
			} else {
				tmp_color = next_color;
				next_color <<= 1;
			}
		}
		u->color = tmp_color;

		/* assign color to the children w/o color */
		for (pn = 1; pn < u->sw->num_ports; pn++) {
			osm_node_t *p_remote_node;
			struct updn_node *rem_u;
			osm_switch_t *p_remote_sw;
			osm_physp_t *p_physp;
			p_remote_node =
			    osm_node_get_remote_node(u->sw->p_node, pn,
						     NULL);

			p_physp = osm_node_get_physp_ptr(u->sw->p_node, pn);

			/* If no remote node OR remote node is not a SWITCH
			   continue to next pn */
			if (!p_remote_node || !p_remote_node->sw ||
			    !p_physp || !osm_link_is_healthy(p_physp))
				continue;

			p_remote_sw = p_remote_node->sw;
			rem_u = (struct updn_node *) p_remote_sw->priv;
			rem_u = &rem_u[0];

			if (DOWN != updn_get_dir(u->sw->rank, rem_u->sw->rank,
						u->sw->id, rem_u->sw->id, 1))
				continue;
			if (rem_u->color == NO_COLOR)
				rem_u->color = u->color;
		}
	}
	OSM_LOG_EXIT(p_log);
}

static int create_port_array(updn_t *updn, cl_qlist_t *port_list,
			     uint8_t min_path_bit,
			     uint8_t max_path_bit)
{
	int i = 0, j = 0, port_array_len;
	cl_list_item_t *list_item;
	endport_t *port_array = NULL;
	osm_log_t *p_log = &updn->p_osm->log;
	OSM_LOG_ENTER(p_log);

	/* create endpoint array */
	if (!port_list)
		return -1;

	port_array_len = IB_LID_UCAST_END_HO + 1;

	port_array = malloc(port_array_len * sizeof(endport_t));
	if (!port_array)
		return -1;

	updn->offset_arr = calloc(1,
			   port_array_len * sizeof(group_offset_t));
	updn->offset_arr_len = 0;

	if (!updn->offset_arr) {
		free(port_array);
		return -1;
	}

	for (list_item = cl_qlist_head(port_list);
	     list_item != cl_qlist_end(port_list);
	     list_item = cl_qlist_next(list_item)) {
		osm_port_t *p_dest_port =
			cl_item_obj(list_item, p_dest_port, list_item);
		osm_switch_t *dest_sw;
		osm_physp_t *p_dest_sw_physp, *p_dest_physp;
		uint8_t end_lid_offset;

		if (!p_dest_port->p_node->sw) {
			p_dest_sw_physp =
				osm_physp_get_remote(p_dest_port->p_physp);
			if (!p_dest_sw_physp)
				continue;
			dest_sw = p_dest_sw_physp->p_node->sw;
		}
		else
			dest_sw = p_dest_port->p_node->sw;

		p_dest_physp = osm_port_get_physp(p_dest_port);

		end_lid_offset = (1 << osm_physp_get_lmc(p_dest_physp)) - 1;
		if (min_path_bit > end_lid_offset)
			continue;

		if (max_path_bit < end_lid_offset)
			end_lid_offset = max_path_bit;

		for(j = min_path_bit; j <= end_lid_offset; j++) {
			port_array[i].path_bit = j;
			port_array[i].p_port = p_dest_port;
			port_array[i++].adj_sw = (updn_node_t *) dest_sw->priv;
		}
	}
	updn->port_arr_len = i;
	updn->port_arr = port_array;
	OSM_LOG_EXIT(p_log);
	return 0;

}

/*
 * When converge_routes is TRUE, an adjacent switch with higher value of dlid_routes
 * should be selected. Otherwise an adjacent switch with lower value of dlid_routes
 * should be chosen
*/

inline static boolean_t updn_lid_tracking_compare_dlid_routes(
					 uint32_t curr_dlid_routes,
					 uint32_t adj_sw_dlid_routes,
					 boolean_t converge_routes)
{
	return (converge_routes) ?
		(curr_dlid_routes < adj_sw_dlid_routes):
		(curr_dlid_routes > adj_sw_dlid_routes);
}

static int updn_lid_tracking_converge_routes(updn_t *p_updn, struct updn_node *u)
{
	osm_ucast_mgr_t *p_mgr = &p_updn->p_osm->sm.ucast_mgr;
	osm_switch_t *p_sw = u->sw;
	osm_port_t *p_dest_port;
	osm_physp_t *p_exit_physp;
	osm_node_t *p_adj_node, *curr_adj_node;
	osm_switch_t *p_adj_sw, *p_dest_sw;
	updn_node_t *adj_node;
	boolean_t lft_set;
	uint16_t dest_port_base_lid, dest_port_lid;
	int p, port, exit_port, i;
	uint32_t weights[IB_NODE_NUM_PORTS_MAX] = { 0 };
	uint8_t path_bit;

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Processing routes converge for switch 0x%" PRIx64 "\n",
		cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));

	for (i = 0; i < p_updn->port_arr_len; i++) {

		p_dest_port = p_updn->port_arr[i].p_port;
		path_bit = p_updn->port_arr[i].path_bit;

		dest_port_base_lid = cl_ntoh16(p_dest_port->lid);
		if (!dest_port_base_lid)
			continue;

		dest_port_lid = dest_port_base_lid + path_bit;

		if (p_dest_port->p_node->sw)
			continue;

		/* non-leaf switch - some routes pass to destination */
		if (u->dlid_routes[dest_port_lid])
			continue;

		if (p_sw->new_lft[dest_port_lid] == OSM_NO_PATH)
			continue;

		p_dest_sw = (p_updn->port_arr[i].adj_sw)->sw;

		if (!p_dest_sw)
			continue;

		/* Destination port connected directly to this switch */
		if (p_dest_sw == p_sw)
			continue;

		lft_set = FALSE;
		exit_port = OSM_NO_PATH;
		curr_adj_node = NULL;

		for (p = 1; !lft_set && p < p_sw->num_ports; p++) {
			port = osm_switch_get_dimn_port(p_sw, p);
			if (!port)
				continue;
			/* Check if port is connected and healthy */
			p_exit_physp = osm_node_get_physp_ptr(p_sw->p_node, port);
			if (!p_exit_physp || !osm_physp_is_healthy(p_exit_physp) ||
			    !p_exit_physp->p_remote_physp)
				continue;

			/* Get remote switch */
			p_adj_node = osm_physp_get_node_ptr(p_exit_physp->p_remote_physp);

			p_adj_sw = p_adj_node->sw;

			if (!p_adj_sw)
				continue;

			adj_node = (updn_node_t *)p_adj_sw->priv;

			/* Validate port direction */
			if (UP !=
			    updn_get_dir(p_sw->rank, p_adj_sw->rank,
					 u->sw->id, adj_node->sw->id, 1))
				continue;
			/* No real path through adj_node to destination */
			if (adj_node->dlid_routes[dest_port_lid] == 0)
				continue;
			/* if more than one switch are delivering DLID */
			/* stick to the first one found */
			if (curr_adj_node && curr_adj_node != p_adj_node)
				continue;
			/*
			* take care of parallel links to the
			* switch processing the node
			*/
			if (p_sw->new_lft[dest_port_lid] == port)
				lft_set = TRUE;
			else {
				if ((exit_port == OSM_NO_PATH) ||
				    weights[port] < weights[exit_port]) {
					exit_port = port;
					curr_adj_node = p_adj_node;
				}
			}
		}
		/* converge route */
		if(!lft_set && exit_port != OSM_NO_PATH) {
			p_sw->new_lft[dest_port_lid] = exit_port;
			weights[exit_port]++;
		}

	}

	return 0;
}
static int updn_lid_tracking_process_switch(updn_t *p_updn, struct updn_node *u,
					    enum updn_switch_dir direction,
					    int port_array_start)
{
	osm_ucast_mgr_t *p_mgr = &p_updn->p_osm->sm.ucast_mgr;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	endport_t *port_arr = p_updn->port_arr;
	int port_arr_len = p_updn->port_arr_len;
	osm_switch_t *p_sw = u->sw;
	osm_port_t *p_dest_port;
	osm_physp_t *p_dest_physp, *p_dest_sw_physp = NULL, *p_exit_physp;
	osm_node_t *p_adj_node;
	osm_switch_t *p_dest_sw, *p_adj_sw;
	updn_node_t *exit_node, *adj_node;
	uint32_t weights[IB_NODE_NUM_PORTS_MAX] = { 0 };
	uint32_t weights_in_color[IB_NODE_NUM_PORTS_MAX] = { 0 };
	uint32_t dlid_routes;
	uint32_t dlid_sw_routes;
	uint32_t total_routes, num_routes;
	uint16_t dest_sw_lid, dest_port_base_lid, dest_port_lid;
	uint8_t least_hops, port, p, num_ports = p_sw->num_ports;
	uint8_t exit_port;
	int i;
	color_t prev_color = NO_COLOR;
	uint8_t path_bit;
	boolean_t found, port_on_direction_exist;

	OSM_ASSERT(p_sw->rank != OSM_SW_NO_RANK);

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Processing switch 0x%" PRIx64 " direction %s offset %d\n",
		cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
		direction_str(direction),
		port_array_start);

	for (i = port_array_start; i < port_arr_len + port_array_start; i++) {

		exit_port = OSM_NO_PATH;
		exit_node = NULL;
		dlid_sw_routes = 0;
		dlid_routes = 0;
		total_routes = 0;
		if (prev_color != port_arr[i % port_arr_len].adj_sw->color) {
			prev_color = port_arr[i % port_arr_len].adj_sw->color;
			memset(weights_in_color, 0, sizeof (weights_in_color));
		}
		p_dest_port = port_arr[i % port_arr_len].p_port;
		path_bit = port_arr[i % port_arr_len].path_bit;

		p_dest_physp = osm_port_get_physp(p_dest_port);
		dest_port_base_lid = cl_ntoh16(p_dest_port->lid);
		if (!dest_port_base_lid)
			continue;

		dest_port_lid = dest_port_base_lid + path_bit;

		/* Destination is this switch */
		if (p_dest_port->p_node == p_sw->p_node) {
			if (direction == UP) {
				p_sw->new_lft[dest_port_lid] = 0;
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"Routing LID %u to port 0 for switch"
					" 0x%" PRIx64 "\n",
					dest_port_lid,
					cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			}
			continue;
		}
		p_dest_sw = port_arr[i % port_arr_len].adj_sw->sw;

		/* Destination port connected directly to this switch */
		if (p_dest_sw == p_sw) {
			if (direction == UP) {
				p_dest_sw_physp = osm_physp_get_remote(p_dest_physp);
				OSM_ASSERT(p_dest_sw_physp);
				p_sw->new_lft[dest_port_lid] =
				    osm_physp_get_port_num(p_dest_sw_physp);
			}
			continue;
		}

		if (p_dest_sw->rank == OSM_SW_NO_RANK)
			/* Unranked switch ... */
			continue;

		dest_sw_lid = cl_ntoh16(osm_node_get_base_lid(p_dest_sw->p_node, 0));
		least_hops = osm_switch_get_least_hops(p_sw, dest_sw_lid);
		if (least_hops == OSM_NO_PATH) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"No path to get to LID %u from switch"
				" 0x%" PRIx64 "\n",
				dest_port_lid,
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			continue;
		}

		port_on_direction_exist = FALSE;

		for (p = 1; p < num_ports + 1; p++) {
			port = osm_switch_get_dimn_port(p_sw, p % num_ports);
			if (!port)
				continue;

			/* Check if port leads to shortest route */
			if (osm_switch_get_hop_count(p_sw, dest_sw_lid, port) !=
						     least_hops)
				continue;

			/* Check if port is connected and healthy */
			p_exit_physp = osm_node_get_physp_ptr(p_sw->p_node, port);
			if (!p_exit_physp || !osm_physp_is_healthy(p_exit_physp) ||
			    !p_exit_physp->p_remote_physp)
				continue;

			/* Get remote switch */
			p_adj_node = osm_node_get_remote_node(p_sw->p_node,
							      port, NULL);
			OSM_ASSERT(p_adj_node && p_adj_node->sw);
			p_adj_sw = p_adj_node->sw;
			adj_node = (updn_node_t *)p_adj_sw->priv;

			/* Validate port direction.
			 * Don't skip the port in case the link
			 * is between "siblings".
			 */
			if (p_sw->rank != p_adj_sw->rank &&
			    direction !=
			    updn_get_dir(p_sw->rank, p_adj_sw->rank,
					 u->sw->id, adj_node->sw->id, 1))
				continue;
			/* found port on the same direction */
			port_on_direction_exist = TRUE;

			/*
			 * Exit port selection criteria by precedence:
			 * 1. Exit port that leads to switch with maximal/minimal
			 *    number of routes to destination port.
			 *    (depends on opt.updn_lid_tracking_converge_routes)
			 * 2. Exit port with least routes to the nodes
			 *    with the same color as the current node
			 * 3. Exit port with least routes.
			 * 4. Exit port that leads to the switch with
			 *    lesser number of routes passing through it
			 * 5. Exit port that leads to a switch with least routes
			 *    that also go through the switch connected to the
			 *    destination port.
			 */
			if (p_subn->opt.updn_lid_tracking_prefer_total_routes)
				found = (exit_port == OSM_NO_PATH ||
				updn_lid_tracking_compare_dlid_routes(
					dlid_routes,
					adj_node->dlid_routes[dest_port_lid],
					p_subn->opt.updn_lid_tracking_converge_routes) ||
				(dlid_routes == adj_node->dlid_routes[dest_port_lid] &&
				(weights_in_color[exit_port] > weights_in_color[port] ||
				(weights_in_color[exit_port] == weights_in_color[port] &&
				(weights[exit_port] > weights[port] ||
				(weights[exit_port] == weights[port] &&
				(total_routes > adj_node->total_routes ||
				(total_routes == adj_node->total_routes &&
				dlid_sw_routes > adj_node->dlid_sw_routes[dest_sw_lid]))))))));
			else
				found = (exit_port == OSM_NO_PATH ||
				updn_lid_tracking_compare_dlid_routes(
					dlid_routes,
					adj_node->dlid_routes[dest_port_lid],
					p_subn->opt.updn_lid_tracking_converge_routes) ||
				(dlid_routes == adj_node->dlid_routes[dest_port_lid] &&
				(weights_in_color[exit_port] > weights_in_color[port] ||
				(weights_in_color[exit_port] == weights_in_color[port] &&
				(weights[exit_port] > weights[port] ||
				(weights[exit_port] == weights[port] &&
				(dlid_sw_routes > adj_node->dlid_sw_routes[dest_sw_lid] ||
				(dlid_sw_routes == adj_node->dlid_sw_routes[dest_sw_lid] &&
				total_routes > adj_node->total_routes))))))));

			if (found) {
				exit_port = port;
				exit_node = adj_node;
				total_routes = adj_node->total_routes;
				dlid_sw_routes = adj_node->dlid_sw_routes[dest_sw_lid];
				dlid_routes = adj_node->dlid_routes[dest_port_lid];
			}
		}

		if (exit_port == OSM_NO_PATH) {
			/*
			 * only if we had at least one exit port to adjacent switch
			 * on the same direction as the explored direction,
			 * and no valid exit port was found report error
			 */
			if (port_on_direction_exist)
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR AA20: Failed to find %s path to get to LID %u"
					" from switch 0x%" PRIx64 "\n",
					direction_str(direction),
					dest_port_lid,
					cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			continue;
		}

		p_sw->new_lft[dest_port_lid] = exit_port;
		/* Update weights */
		num_routes = u->dlid_routes[dest_port_lid] +
			     (u->num_hca ? u->num_hca : !!u->is_leaf);

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Routing %s LID %u via port %u for switch 0x%" PRIx64 " NUM_R %d to switch %s\n",
			direction_str(direction),
			dest_port_lid, exit_port,
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
			num_routes,
			exit_node->sw->p_node->print_desc);


		if (num_routes && !p_dest_port->p_node->sw) {
			exit_node->dlid_routes[dest_port_lid] += num_routes;
			exit_node->dlid_sw_routes[dest_sw_lid] += num_routes;
			exit_node->total_routes += num_routes;
			weights[exit_port] += num_routes;
			weights_in_color[exit_port] += num_routes;
		}
	}

	return 0;
}
/*
 * Array is already sorted by nodes under the same switch
 * The switch port0 nodes are at the end of the array.
 * We need to group all nodes with same color together
 * and keep switches at the end
 */

static int updn_port_color_cmp_func(const void *lhs, const void* rhs)
{
	endport_t *p_left = (endport_t *)lhs;
	endport_t *p_right = (endport_t *)rhs;

	if (p_left->p_port->p_node->sw && p_right->p_port->p_node->sw)
		return 0;

	if (p_left->p_port->p_node->sw && !p_right->p_port->p_node->sw)
		return 1;

	if (!p_left->p_port->p_node->sw && p_right->p_port->p_node->sw)
		return -1;

	if (p_left->adj_sw->color < p_right->adj_sw->color)
		return -1;

	if (p_left->adj_sw->color > p_right->adj_sw->color)
		return 1;
	return 0;
}

static int updn_node_rank_cmp_func(const void *lhs, const void* rhs)
{
	updn_node_t *p_left = *(updn_node_t **)lhs;
	updn_node_t *p_right = *(updn_node_t **)rhs;

	return (updn_get_dir(p_left->sw->rank, p_right->sw->rank,
			     p_left->sw->id, p_right->sw->id, 1) == DOWN ? -1 : 1);
}

static int updn_node_load_and_color_cmp_up_func(const void *lhs, const void* rhs)
{
	updn_node_t *p_left = *(updn_node_t **)lhs;
	updn_node_t *p_right = *(updn_node_t **)rhs;
	unsigned lrank = p_left->sw->rank;
	unsigned rrank = p_right->sw->rank;
	color_t lcolor = p_left->color;
	color_t rcolor = p_right->color;
	uint64_t l_id = p_left->sw->id;
	uint64_t r_id = p_right->sw->id;
	uint32_t l_total = p_left->total_routes;
	uint32_t r_total = p_right->total_routes;

	if (lrank < rrank)
		return -1;

	if (lrank > rrank)
		return 1;

	if (l_total < r_total)
		return -1;

	if (l_total > r_total)
		return 1;

	/* Equal rank, decide by color */
	if (lcolor < rcolor)
		return -1;
	if (lcolor > rcolor)
		return 1;

	/* same color. decide by id */
	if (l_id < r_id)
		return -1;
	return 1;
}

/* when going down, sort according to decreasing order of total routes */
/* we want to serve more loaded switches first */
static int updn_node_load_and_color_cmp_down_func(const void *lhs, const void* rhs)
{
	updn_node_t *p_left = *(updn_node_t **)lhs;
	updn_node_t *p_right = *(updn_node_t **)rhs;
	unsigned lrank = p_left->sw->rank;
	unsigned rrank = p_right->sw->rank;
	color_t lcolor = p_left->color;
	color_t rcolor = p_right->color;
	uint64_t l_id = p_left->sw->id;
	uint64_t r_id = p_right->sw->id;
	uint32_t l_total = p_left->total_routes;
	uint32_t r_total = p_right->total_routes;

	if (lrank < rrank)
		return -1;

	if (lrank > rrank)
		return 1;

	if (l_total > r_total)
		return -1;

	if (l_total < r_total)
		return 1;

	/* Equal rank, decide by color */
	if (lcolor < rcolor)
		return -1;
	if (lcolor > rcolor)
		return 1;

	/* same color. decide by id */
	if (l_id < r_id)
		return -1;
	return 1;
}

static int updn_lid_tracking_build_lft_tables(updn_t *p_updn)
{
	osm_ucast_mgr_t *p_mgr = &p_updn->p_osm->sm.ucast_mgr;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	osm_log_t *p_log = &p_updn->p_osm->log;
	cl_map_item_t *item;
	osm_switch_t *p_sw;
	struct updn_node *u;
	int i, result = 0;
	uint16_t num_lids;
	updn_node_t * prev_adj_sw;
	endport_t *p;
	int port_array_offset, next_offset;
	int prev_rank, max_rank;
	color_t prev_color = MAX_COLOR;
	int found, k;
	int routing_guid_order_used;

	OSM_LOG_ENTER(p_log);

	/* reset new lfts from previous routing */
	osm_ucast_mgr_reset_lfts_all_switches(p_subn);

	if (osm_ucast_mgr_init_port_order_lists(p_mgr, 1)) {
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR AA08: Failed to initialize port order list\n");
		return 1;
	}


	/* Add all ranked updn nodes to an array and Initialize DLID arrays */
	p_updn->num_nodes = 0;
	p_updn->ranked_nodes =
	    (updn_node_t**)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
				  sizeof(updn_node_t *));
	if (!p_updn->ranked_nodes) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR,
			"ERR AA0D: Failed to allocate memory for ranked nodes"
			" array\n");
		result = 1;
		goto exit;
	}

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		OSM_ASSERT(p_sw && p_sw->priv);
		u = (updn_node_t *)p_sw->priv;
		if (p_sw->rank == OSM_SW_NO_RANK)
			continue;

		p_updn->ranked_nodes[p_updn->num_nodes++] = u;
		num_lids = u->sw->max_lid_ho + 1;
		u->dlid_routes = (uint32_t *)calloc(num_lids, sizeof(uint32_t));
		u->dlid_sw_routes = (uint32_t *)calloc(num_lids, sizeof(uint32_t));

		if (!u->dlid_routes || !u->dlid_sw_routes) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR AA09: Failed to allocates destination LIDs"
				" array\n");
			result = 1;
			goto exit;
		}

		u->total_routes = 0;
	}

	qsort(p_updn->ranked_nodes, p_updn->num_nodes, sizeof(updn_node_t *),
	      updn_node_rank_cmp_func);

	updn_assign_color(p_updn);

	if (create_port_array(p_updn, &p_mgr->port_order_lists[0],
			      p_subn->min_path_bit,
			      p_subn->max_path_bit)) {
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR AA10: Failed to create port array for port list\n");
		result = 1;
		goto exit;
	}

	routing_guid_order_used = p_mgr->guid_routing_order_count;
	/* sort endports by color if no guid_routing_order */
	if(!routing_guid_order_used) {
		qsort(p_updn->port_arr, p_updn->port_arr_len, sizeof(endport_t),
		      updn_port_color_cmp_func);
		/* build offset array */
		p_updn->offset_arr_len = 0;
		prev_adj_sw = NULL;

		for(i=0; i < p_updn->port_arr_len; i++) {
			p = &p_updn->port_arr[i];

			if (p->p_port->p_node->sw) {
				continue;
			}
			if (prev_adj_sw != p->adj_sw) {
				group_offset_t *g =
					&p_updn->offset_arr[p_updn->offset_arr_len];
				prev_adj_sw = p->adj_sw;
				g->offset = i;
				g->color = p->adj_sw->color;
				p_updn->offset_arr_len++;
			}
		}
	}

	next_offset = 0;
	prev_rank = -1;
	prev_color = 0;
	/* From leafs to spines */
	for (i = p_updn->num_nodes - 1; i >= 0; i--) {
		u = p_updn->ranked_nodes[i];
		/*
		* route switches with higher total routes value first
		* under the same rank
		* First sort will be by color as total routes are yet 0
		*/

		if (u->sw->rank != 0 && u->sw->rank != prev_rank) {
			prev_rank = u->sw->rank;
			qsort(&p_updn->ranked_nodes[0], i + 1,
			      sizeof(updn_node_t *),
			      updn_node_load_and_color_cmp_up_func);
			u = p_updn->ranked_nodes[i];
		}

		if (routing_guid_order_used) {
			port_array_offset = 0;
			goto Next1;
		}
		if (u->is_leaf && u->sw->rank) {
			found = 0;
			/* find first offset after the nodes with switch color */
			for (k = 0; k < p_updn->offset_arr_len; k++) {
				while (p_updn->offset_arr[k].color
					== u->color && k < p_updn->offset_arr_len) {
					found = 1;
					k++;
				}
				if (found) {
					next_offset = k % p_updn->offset_arr_len;
					break;
				}
			}
		}
		port_array_offset = p_updn->offset_arr[next_offset].offset;
Next1:
		updn_lid_tracking_process_switch(p_updn, u, UP,
						 port_array_offset);
	}

	/* Clear switch weights */
	for (i = 0; i < p_updn->num_nodes; i++) {
		u = p_updn->ranked_nodes[i];
		if (u->sw->rank != 0 ) {
			num_lids = u->sw->max_lid_ho + 1;
			memset(u->dlid_routes,0, num_lids * sizeof(uint32_t));
			memset(u->dlid_sw_routes,0, num_lids * sizeof(uint32_t));
			u->total_routes = 0;
		}
	}
	prev_rank = -1;
	max_rank = p_updn->ranked_nodes[p_updn->num_nodes - 1]->sw->rank;
	/* From spines to leafs */
	for (i = 0; i < p_updn->num_nodes; i++) {
		u = p_updn->ranked_nodes[i];
		/*
		 * when rank/color changes we need to find first offset
		 * corresponding to the host group with subcolor
		*/
		if (u->sw->rank != max_rank && u->sw->rank != prev_rank) {
			prev_rank = u->sw->rank;
			qsort(&p_updn->ranked_nodes[i], p_updn->num_nodes - i,
			      sizeof(updn_node_t *),
			      updn_node_load_and_color_cmp_down_func);
			u = p_updn->ranked_nodes[i];
		}

		if (routing_guid_order_used) {
			port_array_offset = 0;
			goto Next2;
		}
		if (next_offset == p_updn->offset_arr_len ||
		    prev_rank != u->sw->rank ||
		    prev_color != u->color) {
			prev_rank = u->sw->rank;
			prev_color = u->color;
			for (k = 0; k < p_updn->offset_arr_len; k++) {
				if (p_updn->offset_arr[k].color &
					u->color)
					break;
			}
			next_offset = k;
		}
		/* switch disconnected from topology. Possible with REC */
		if (next_offset == p_updn->offset_arr_len)
			continue;

		port_array_offset = p_updn->offset_arr[next_offset++].offset;
Next2:
		updn_lid_tracking_process_switch(p_updn, u, DOWN,
						 port_array_offset);

	}

	if (!p_subn->opt.updn_lid_tracking_converge_routes)
		goto exit;

	/* Do another pass UP to converge routes  */
	for (i = p_updn->num_nodes - 1; i >= 0; i--) {
		u = p_updn->ranked_nodes[i];
		if (u->sw->rank != 0 )
			updn_lid_tracking_converge_routes(p_updn, u);
	}

exit:
	osm_ucast_mgr_release_port_order_lists(p_mgr, 1);

	OSM_LOG_EXIT(p_log);

	return result;
}

static void updn_calculate_missing_routes(updn_t * p_updn)
{
	osm_ucast_mgr_t *p_mgr = &p_updn->p_osm->sm.ucast_mgr;
	osm_log_t *p_log = &p_updn->p_osm->log;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);

	OSM_LOG_ENTER(p_log);

	if (p_updn->num_roots > 1 &&
	    p_subn->opt.calc_missing_routes &&
	    osm_ucast_mgr_calculate_missing_routes(p_mgr)) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR,
			"ERR AA06: leaf switches were not set properly, "
			"no missing LID routes will be calculated\n");
	}

	OSM_LOG_EXIT(p_log);
}

/* UPDN callback function */
static int updn_fwd_tables(void * ctx)
{
	updn_t *p_updn = ctx;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	osm_log_t *p_log = &p_updn->p_osm->log;
	int ret = 0;

	OSM_LOG_ENTER(p_log);

	if (p_subn->opt.updn_lid_tracking_mode) {
		OSM_LOG(p_log, OSM_LOG_VERBOSE,
			"Up/Down routing with LID tracking\n");
		ret = updn_lid_tracking_build_lft_tables(p_updn);
	} else {
		ret = osm_ucast_mgr_build_lft_tables(&p_updn->p_osm->sm.ucast_mgr);
	}

	if (ret < 0)
		goto Exit;

	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Connecting switches that are unreachable within "
		"Up/Down rules\n");
	updn_calculate_missing_routes(p_updn);

Exit:
	updn_delete_all(p_updn);

	OSM_LOG_EXIT(p_log);
	return ret;
}

static int ar_updn_fwd_tables(void * ctx)
{
	updn_t *p_updn = ctx;
	osm_sm_t *sm = &p_updn->p_osm->sm;
	boolean_t all_sw_support_hbf = TRUE;
	cl_map_item_t *p_map_item;
	boolean_t prev_ignore_existig_lfts = sm->p_subn->ignore_existing_lfts;
	int result;

	/* determine whether to do rebalancing */
	if (sm->p_subn->hbf_reroute_optimization) {
		all_sw_support_hbf = TRUE;
		for (p_map_item = cl_qmap_head(&sm->p_subn->sw_guid_tbl);
		     p_map_item != cl_qmap_end(&sm->p_subn->sw_guid_tbl);
		     p_map_item = cl_qmap_next(p_map_item)) {
			osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
			osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;

			if (!osm_ar_info_is_hbf_support(&p_ar_data->req_ar_info)) {
				all_sw_support_hbf = FALSE;
				break;
			}
		}

		if (all_sw_support_hbf)
			sm->p_subn->ignore_existing_lfts = FALSE;
	}

	result = updn_fwd_tables(ctx);
	if (result)
		sm->p_subn->ignore_existing_lfts = prev_ignore_existig_lfts;

	return result;
}

static void updn_delete(void *context)
{
	free(context);
}

static ib_api_status_t updn_mcast_stree(void *context,
					IN OUT struct osm_mgrp_box *mgb)
{
	updn_t *p_updn = (updn_t *)context;
	return osm_mcast_mgr_build_clf_stree(updn_get_subnet(p_updn), mgb);
}

static int updn_add_to_sw_reachable_router(osm_subn_t *p_subn,
					   cl_qmap_t *p_sw_to_rtrs,
					   osm_switch_t *p_sw,
					   osm_router_t *p_router)
{
	cl_map_item_t *p_item;
	osm_sw_rtrs_list_t *p_rtrs_list;
	ib_net64_t sw_guid = osm_node_get_node_guid(p_sw->p_node);

	p_item = cl_qmap_get(p_sw_to_rtrs, sw_guid);

	if (p_item == cl_qmap_end(p_sw_to_rtrs)) {
		p_rtrs_list = (osm_sw_rtrs_list_t *)calloc(1, sizeof(osm_sw_rtrs_list_t));

		if (!p_rtrs_list) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR AA07: Not enough memory to add router GUID 0x%" PRIx64
				" to switch GUID 0x%" PRIx64 " list\n",
				cl_ntoh64(osm_node_get_node_guid(p_router->p_port->p_node)),
				cl_ntoh64(sw_guid));
			return -1;
		}
		if (cl_list_init(&p_rtrs_list->rtrs_list, 1)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR AA12: Failed to initialize list\n");
			free(p_rtrs_list);
			return -1;
		}
		p_rtrs_list->sw_guid = sw_guid;
		p_rtrs_list->p_sw = p_sw;
                cl_qmap_insert(p_sw_to_rtrs, p_rtrs_list->sw_guid, &p_rtrs_list->map_item);
	} else {
		p_rtrs_list = (osm_sw_rtrs_list_t *)p_item;
	}

	/* check that the router is not in the list already */
	if (osm_router_is_router_in_the_list(p_router, &p_rtrs_list->rtrs_list))
		return 0;

	if (cl_list_insert_tail(&p_rtrs_list->rtrs_list, p_router)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR AA13: Failed to add router 0x%" PRIx64 " to "
			"reachable routers list for switch 0x%" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(p_router->p_port->p_node)),
			cl_ntoh64(sw_guid));
		return -1;
	}

	return 0;
}

/*
 * This function maps all the switches to the routers that are reachable from it.
 * A router is reachable from a switch, if the path from a switch to router is an up-down path.
 * The updown and ftree routing engines build minhop table to haveonly such paths.
 * If in the minhop table of some switch A the switch that is connected to the router (switch B)
 * doesn't appear, then this router is not reachable from switch A.
 */
static void updn_mark_reachable_routers(IN osm_subn_t *p_subn, OUT cl_qmap_t *p_sw_to_rtrs)
{
	cl_map_item_t *p_sw_item;
	osm_switch_t  *p_sw;
	cl_map_item_t *p_map_item;
	osm_router_t  *p_router;
	ib_net16_t    rtr_sw_lid;
	uint8_t       hops;

	for (p_sw_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_sw_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_sw_item = cl_qmap_next(p_sw_item)) {
		p_sw = (osm_switch_t *) p_sw_item;

		if (lookup_held_back_switch(p_subn, p_sw->p_node)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Skipping held back switch 0x%" PRIx64 ".\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			continue;
		}

		/* go over all routers and add routers that switch has valid route to */
		for (p_map_item = cl_qmap_head(&p_subn->rtr_guid_tbl);
		     p_map_item != cl_qmap_end(&p_subn->rtr_guid_tbl);
		     p_map_item = cl_qmap_next(p_map_item)) {

			p_router = (osm_router_t *)p_map_item;
			/*
			 * Skip the router that doesn't support FLID or
			 * wasn't successfully configured with FLIDs ranges in router info set.
			 */
			if (!ib_general_info_is_cap_enabled(
				&p_router->p_port->p_node->general_info,
				IB_GENERAL_INFO_IS_ROUTER_LID_SUPPORTED_BIT) ||
			    !osm_router_is_router_flid_enabled(p_subn, p_router)) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
					"Skipping router 0x%" PRIx64 " that doesn't have "
					"FLID capability or wasn't properly configured "
					"with FLID ranges.\n",
					cl_ntoh64(
						osm_node_get_node_guid(p_router->p_port->p_node)));
				continue;
			}
			if (!p_router->p_port->p_remote_sw_node) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR AA24: The router port 0x%" PRIx64
					" has no remote switch node!\n",
					cl_ntoh64(osm_port_get_guid(p_router->p_port)));
				continue;
			}

			if (p_router->p_port->disc_through != OSM_DISC_NORMAL) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
					"Skipping router 0x%" PRIx64
					" that wasn't properly discovered (unhealthy).\n",
					cl_ntoh64(
						osm_node_get_node_guid(p_router->p_port->p_node)));
				continue;
			}

			/* get the lid of a switch connected to router */
			rtr_sw_lid = osm_node_get_base_lid(p_router->p_port->p_remote_sw_node, 0);
			hops = osm_switch_get_least_hops(p_sw, cl_ntoh16(rtr_sw_lid));

			/* no valid path to this router */
			if (hops == OSM_NO_PATH)
				continue;

			/* add this router to the list of reachable routers from this switch */
			updn_add_to_sw_reachable_router(p_subn, p_sw_to_rtrs, p_sw, p_router);
		}
	}
}

/*
 * This function assigns FLIDs for the switches and updates calculation
 * of the routers' router lid tables with these local FLIDs.
 */
static void updn_assign_flids_to_switches_routers(IN osm_subn_t *p_subn,
						  IN cl_qmap_t *p_sw_to_rtrs)
{
	cl_list_t *p_leaf_switches_list;
	cl_list_iterator_t leaf_switch_iter;
	osm_switch_t *p_sw;
	osm_port_t *p_sw_port0;
	osm_lid_mgr_t *p_lid_mgr = &p_subn->p_osm->sm.lid_mgr;
	uint16_t flid;
	osm_router_t  *p_router;
	cl_map_item_t *p_rtrs_item;
	ib_net64_t sw_guid;
	osm_sw_rtrs_list_t *p_rtrs_list;
	cl_list_iterator_t rtr_list_iter;
	boolean_t ran_out_of_lids = FALSE;
	unsigned int assigned_flids_number = 0;

	OSM_LOG_ENTER(&p_subn->p_osm->log);
	/* go over leaf switches and assign them with FLID */
	p_leaf_switches_list = &p_subn->leaf_switches_list;

	for (leaf_switch_iter = cl_list_head(p_leaf_switches_list);
	     leaf_switch_iter != cl_list_end(p_leaf_switches_list);
	     leaf_switch_iter = cl_list_next(leaf_switch_iter)) {
		p_sw = (osm_switch_t *)cl_list_obj(leaf_switch_iter);
		p_sw_port0 = osm_switch_get_port0(p_sw, p_subn);
		sw_guid = osm_node_get_node_guid(p_sw->p_node);

		if (!osm_switch_has_cas(p_sw)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Do not assign FLID to switch GUID 0x%" PRIx64
				" that has no CAs on it.\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			continue;
		}

		/* go over all reachable routers of this switch */
		p_rtrs_item = cl_qmap_get(p_sw_to_rtrs, sw_guid);
		/* no reachable routers from this leaf switch */
		if (p_rtrs_item == cl_qmap_end(p_sw_to_rtrs)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR AA18: "
				"The switch GUID 0x%" PRIx64 " doesn't have reachable routers. "
				"Can't assign FLID to it.\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			continue;
		}

		flid = osm_lid_mgr_get_switch_flid(p_lid_mgr, p_sw_port0);
		if (!flid) {
			/*
			 * Report first switch that the allocation of FLID failed for,
			 * in order not to flood with these errors.
			 */
			if (!ran_out_of_lids) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR AA0F: "
					"Failed to allocate FLID for switch GUID 0x%" PRIx64 ".\n",
					cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));

				ran_out_of_lids = TRUE;
			}

			/* next switch, that may get FLID allocated from guid2flid db */
			continue;
		}
		assigned_flids_number ++;

		p_rtrs_list = (osm_sw_rtrs_list_t *)p_rtrs_item;
		for (rtr_list_iter = cl_list_head(&p_rtrs_list->rtrs_list);
		     rtr_list_iter != cl_list_end(&p_rtrs_list->rtrs_list);
		     rtr_list_iter = cl_list_next(rtr_list_iter)) {
			p_router = (osm_router_t *)cl_list_obj(rtr_list_iter);
			/*
			 * assign flid after we get at least one router
			 * that is reachable from the switch.
			 */
			p_sw_port0->flid = cl_hton16(flid);

			/* update this router calculated router lid table with the switch FLID */
			osm_router_set_flid(p_subn, p_router, flid);
		}
	}

	/* update the assigned FLIDs in the DB */
	osm_db_store(p_lid_mgr->p_g2fl, p_subn->opt.fsync_high_avail_files);

	if (ran_out_of_lids)
		/* Notify UFM that we ran out of FLIDs */
		osm_report_notice_1313(p_subn, SM_VS_TRAP_RTR_LID_NUMBER_LIDS_EXCEEDED);
	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
		"Number of assigned FLIDs is %u.\n", assigned_flids_number);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

/*
 * This function assigns FLIDs to the CA ports based on previously assigned FLIDs for the switches.
 */
static void updn_assign_ca_flid(IN osm_subn_t *p_subn) {
	cl_map_item_t *item;
	osm_port_t *p_port;
	osm_port_t *p_sw_port0;

	OSM_LOG_ENTER(&p_subn->p_osm->log);
	/* go over CAs and assign them with FLID */
	for (item = cl_qmap_head(&p_subn->port_guid_tbl);
	     item != cl_qmap_end(&p_subn->port_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_port = (osm_port_t *)item;
		if (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA) {
			if (!ib_general_info_is_cap_enabled(&p_port->p_node->general_info,
				IB_GENERAL_INFO_IS_ROUTER_LID_SUPPORTED_BIT) ||
			    !p_port->p_remote_sw_node) {
				p_port->flid = 0;
				continue;
			}
			p_sw_port0 = osm_switch_get_port0(p_port->p_remote_sw_node->sw,
					p_subn);
			/*
			 * In case remote switch doesn't have an FLID assigned (flid == 0),
			 * don't assign port with FLID 0, since the flid should be 'sticky'.
			 * The switch may have flid == 0 because it lost reachability to all routers.
			 * If the port never had flid assigned it is 0 anyway, and if port->flid != 0,
			 * don't reset it.
			 * Sticky flid - don't remove previously configured flid on host
			 * in order not to break traffic on opened qp with gid (for example).
			 * TODO: it may be a CA connected to a spine, consider assigning FLID to this switch
			 */
			if (p_sw_port0->flid != 0) {
				p_port->flid = p_sw_port0->flid;
			}
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"port 0x%016" PRIx64 " has flid %u, remote switch 0x%016" PRIx64
				" has flid %u\n",
				cl_ntoh64(osm_port_get_guid(p_port)), p_port->flid,
				cl_ntoh64(osm_node_get_node_guid(p_port->p_remote_sw_node)),
				cl_ntoh16(p_sw_port0->flid));
		}
	}
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void updn_clear_sw_to_rtrs_tbl(IN OUT cl_qmap_t *p_key_to_rtrs)
{
	cl_map_item_t *p_item;
	osm_sw_rtrs_list_t *p_sw_rtrs_list;

	while (cl_qmap_count(p_key_to_rtrs)) {
		p_item = cl_qmap_head(p_key_to_rtrs);
		cl_qmap_remove_item(p_key_to_rtrs, p_item);
		p_sw_rtrs_list = (osm_sw_rtrs_list_t *)p_item;
		cl_list_remove_all(&p_sw_rtrs_list->rtrs_list);
		cl_list_destroy(&p_sw_rtrs_list->rtrs_list);
		free(p_item);
	}
}

static void updn_route_flids_by_switch(IN osm_sw_rtrs_list_t *p_sw_rtrs_list,
				       IN osm_subn_t *p_subn,
				       IN flid_routing_block_t *p_route_block)
{
	osm_router_t *p_router;
	osm_ucast_mgr_t *p_mgr = &p_subn->p_osm->sm.ucast_mgr;
	osm_ar_subgroup_t *p_subgroup, *ar_groups = p_route_block->ar_groups;
	osm_switch_t *p_sw = p_sw_rtrs_list->p_sw;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	ib_net64_t sw_guid = osm_node_get_node_guid(p_sw->p_node);
	cl_list_iterator_t list_iter;
	int foreign_flid_count = osm_router_get_foreign_flid_count(p_subn);
	uint16_t flid, rtr_sw_lid, flid_group_id, group_cap, sw_lid, group_top, rtr_lid = 0;
	uint16_t *flids_routed_ar = p_route_block->flids_routed_ar;
	uint8_t port, port_from_group, num_port_set, *flid_routed = p_route_block->flid_routed;
	uint8_t *static_ports = p_route_block->static_ports;
	uint32_t flids_routed_ar_index, flid_index, global_flids = p_subn->opt.global_flid_end + 1;

	/* look at the group top configured by routing engine */
	group_top = ib_ar_info_get_group_top_ho(&p_ar_data->req_ar_info);
	sw_lid = cl_ntoh16(osm_node_get_base_lid(p_sw->p_node, 0));

	/* If this switch has no reachable routers skip it */
	if (cl_is_list_empty(&p_sw_rtrs_list->rtrs_list))
		return;

	port = 0;
	flids_routed_ar_index = 0;

	/* Iterate through the routers that this switch can reach */
	for (list_iter = cl_list_head(&p_sw_rtrs_list->rtrs_list);
	     list_iter != cl_list_end(&p_sw_rtrs_list->rtrs_list);
	     list_iter = cl_list_next(list_iter)) {
		p_router = (osm_router_t *)cl_list_obj(list_iter);

		if (!p_router->enabled_flids)
			continue;

		rtr_lid = cl_ntoh16(osm_port_get_base_lid(p_router->p_port));
		rtr_sw_lid = cl_ntoh16(osm_node_get_base_lid(
					p_router->p_port->p_remote_sw_node, 0));
		port = p_sw->new_lft[rtr_lid];

		flid_index = 0;
		/* Iterate through the enabled FLIDs on this router */
		while (p_router->enabled_flids[flid_index] != 0 &&
		       flid_index < foreign_flid_count) {
			flid = p_router->enabled_flids[flid_index++];

			/* Be sure to prioritize any static routing that took place */
			if (flid_routed[flid] & OSM_RTR_FLID_STATIC)
				continue;

			/* update regular LFT with the valid router LID */
			p_sw->new_lft[flid] = port;

			/* Set static routing if necessary */
			if (p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED) {
				if (rtr_sw_lid == sw_lid) {
					static_ports[flid] = port;
					flid_routed[flid] = OSM_RTR_FLID_STATIC;
				}
				continue;
			}

			/*
			 * If this is the switch that is connected directly to the router
			 * Make sure to route only through this router
			 * Currently FW can't support AR Group with virtual port
			 * (of the router) set.
			 * Skip creating AR group for this FLID on the switch with router.
			 * The AR Group will be needed only for pFRN cross subnet.
			 */
			if (rtr_sw_lid == sw_lid) {
				osm_ar_lft_set(p_ar_data, flid, 0,
					       port, OSM_AR_IB_LID_STATE_STATIC, 0);
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
					"On switch 0x%" PRIx64 " (lid %u), routing to FLID %u "
					"via static port %u ONLY "
					" (through router 0x%" PRIx64
					", rtr lid %u, rtr_sw_lid %u)\n",
					cl_ntoh64(sw_guid), sw_lid, flid, port,
					cl_ntoh64(osm_node_get_node_guid(
							p_router->p_port->p_node)),
					rtr_lid, rtr_sw_lid);
				flid_routed[flid] = OSM_RTR_FLID_STATIC;
				continue;
			}

			num_port_set = osm_switch_get_minhop_ports(p_sw, rtr_sw_lid,
					&ar_groups[flid], &port_from_group);
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"On switch 0x%" PRIx64 ", routing to FLID %u "
				"via static port %u (%u total minhop ports) "
				"(through router 0x%" PRIx64 ", rtr lid %u, rtr_sw_lid %u)\n",
				cl_ntoh64(sw_guid), flid, port, num_port_set,
				cl_ntoh64(osm_node_get_node_guid(
						p_router->p_port->p_node)), rtr_lid, rtr_sw_lid);

			/* Ensure that the static port used is one from the group */
			if (!ar_is_port_in_subgroup(&ar_groups[flid], port)) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
					"On switch 0x%" PRIx64 ", route FLID %u via static port %u, "
					"which is not part of the flid group. "
					"Change it to port %u.\n",
					cl_ntoh64(sw_guid), flid, port, static_ports[flid]);
				/*
				 * Change the static port to be part of group,
				 * if the group is empty, use the static port
				 * that was calculated in unicast routing.
				 */
				static_ports[flid] = port_from_group ? port_from_group : port;
			} else {
				static_ports[flid] = port;
			}

			/*
			 * To avoid setting the ar group a lot of times we will set
			 * them all at the end. Here we keep track of the FLIDs routed.
			 */
			if (!flid_routed[flid]) {
				flids_routed_ar[flids_routed_ar_index++] = flid;
				flid_routed[flid] |= OSM_RTR_FLID_ROUTED;
			}
		} /* end of enabled FLIDs loop */
	} /* end of reachable routers loop */

	/*
	 * The AR groups for FLIDs have been calculated. Now configure the tables.
	 * FLIDs that are statically routed have already been updated in the tables.
	 */
	for (flid_index = 0; flid_index < global_flids; flid_index++) {
		flid = flids_routed_ar[flid_index];

		if (!flid)
			break;

		/*
		 * If this switch has a router connected to it
		 * or does not have AR capability static routing prevails
		 */
		if (flid_routed[flid] & OSM_RTR_FLID_STATIC)
			continue;

		flid_group_id = osm_ucast_mgr_get_groupid(p_mgr, flid, 0);
		group_cap = ib_ar_info_get_group_cap_ho(&p_ar_data->ar_info);

		if (flid_group_id == UCAST_MGR_INVALID_GROUP_ID) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Failed to find a group id for flid %u.\n",
				flid);
			continue;
		}

		/* group id exists, check capability */
		if (flid_group_id >= group_cap) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"On the switch 0x%" PRIx64 " the group id %u for FLID %u, "
				"is greater than the capability %u. Skip this switch.\n",
				cl_ntoh64(sw_guid), flid_group_id, flid, group_cap);
			/* need to configure static routing at least */
			osm_ar_lft_set(p_ar_data, flid, 0,
				       static_ports[flid], OSM_AR_IB_LID_STATE_STATIC, 0);
			continue;
		}

		/* set the FLID AR group on the switch, fill out only sub group 0 */
		p_subgroup = osm_ar_subgroup_get(p_ar_data, flid_group_id, 0);
		memcpy(p_subgroup, &ar_groups[flid], sizeof(*p_subgroup));
		/*
		 * Set this flid routing entry in plft 0 (the only one in ar updn).
		 * The static port can be one of the static ports for one of the routers.
		 */
		osm_ar_lft_set(p_ar_data, flid, 0,
			       static_ports[flid], OSM_AR_IB_LID_STATE_FREE, flid_group_id);
		group_top = flid_group_id > group_top ? flid_group_id : group_top;
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"On switch 0x%" PRIx64 " routing to FLID %u: "
			"port %u group_id %u, current group top %u ar data top %u\n",
			cl_ntoh64(sw_guid), flid, static_ports[flid], flid_group_id, group_top,
			ib_ar_info_get_group_top_ho(&p_ar_data->req_ar_info));
	}

	if (group_top > ib_ar_info_get_group_top_ho(&p_ar_data->req_ar_info)) {
		osm_ar_set_required_group_top(p_ar_data, group_top);
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"On switch 0x%" PRIx64 " set group top %u.\n",
			cl_ntoh64(sw_guid), group_top);
	}
}

static int updn_route_flids_by_switch_par(IN void ** task, uint32_t task_len,
					  IN void * context, IN unsigned thread_id)
{
	osm_sw_rtrs_list_t ** task_list = (osm_sw_rtrs_list_t **)task;
	osm_sw_rtrs_list_t *p_sw_rtrs_list;
	osm_subn_t *p_subn = context;
	flid_routing_block_t route_block;
	uint32_t i, global_flids = p_subn->opt.global_flid_end + 1;

	if (!flid_routing_block_init(&route_block, global_flids)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR AA25: "
			"Failed to allocate memory for FLID routing\n");
		return 1;
	}

	for (i = 0; i < task_len; i++) {
		p_sw_rtrs_list = task_list[i];
		flid_routing_block_clear(&route_block, global_flids);
		updn_route_flids_by_switch(p_sw_rtrs_list, p_subn, &route_block);
	}

	flid_routing_block_free(&route_block);
	return 0;
}

static void updn_map_flid_to_groupids(IN osm_subn_t *p_subn)
{
	osm_ucast_mgr_t *p_mgr = &p_subn->p_osm->sm.ucast_mgr;
	cl_map_item_t *p_map_item;
	osm_router_t * p_router;
	uint32_t flid_index;
	uint16_t flid, flid_group_id;
	int foreign_flid_count = osm_router_get_foreign_flid_count(p_subn);

	/* Iterate through all routers */
	for (p_map_item = cl_qmap_head(&p_subn->rtr_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->rtr_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_router = (osm_router_t *) p_map_item;

		if (!p_router->enabled_flids)
			continue;

		flid_index = 0;
		/* Iterate through the enabled FLIDs on this router */
		while (p_router->enabled_flids[flid_index] != 0 &&
		       flid_index < foreign_flid_count) {
			flid = p_router->enabled_flids[flid_index++];

			flid_group_id = osm_ucast_mgr_get_groupid(p_mgr, flid, 0);

			if (flid_group_id == UCAST_MGR_INVALID_GROUP_ID) {
				flid_group_id = osm_ucast_mgr_find_free_groupid(p_mgr);
				if (flid_group_id == UCAST_MGR_INVALID_GROUP_ID) {
					OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
						"Failed to find a group id for flid %u.\n",
						flid);
					continue;
				}
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
					"The group id for flid %u didn't exist. Assign %u.\n",
					flid, flid_group_id);
				/*
				 * Group id exists, store it for plft 0,
				 * as we have only plft 0 in trees.
				 */
				osm_ucast_mgr_set_groupid(p_mgr, flid, flid_group_id, 0);
			}

		}
	}
}

/*
 * This function calculates the routing to the foreign FLIDs.
 * Foreign FLID is an FLID that is set in some router's router lid enable table MAD,
 * that is not in this subnet local FLID range.
 * It is based on calculation of routing in this subnet by routing engine.
 * It uses already calculated ar lfts (or regular lfts depending on switch being ar enabled).
 * The routing toward some foreign FLID F1 on switch S,
 * uses AR Group that consists of all minhop ports (on S) towards all the routers,
 * that have this FLID F1 enabled in their router lid enable table.
 */
static void updn_route_flids(IN osm_subn_t *p_subn, IN cl_qmap_t *p_sw_to_rtrs)
{
	osm_ucast_mgr_t *p_ucast_mgr = &p_subn->p_osm->sm.ucast_mgr;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	updn_map_flid_to_groupids(p_subn);

	/* Multithreaded part of the algorithm */
	cl_mp_thread_pool_qmap_apply(&p_ucast_mgr->thread_pool, p_sw_to_rtrs,
				     CL_MP_THREAD_POOL_BATCH_PER_THREAD,
				     updn_route_flids_by_switch_par, p_subn);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

/*
 * This function goes over switches, clears previously calculated flid
 * and prepares routing tables if needed.
 */
static void updn_clear_calculated_flid_data_prepare_switches(IN OUT osm_subn_t *p_subn)
{
	cl_map_item_t *p_map_item;
	osm_router_t *p_router;
	uint32_t num_blocks;
	uint32_t size;
	osm_switch_t *p_sw;
	osm_port_t *p_sw_port0;
	osm_advanced_routing_data_t *p_ar_data;

	OSM_LOG_ENTER(&p_subn->p_osm->log);
	for (p_sw = (osm_switch_t *) cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_sw != (osm_switch_t *) cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_sw = (osm_switch_t *) cl_qmap_next(&p_sw->map_item)) {
		/* reset previously calculated FLID */
		p_sw_port0 = osm_switch_get_port0(p_sw, p_subn);
		p_sw_port0->flid = 0;
		p_ar_data = &p_sw->ar_data;

		/*
		 * Prepare routing tables.
		 * A precaution taken in case the change bit (shouldn't happen)
		 * in router info weren't set
		 * and the ucast cache is enabled and valid, then re-routing won't happen
		 * and the memory won't be extended.
		 */
		if (p_subn->max_lid > p_sw->max_lid_ho) {
			if (osm_switch_prepare_path_rebuild(p_sw, p_subn->max_lid, FALSE)) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR AA22: "
					"Failed to allocate memory for lft table.\n");
				return;
			}
			/* Extend AR LFT tables */
			if (p_sw->ar_configured == OSM_SW_AR_CONF_CONFIGURED) {
				if (osm_ar_reallocate_lft_tables(p_ar_data,
							p_subn->max_lid + 1, FALSE))
					OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
						"ERR AA23: Failed to allocate memory for "
						"adaptive routing forwarding tables.\n");
			}
		}
	}

	for (p_map_item = cl_qmap_head(&p_subn->rtr_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->rtr_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {

		p_router = (osm_router_t *)p_map_item;
		num_blocks = RTR_TBL_NUM_BLOCKS_RANGE(p_router->local_flid_start,
						      p_router->local_flid_end,
						      IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM);
		size = num_blocks * sizeof(ib_rtr_router_lid_tbl_block_t);

		if (size)
			memset(p_router->new_router_lid_tbl, 0, size);
	}
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

/* This function must be called under routing lock */
void osm_updn_calculate_flids(IN OUT osm_subn_t *p_subn)
{
	OSM_LOG_ENTER(&p_subn->p_osm->log);
	cl_qmap_t sw_to_rtrs;

	cl_qmap_init(&sw_to_rtrs);

	/* clear the previously calculated data */
	updn_clear_calculated_flid_data_prepare_switches(p_subn);

	updn_mark_reachable_routers(p_subn, &sw_to_rtrs);

	if (osm_local_flid_enabled(p_subn)) {
		updn_assign_flids_to_switches_routers(p_subn, &sw_to_rtrs);
		updn_assign_ca_flid(p_subn);
	}

	/* calculation of routing to foreign FLIDs */
	updn_route_flids(p_subn, &sw_to_rtrs);

	updn_clear_sw_to_rtrs_tbl(&sw_to_rtrs);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void updn_calculate_flids(void *ctx)
{
	updn_t *p_updn = ctx;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);

	OSM_LOG_ENTER(&p_updn->p_osm->log);	
	osm_updn_calculate_flids(p_subn);
	OSM_LOG_EXIT(&p_updn->p_osm->log);
}

int osm_ucast_updn_setup(struct osm_routing_engine *r, osm_opensm_t *osm)
{
	updn_t *updn;

	updn = malloc(sizeof(updn_t));
	if (!updn)
		return -1;
	memset(updn, 0, sizeof(updn_t));

	updn->p_osm = osm;
	updn->p_subn = osm->sm.ucast_mgr.p_subn;

	r->context = updn;
	r->destroy = updn_delete;
	r->build_lid_matrices = updn_lid_matrices;
	r->ucast_build_fwd_tables = updn_fwd_tables;
	r->mcast_build_stree = updn_mcast_stree;

	return 0;
}

int osm_ucast_ar_updn_setup(struct osm_routing_engine *r,
			    osm_opensm_t *osm)
{
	int rc;

	rc = osm_ucast_updn_setup(r, osm);

	if (rc)
		return rc;

	/* Override the updn_fwd_calculation function*/
	r->ucast_build_fwd_tables = ar_updn_fwd_tables;

	/* set suitable auto AR parameters */
	r->attributes.ar_mode_auto = OSM_ADV_ROUTING_MODE_ENABLED;
	r->attributes.shield_mode_auto = OSM_ADV_ROUTING_MODE_RN;
	r->attributes.pfrn_mode = TRUE;
	r->adv_routing_engine_auto = OSM_ADV_ROUTING_TYPE_AR_TREE;
	r->calculate_flids = updn_calculate_flids;

	return 0;
}
