/*
 * Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    osm vendor specific (class 0x0A)
 *
 * Author:
 *    Or Nechemia, Mellanox
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_debug.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_VENDOR_SPECIFIC_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_vendor_specific.h>
#include <opensm/osm_key_mgr.h>

#define VENDOR_SPECIFIC_INITIAL_TID_VALUE 0xc01d
#define OSM_VS_DEFAULT_PRINT_SUPPRESS_TIMEOUT 5000000

extern int osm_print_num_received(IN uint32_t num_received);
extern uint64_t osm_trap_get_key(IN uint16_t lid, IN uint8_t port_num,
				    IN osm_madw_t * p_ntci);
void dump_vs_mad(ib_vs_mad_t * p_vs_mad, uint32_t buf_size, osm_log_t * p_log,
		 ib_net64_t dest_guid);
static ib_api_status_t vendor_specific_send_set_keyinfo(osm_subn_t * p_subn,
							osm_port_t * p_port);

void vs_mad_post(osm_vendor_specific_t * p_vs, osm_madw_t * p_madw,
		 osm_physp_t * p_physp, osm_port_t * p_port,
		 ib_net16_t attr_id, ib_net32_t attr_mod,
		 uint8_t method_id)
{
	osm_opensm_t *p_osm = p_vs->p_osm;
	ib_vs_mad_t *p_vs_mad;
	uint8_t port;
	ib_net16_t dest_lid;
	osm_node_t *p_node = p_port->p_node;

	OSM_LOG_ENTER(&p_osm->log);

	port = osm_physp_get_port_num(p_physp);

	p_vs_mad = osm_madw_get_vs_mad_ptr(p_madw);

	p_vs_mad->header.base_ver = 1;
	p_vs_mad->header.mgmt_class = IB_MCLASS_VS;
	p_vs_mad->header.class_ver = 1;
	p_vs_mad->header.method = method_id;
	p_vs_mad->header.status = 0;
	p_vs_mad->header.class_spec = 0;
	p_vs_mad->header.trans_id =
		cl_hton64((uint64_t) cl_atomic_inc(&p_vs->trans_id) &
			  (uint64_t) (0xFFFFFFFF));
	if (p_vs_mad->header.trans_id == 0)
		p_vs_mad->header.trans_id =
			cl_hton64((uint64_t) cl_atomic_inc(&p_vs->trans_id) &
				  (uint64_t) (0xFFFFFFFF));
	p_vs_mad->header.attr_id = attr_id;
	p_vs_mad->header.additional_status = 0;
	p_vs_mad->header.attr_mod = attr_mod;

	dest_lid = osm_node_get_base_lid(p_node, port);
	if (p_osm->subn.opt.vs_key_enable)
		p_vs_mad->vend_key = p_port->vs_key;

	p_madw->mad_addr.dest_lid = dest_lid;
	p_madw->mad_addr.addr_type.gsi.remote_qp = IB_QP1;
	p_madw->mad_addr.addr_type.gsi.remote_qkey =
		IB_QP1_WELL_KNOWN_Q_KEY;
	p_madw->resp_expected = TRUE;
	p_madw->fail_msg = CL_DISP_MSGID_NONE;

	p_madw->context.vs_context.node_guid = osm_node_get_node_guid(p_node);
	p_madw->context.vs_context.port_guid = osm_physp_get_port_guid(p_physp);
	p_madw->context.vs_context.port = port;
	p_madw->context.vs_context.mad_method = method_id;
	p_madw->context.vs_context.attr_mod = attr_mod;

	p_madw->dest_guid = p_madw->context.vs_context.node_guid;

	cl_spinlock_acquire(&p_vs->mad_queue_lock);
	cl_atomic_inc(&p_vs->outstanding_mads);
	cl_qlist_insert_tail(&p_vs->mad_queue, &p_madw->list_item);
	cl_spinlock_release(&p_vs->mad_queue_lock);

	cl_event_signal(&p_vs->vs_poller_wakeup);

	OSM_LOG_EXIT(&p_osm->log);
}

static ib_api_status_t vendor_specific_send_set_cpi(osm_subn_t * p_subn,
						    osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_subn->p_osm;
	osm_vendor_specific_t *p_vs= &p_subn->p_osm->vs;
	const ib_vs_mad_t *p_vs_mad;
	ib_class_port_info_t *cpi_data;
	osm_madw_t *p_madw;
	ib_gid_t gid;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(&p_osm->log);

	p_madw = osm_mad_pool_get(&p_osm->mad_pool, p_vs->bind_handle[p_subn->sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (!p_madw) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 980F: "
			"Unable to allocate MAD\n");
		status = IB_INSUFFICIENT_MEMORY;
		goto Exit;
	}
	p_vs_mad = osm_madw_get_vs_mad_ptr(p_madw);
	cpi_data = (ib_class_port_info_t *)ib_vs_mad_get_data_ptr(p_vs_mad);
	memcpy(cpi_data, &p_port->vs_data.cpi, sizeof(ib_class_port_info_t));

	gid.unicast.prefix = 0;
	gid.unicast.interface_id = 0;
	cpi_data->trap_gid = gid;
	cpi_data->trap_tc_sl_fl = 0;
	cpi_data->trap_lid = p_subn->sm_base_lid;
	cpi_data->trap_pkey = cl_hton16(0xFFFF);
	cpi_data->trap_hop_qp = cl_hton32(1);
	cpi_data->trap_qkey = IB_QP1_WELL_KNOWN_Q_KEY;

	if (!memcmp(cpi_data, &p_port->vs_data.cpi, sizeof(ib_class_port_info_t))) {
		p_port->vs_data.resend_mask &= ~OSM_VS_RESEND_TRAP_LID;
		osm_mad_pool_put(&p_osm->mad_pool, p_madw);
		goto Exit;
	}

	cl_plock_acquire(&p_osm->lock);
	vs_mad_post(p_vs, p_madw, p_port->p_physp, p_port,
		    IB_MAD_ATTR_CLASS_PORT_INFO, 0, IB_MAD_METHOD_SET);
	cl_plock_release(&p_osm->lock);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

static boolean_t is_vskey_supported(osm_port_t * p_port)
{
	return p_port->vs_data.cpi.cap_mask & IB_VS_CAPMASK_VENDKEY_SUPPORTED;
}

static void vendor_specific_rcv_cpi(osm_vendor_specific_t * p_vs,
				    osm_madw_t * p_madw,
				    osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_vs->p_osm;
	const ib_vs_mad_t *p_vs_mad;
	ib_class_port_info_t *data;
	osm_vs_context_t *p_vs_context;

	OSM_LOG_ENTER(&p_osm->log);

	p_vs_mad = (ib_vs_mad_t *)osm_madw_get_mad_ptr(p_madw);
	data = ib_vs_mad_get_data_ptr(p_vs_mad);
	p_vs_context = (osm_vs_context_t *)&p_madw->context;

	if (p_vs_mad->header.status != 0) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9826: "
			"Received unsuccessful ClassPortInfo GetResp\n");
		goto Exit;
	}

	if (p_vs_context->mad_method == IB_MAD_METHOD_GET) {
		p_port->vs_data.resend_mask &= ~OSM_VS_RESEND_VS_KEY_SUPPORT;

		/* send classportinfo to update trap address */
		p_port->vs_data.resend_mask |= OSM_VS_RESEND_TRAP_LID;
		vendor_specific_send_set_cpi(&p_osm->subn, p_port);
		memcpy(&p_port->vs_data.cpi, (ib_class_port_info_t *)data,
		       sizeof(ib_class_port_info_t));

	} else if (p_vs_context->mad_method == IB_MAD_METHOD_SET) {
		p_port->vs_data.resend_mask &= ~OSM_VS_RESEND_TRAP_LID;
		if (!p_port->vs_data.key_configured)
			p_port->vs_data.resend_mask |= OSM_VS_RESEND_VS_KEY;
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static void vendor_specific_rcv_keyinfo(osm_vendor_specific_t * p_vs,
					osm_madw_t * p_madw,
					osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_vs->p_osm;
	osm_subn_opt_t *p_opt = &p_osm->subn.opt;
	const ib_mad_t *p_mad;
	const ib_vs_mad_t *p_vs_mad;
	osm_vs_context_t *p_vs_context;
	ib_vend_key_info_t *p_mad_info_data;

	OSM_LOG_ENTER(&p_osm->log);
	OSM_ASSERT(p_madw);
	p_mad = osm_madw_get_mad_ptr(p_madw);
	p_vs_context = (osm_vs_context_t *)&p_madw->context;

	if (p_mad->status != 0) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9805: "
			"Received keyinfo GetResp with status (0x%x)\n",
			cl_ntoh16(p_mad->status));
		goto Exit;
	}

	if (p_vs_context->mad_method == IB_MAD_METHOD_GET) {
		p_vs_mad = osm_madw_get_vs_mad_ptr(p_madw);
		p_mad_info_data = ib_vs_mad_get_data_ptr(p_vs_mad);

		/*
		 * If response holds a key - it means key is no longer protected
		 * and this is the port's key.
		 * If response's key is 0, there are 2 options -
		 * 	1. protect_bit is 1, means key is protected and SM holds the right key.
		 * 	2. otherwise, key is 0.
		 */

		if (p_mad_info_data->protect_bit) {
			/* Key parameters were not changed */
			if (p_mad_info_data->lease_period == cl_ntoh16(p_opt->vs_key_lease_period) &&
			    p_mad_info_data->protect_bit ==
			    cl_hton16((uint16_t)p_opt->vs_key_ci_protect_bits << 15)) {

				/*
				 * Send CPI Get in order to check if need to update
				 * values of trap_lid etc.
				 */
				p_port->vs_data.resend_mask |= OSM_VS_RESEND_VS_KEY_SUPPORT;
			}

		} else {
			/* Updating the port with current key */
			p_port->vs_key = p_mad_info_data->vend_key;
		}

		p_port->vs_data.resend_mask &= ~OSM_VS_RESEND_GET_KEY_INFO;
	} else if (p_vs_context->mad_method == IB_MAD_METHOD_SET &&
	    (p_port->vs_data.resend_mask & OSM_VS_RESEND_VS_KEY)) {
		p_port->vs_data.resend_mask &= ~OSM_VS_RESEND_VS_KEY;

		/* Done key configuration */
		p_port->vs_data.key_configured = TRUE;
		p_vs->num_configured++;
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

void osm_vendor_specific_decrement_outstanding_mads(osm_vendor_specific_t * p_vs)
{
	uint32_t outstanding;

	outstanding = cl_atomic_dec(&p_vs->outstanding_mads);
	if (!outstanding)
		cl_event_signal(&p_vs->outstanding_mads_done_event);

	cl_atomic_dec(&p_vs->outstanding_mads_on_wire);
	cl_event_signal(&p_vs->sig_mads_on_wire_continue);
}

static void vendor_specific_rcv_statistic_counter_increase(osm_opensm_t * p_osm,
							   ib_net16_t attr_id)
{
	/* Vendor specific class is currently sent only on the main port */
	uint64_t *stats_counters = p_osm->stats_counters[0];

	switch (attr_id) {
	case IB_MAD_ATTR_CLASS_PORT_INFO:
		stats_counters[VS_CLASS_PORT_INFO_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_VEND_KEY_INFO:
		stats_counters[VS_KEY_INFO_RCV_COUNT]++;
		break;
	default:
		break;
	}
}

static void vendor_specific_send_statistic_counter_increase(osm_opensm_t * p_osm,
							    ib_net16_t attr_id)
{
	/* Vendor specific class is currently sent only on the main port */
	uint64_t *stats_counters = p_osm->stats_counters[0];

	switch (attr_id) {
	case IB_MAD_ATTR_CLASS_PORT_INFO:
		stats_counters[VS_CLASS_PORT_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_VEND_KEY_INFO:
		stats_counters[VS_KEY_INFO_SEND_COUNT]++;
		break;
	default:
		break;
	}
}

static void vs_rcv_mad(void * context, void * data)
{

	osm_vendor_specific_t *p_vs = context;
	osm_opensm_t *p_osm = p_vs->p_osm;
	osm_madw_t *p_madw = data;
	ib_vs_mad_t *p_vs_mad;
	osm_madw_context_t *p_mad_context = &p_madw->context;
	ib_mad_t *p_mad = osm_madw_get_mad_ptr(p_madw);
	ib_net64_t node_guid = p_mad_context->vs_context.node_guid;
	ib_net64_t port_guid = p_mad_context->vs_context.port_guid;
	uint8_t port = p_mad_context->vs_context.port;
	osm_port_t *p_port;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"Processing received MAD status 0x%x for "
		"attr ID 0x%x mod 0x%x node 0x%" PRIx64 " port %u\n",
		cl_ntoh16(p_mad->status), cl_ntoh16(p_mad->attr_id),
		cl_ntoh32(p_mad_context->vs_context.attr_mod),
		cl_ntoh64(node_guid), port);

	p_vs_mad = osm_madw_get_vs_mad_ptr(p_madw);
	vendor_specific_rcv_statistic_counter_increase(p_osm,
						       p_vs_mad->header.attr_id);

	cl_plock_acquire(&p_osm->lock);

	p_port = osm_get_port_by_guid(&p_osm->subn, port_guid);
	if (!p_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9807: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(port_guid));
		cl_plock_release(&p_osm->lock);
		goto Exit;
	}

	p_port->vs_timeout_count = 0;

	if (p_vs_mad->header.status) {
		if (p_vs_mad->header.status & IB_MAD_STATUS_UNSUP_CLASS_VER
		    || p_vs_mad->header.status & IB_MAD_STATUS_UNSUP_METHOD
		    || p_vs_mad->header.status & IB_MAD_STATUS_UNSUP_METHOD_ATTR)
			p_port->vs_unavailable_flag = TRUE;
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9828:"
			" Received invalid status from port GUID 0x%" PRIx64 ","
			" status:0x%x, attr_id:0x%x\n",
			cl_ntoh64(port_guid),
			cl_ntoh16(p_vs_mad->header.status),
			cl_ntoh16(p_vs_mad->header.attr_id));
		cl_plock_release(&p_osm->lock);
		goto Exit;
	} else {
		p_port->vs_unavailable_flag = FALSE;
	}

	if (p_vs_mad->header.attr_id == IB_MAD_ATTR_CLASS_PORT_INFO) {
		vendor_specific_rcv_cpi(p_vs, p_madw, p_port);
	} else if (p_vs_mad->header.attr_id == IB_MAD_ATTR_VEND_KEY_INFO) {
		vendor_specific_rcv_keyinfo(p_vs, p_madw, p_port);
	} else {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9808: "
			"Unexpected MAD attribute ID %u received\n",
			cl_ntoh16(p_vs_mad->header.attr_id));
	}

	cl_plock_release(&p_osm->lock);
Exit:
	osm_mad_pool_put(&p_osm->mad_pool, p_madw);
	osm_vendor_specific_decrement_outstanding_mads(p_vs);
	OSM_LOG_EXIT(&p_osm->log);
}

static ib_api_status_t vendor_specific_send_repress(IN osm_vendor_specific_t * p_vs,
						    IN osm_madw_t * p_req_madw)
{
	osm_madw_t *resp_madw;
	ib_api_status_t status = IB_SUCCESS;
	osm_opensm_t *p_osm = p_vs->p_osm;
	ib_mad_t *p_resp_mad;
	ib_vs_mad_t *p_vs_mad;
	ib_net16_t slid;
	osm_port_t *p_port;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_ASSERT(p_req_madw);

	/* do nothing if we are exiting ... */
	if (osm_exit_flag)
		goto Exit;

	slid = p_req_madw->mad_addr.dest_lid;
	p_port = osm_get_port_by_lid(&p_osm->subn, slid);
	if (!p_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 982A: "
			"Could not find port of source lid %u\n",
			cl_ntoh16(slid));
		status = IB_INVALID_LID;
		goto Exit;
	}

	resp_madw = osm_mad_pool_get(&p_osm->mad_pool,
				     p_vs->bind_handle[p_vs->p_osm->subn.sm_port_idx],
				     MAD_BLOCK_SIZE, NULL);
	if (resp_madw == NULL) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 9812: Unable to acquire MAD\n");
		status = IB_INSUFFICIENT_RESOURCES;
		goto Exit;
	}

	/*
	   Copy the request mad to the response mad, then just
	   update the necessary fields.
	 */
	memcpy((void *)resp_madw->p_mad, (void *)p_req_madw->p_mad, sizeof(ib_mad_t));
	p_resp_mad = (ib_mad_t *)resp_madw->p_mad;

	resp_madw->mad_addr.dest_lid = slid;
	resp_madw->mad_addr.addr_type.gsi.remote_qp = IB_QP1;
	resp_madw->mad_addr.addr_type.gsi.remote_qkey =
		IB_QP1_WELL_KNOWN_Q_KEY;

	p_resp_mad->method = IB_MAD_METHOD_TRAP_REPRESS;
	resp_madw->status = 0;
	resp_madw->resp_expected = FALSE;
	resp_madw->fail_msg = CL_DISP_MSGID_NONE;

	p_vs_mad = osm_madw_get_vs_mad_ptr(resp_madw);
	p_vs_mad->vend_key = p_port->vs_key;

	/* post trap repress to vs poller */
	cl_spinlock_acquire(&p_vs->mad_queue_lock);
	cl_qlist_insert_tail(&p_vs->mad_queue, &resp_madw->list_item);
	cl_spinlock_release(&p_vs->mad_queue_lock);

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE, "Responding with trap repress "
		"to trap from LID %u\n", cl_ntoh16(slid));

	cl_event_signal(&p_vs->vs_poller_wakeup);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

static int vendor_specific_create_vs_notice_1600(osm_sm_t * sm,
						 ib_mad_notice_attr_t * p_notice,
						 ib_mad_notice_attr_t * p_ntci_1600)
{
	ib_net16_t slid;
	osm_port_t *p_port;

	slid = p_notice->data_details.vendor_data_details.v_data_details.ntc_vskey.slid;
	p_port = osm_get_port_by_lid(sm->p_subn, slid);
	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9816: "
			"Could not find source physical port\n");
		return 1;
	}

	/* Copy attributes, not include data_details and issuer gid*/
	memcpy((void *)p_ntci_1600,(void *)p_notice , 10);

	p_ntci_1600->generic_type &= 0x7F;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.slid = slid;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.attribute_id =
		p_notice->data_details.vendor_data_details.v_data_details.ntc_vskey.attribute_id;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.method =
		p_notice->data_details.vendor_data_details.v_data_details.ntc_vskey.method;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.mclass = IB_MCLASS_VS;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.sgid =
		p_notice->data_details.vendor_data_details.v_data_details.ntc_vskey.source_gid;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.physp_sguid =
		p_port->guid;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.key =
		p_notice->data_details.vendor_data_details.v_data_details.ntc_vskey.key;
	p_ntci_1600->data_details.vendor_data_details.trap_num = cl_ntoh16(1600);
	p_ntci_1600->issuer_gid = p_notice->issuer_gid;

	return 0;
}

static void vendor_specific_trap_process_req(void *context, void *data)
{
	osm_vendor_specific_t *p_vs = (osm_vendor_specific_t *)context;
	osm_opensm_t *p_osm = p_vs->p_osm;
	osm_sm_t *sm = &p_osm->sm;
	osm_madw_t *p_madw = data;
	uint8_t payload[sizeof(ib_mad_notice_attr_t)];
	ib_mad_notice_attr_t *p_ntci = (ib_mad_notice_attr_t *)payload;
	ib_mad_notice_attr_t p_ntci_1600;
	ib_vs_mad_t *p_vs_mad;
	osm_madw_t tmp_madw; /* we need a copy to last after repress */
	ib_api_status_t status;
	uint64_t trap_key, tid;
	uint32_t num_received;
	uint8_t method;
	ib_net16_t source_lid, trap_num;
	osm_port_t * source_port;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_ASSERT(p_madw);
	OSM_ASSERT(p_madw->p_mad->mgmt_class == IB_MCLASS_VS);

	memset(payload, 0, sizeof(payload));
	memset(&tmp_madw, 0, sizeof(tmp_madw));

	p_vs_mad = (ib_vs_mad_t *)p_madw->p_mad;
	method = p_madw->p_mad->method;
	if (method != IB_MAD_METHOD_TRAP) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 981D: "
			"Unsupported method 0x%X\n", method);
		goto Exit;
	}

	memcpy(payload, p_vs_mad->data, sizeof(payload));
	p_ntci = (ib_mad_notice_attr_t *)payload;
	trap_num = p_ntci->g_or_v.generic.trap_num;
	if (trap_num != VS_TRAP_NUM_KEY_VIOLATION) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 981E: "
			"Unsupported trap number 0x%X\n",
			cl_ntoh16(trap_num));
		goto Exit;
	}

	memcpy(&tmp_madw, p_madw, sizeof(tmp_madw));
	osm_dump_key_violation(&p_osm->log, p_ntci, FILE_ID, OSM_LOG_VERBOSE,
			       IB_MCLASS_VS);

	source_lid = p_madw->mad_addr.dest_lid;
	source_port = osm_get_port_by_lid(sm->p_subn, source_lid);
	if (!source_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9827: "
			"Trap source port could not be found by lid %u\n",
			cl_ntoh16(source_lid));
		goto Exit;
	}

	/*
	 * If SM does not hold the correct key, sending trap repress and
	 * Set KeyInfo will just generate more key violations.
	 */
	if (!source_port->vs_data.key_configured) {
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE, "Ignores trap,"
			" key is not configured yet for port 0x%"PRIx64"\n",
			cl_ntoh64(source_port->guid));
		goto Exit;
	}

	status = vendor_specific_send_repress(p_vs, p_madw);
	if (status != IB_SUCCESS) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 981A: "
			"Failed sending repress (%s) to port lid %u\n",
			ib_get_err_str(status), cl_ntoh16(source_lid));
		goto Exit;
	}

	status = vendor_specific_send_set_keyinfo(sm->p_subn, source_port);
	if (status != IB_SUCCESS) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 982B:"
			" Failed sending SET KeyInfo (%s) to port lid %u,"
			" due to trap receive\n",
			ib_get_err_str(status), cl_ntoh16(source_lid));
		goto Exit;
	}

	/* try to find it in the aging tracker */
	trap_key = osm_trap_get_key(source_lid, source_port->p_physp->port_num,
				    p_madw);
	num_received = cl_event_wheel_num_regs(&sm->trap_aging_tracker,
						trap_key);
	if (num_received >= 10 &&
	    osm_print_num_received(num_received)) {
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
			"Received trap %u times consecutively\n",
			num_received);
	}

	/* restart the aging anyway */
	cl_event_wheel_reg(&sm->trap_aging_tracker, trap_key,
			   cl_get_time_stamp() + OSM_VS_DEFAULT_PRINT_SUPPRESS_TIMEOUT,
			   NULL, NULL);

	/* If was already registered do nothing more */
	if (num_received >= 10)
		goto Exit;

	tid = cl_ntoh64(p_madw->p_mad->trans_id);

	if(vendor_specific_create_vs_notice_1600(sm, p_ntci, &p_ntci_1600))
		goto Exit;

	cl_plock_acquire(&p_osm->lock);
	status = osm_report_notice(sm->p_log, sm->p_subn, &p_ntci_1600,
				   (uint64_t)cl_atomic_inc(&sm->p_subn->tid) &
				   (uint64_t)(0xFFFFFFFF),
				   tid);
	cl_plock_release(&p_osm->lock);
	if (status != IB_SUCCESS) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9821: "
			"Error sending trap reports (%s)\n",
			ib_get_err_str(status));
		goto Exit;
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static void vendor_specific_poller_send(osm_vendor_specific_t * p_vs,
					osm_madw_t * p_madw)
{
	osm_opensm_t *p_osm = p_vs->p_osm;
	osm_subn_opt_t *p_opt = &p_osm->subn.opt;
	cl_status_t sts;
	ib_api_status_t status;
	osm_madw_context_t mad_context = p_madw->context;
	ib_net64_t guid;
	uint32_t port_num = 0;
	uint8_t mgmt_class = p_madw->p_mad->mgmt_class;;
	uint16_t attr_id = p_madw->p_mad->attr_id;
	boolean_t resp_expected;

	if (mgmt_class != IB_MCLASS_VS) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9813: "
			"Vendor Specific manager got MAD of management class (%u). "
			"Failed to send mad\n", mgmt_class);
		return;
	}
	resp_expected = p_madw->resp_expected;
	status = osm_vendor_send(p_vs->bind_handle[p_osm->subn.sm_port_idx], p_madw, p_madw->resp_expected);
	if (status == IB_SUCCESS) {
		vendor_specific_send_statistic_counter_increase(p_osm, attr_id);
		if (resp_expected)
			cl_atomic_inc(&p_vs->outstanding_mads_on_wire);
		while (p_vs->outstanding_mads_on_wire >
		       (int32_t)p_opt->vs_max_outstanding_mads) {
wait:
			sts = cl_event_wait_on(&p_vs->sig_mads_on_wire_continue,
					       EVENT_NO_TIMEOUT, TRUE);
			if (sts != CL_SUCCESS)
				goto wait;
		}
	} else {
		guid = mad_context.vs_context.node_guid;
		port_num = mad_context.vs_context.port;

		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9814: "
			"send failed to node 0x%" PRIx64 " port %u\n",
			cl_ntoh64(guid), port_num);
	}
}

static char * vs_get_vs_attr_str(IN ib_net16_t attr)
{
	switch (attr) {
	case IB_MAD_ATTR_CLASS_PORT_INFO:
		return "VSClassPortInfo";
	case IB_MAD_ATTR_VEND_KEY_INFO:
		return "VSKeyInfo";
	case IB_MAD_ATTR_NOTICE:
		return "Notice";
	default:
		return "?";
	}
}

void dump_vs_mad(ib_vs_mad_t * p_vs_mad, uint32_t buf_size, osm_log_t * p_log,
		 ib_net64_t dest_guid)
{
	char buf[BUF_SIZE];
	buf[0] = '\0';
	unsigned n;
	int i;

	OSM_ASSERT(p_vs_mad)

	n = sprintf(buf,
		    "VS Mad dump:\n"
		    "\t\t\t\tbase_ver................0x%X\n"
		    "\t\t\t\tmgmt_class..............0x%X\n"
		    "\t\t\t\tclass_ver...............0x%X\n"
		    "\t\t\t\tmethod..................0x%X (%s)\n",
		    p_vs_mad->header.base_ver, p_vs_mad->header.mgmt_class,
		    p_vs_mad->header.class_ver, p_vs_mad->header.method,
		    ib_get_sm_method_str(p_vs_mad->header.method));

	n += snprintf(buf + n, buf_size - n,
		     "\t\t\t\tstatus..................0x%X\n",
		     cl_ntoh16(p_vs_mad->header.status));

	n += snprintf(buf + n, buf_size - n,
		      "\t\t\t\ttrans_id................0x%" PRIx64 "\n"
		      "\t\t\t\tattr_id.................0x%X (%s)\n"
		      "\t\t\t\tadditional_status.......0x%X\n"
		      "\t\t\t\tattr_mod................0x%X\n"
		      "\t\t\t\tvend_key................0x%016" PRIx64 "\n"
		      "\t\t\t\tdest_guid...............0x%016" PRIx64 "\n",
		      cl_ntoh64(p_vs_mad->header.trans_id),
		      cl_ntoh16(p_vs_mad->header.attr_id),
		      vs_get_vs_attr_str(p_vs_mad->header.attr_id),
		      cl_ntoh16(p_vs_mad->header.additional_status),
		      cl_ntoh32(p_vs_mad->header.attr_mod),
		      cl_ntoh64(p_vs_mad->vend_key),
		      cl_ntoh64(dest_guid));

	for (i = 0; i < 64 ; i += 16) {
		n += snprintf(buf + n, buf_size - n,
			      "\n\t\t\t\t%02X %02X %02X %02X "
			      "%02X %02X %02X %02X"
			      "   %02X %02X %02X %02X %02X %02X %02X %02X\n",
			      p_vs_mad->data[i],
			      p_vs_mad->data[i + 1],
			      p_vs_mad->data[i + 2],
			      p_vs_mad->data[i + 3],
			      p_vs_mad->data[i + 4],
			      p_vs_mad->data[i + 5],
			      p_vs_mad->data[i + 6],
			      p_vs_mad->data[i + 7],
			      p_vs_mad->data[i + 8],
			      p_vs_mad->data[i + 9],
			      p_vs_mad->data[i + 10],
			      p_vs_mad->data[i + 11],
			      p_vs_mad->data[i + 12],
			      p_vs_mad->data[i + 13],
			      p_vs_mad->data[i + 14],
			      p_vs_mad->data[i + 15]);
	}

	osm_log(p_log, OSM_LOG_FRAMES, "%s\n", buf);

}

static void vs_poller(void *p_ptr)
{
	osm_vendor_specific_t * p_vs = p_ptr;
	osm_opensm_t *p_osm = p_vs->p_osm;
	osm_madw_t *p_madw;
	ib_mad_t *p_mad;

	OSM_LOG_ENTER(&p_osm->log);

	if (p_vs->thread_state == OSM_THREAD_STATE_NONE)
		p_vs->thread_state = OSM_THREAD_STATE_RUN;

	while (p_vs->thread_state == OSM_THREAD_STATE_RUN) {
		cl_spinlock_acquire(&p_vs->mad_queue_lock);
		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_vs->mad_queue);
		cl_spinlock_release(&p_vs->mad_queue_lock);

		if (p_madw != (osm_madw_t *) cl_qlist_end(&p_vs->mad_queue)) {
			p_mad = (ib_mad_t *)osm_madw_get_vs_mad_ptr(p_madw);

			if (OSM_LOG_IS_ACTIVE_V2(&p_osm->log, OSM_LOG_FRAMES)) {
				dump_vs_mad((ib_vs_mad_t *)p_mad, BUF_SIZE,
					    &p_osm->log, p_madw->dest_guid);
			}
			vendor_specific_poller_send(p_vs, p_madw);
		} else {
			cl_event_wait_on(&p_vs->vs_poller_wakeup,
					 EVENT_NO_TIMEOUT, TRUE);
		}
	}

	OSM_LOG_EXIT(&p_osm->log);
}

ib_api_status_t osm_vendor_specific_init(osm_vendor_specific_t * p_vs,
					 struct osm_opensm * p_osm,
					 const osm_subn_opt_t * p_opt)
{
	ib_api_status_t status;

	OSM_LOG_ENTER(&p_osm->log);

	memset(p_vs, 0, sizeof(*p_vs));

	p_vs->p_osm = p_osm;
	p_vs->trans_id = VENDOR_SPECIFIC_INITIAL_TID_VALUE;
	memcpy(p_vs->vendor, p_osm->p_gmp_vendor, sizeof(p_vs->vendor));

	p_vs->vs_disp_h = cl_disp_register(&p_osm->gmp_disp,
					   OSM_MSG_MAD_VS,
					   vs_rcv_mad, p_vs);
	if (p_vs->vs_disp_h == CL_DISP_INVALID_HANDLE) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9801:"
			"could not register dispatcher handler\n");
		status = IB_INSUFFICIENT_RESOURCES;
		goto Exit;
	}

	p_vs->vs_trap_disp_h = cl_disp_register(&p_osm->gmp_trap_disp,
						OSM_MSG_MAD_VS_NOTICE,
						vendor_specific_trap_process_req,
						p_vs);
	if (p_vs->vs_trap_disp_h == CL_DISP_INVALID_HANDLE){
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9802:"
			"could not register dispatcher handler\n");
		return IB_INSUFFICIENT_RESOURCES;
	}

	cl_qlist_init(&p_vs->mad_queue);

	status = cl_spinlock_init(&p_vs->mad_queue_lock);
	if (status != IB_SUCCESS)
		goto Exit;

	cl_event_construct(&p_vs->vs_poller_wakeup);
	status = cl_event_init(&p_vs->vs_poller_wakeup, FALSE);
	if (status != IB_SUCCESS)
		goto Exit;

	cl_event_construct(&p_vs->outstanding_mads_done_event);
	status = cl_event_init(&p_vs->outstanding_mads_done_event, FALSE);
	if (status != IB_SUCCESS)
		goto Exit;

	cl_event_construct(&p_vs->sig_mads_on_wire_continue);
	status = cl_event_init(&p_vs->sig_mads_on_wire_continue, FALSE);
	if (status != IB_SUCCESS)
		goto Exit;

	p_vs->thread_state = OSM_THREAD_STATE_NONE;

	status = cl_thread_init(&p_vs->vs_poller, vs_poller, p_vs,
				"vs mgr poller");
	if (status != IB_SUCCESS)
		goto Exit;

	p_vs->recalculate_keys = TRUE;
	p_vs->num_configured = 0;

	status = IB_SUCCESS;
Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

void osm_vendor_specific_destroy(osm_vendor_specific_t * p_vs)
{
	osm_opensm_t *p_osm = p_vs->p_osm;
	osm_madw_t *p_madw;

	OSM_LOG_ENTER(&p_osm->log);

	p_vs->thread_state = OSM_THREAD_STATE_EXIT;

	cl_event_signal(&p_vs->sig_mads_on_wire_continue);
	cl_event_signal(&p_vs->vs_poller_wakeup);

	cl_thread_destroy(&p_vs->vs_poller);

	cl_spinlock_acquire(&p_vs->mad_queue_lock);
	while (!cl_is_qlist_empty(&p_vs->mad_queue)) {
		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_vs->mad_queue);
		osm_mad_pool_put(&p_osm->mad_pool, p_madw);
	}
	cl_spinlock_release(&p_vs->mad_queue_lock);

	cl_spinlock_destroy(&p_vs->mad_queue_lock);

	cl_event_destroy(&p_vs->vs_poller_wakeup);
	cl_event_destroy(&p_vs->outstanding_mads_done_event);
	cl_event_destroy(&p_vs->sig_mads_on_wire_continue);

	OSM_LOG_EXIT(&p_osm->log);
}

static void vs_mad_recv_callback(osm_madw_t * p_madw, void * bind_context,
				 osm_madw_t * p_req_madw)
{
	osm_vendor_specific_t *p_vs = bind_context;
	osm_opensm_t *p_osm = p_vs->p_osm;
	ib_net64_t dest_guid;

	OSM_LOG_ENTER(&p_osm->log);
	OSM_ASSERT(p_madw);

	dest_guid = (p_madw->p_mad->method == IB_MAD_METHOD_GET_RESP) ?
		    p_req_madw->dest_guid : p_madw->dest_guid;

	if (OSM_LOG_IS_ACTIVE_V2(&p_osm->log, OSM_LOG_FRAMES))
		dump_vs_mad((ib_vs_mad_t *)p_madw->p_mad, BUF_SIZE, &p_osm->log, dest_guid);

	switch (p_madw->p_mad->method) {
	case IB_MAD_METHOD_GET_RESP:
		OSM_ASSERT(p_req_madw != NULL);
		osm_madw_copy_context(p_madw, p_req_madw);
		osm_mad_pool_put(&p_osm->mad_pool, p_req_madw);

		/* posts to disp queue so disp handles through vs_rcv_mad*/
		if (cl_disp_post(p_vs->vs_disp_h, OSM_MSG_MAD_VS,
				 p_madw, NULL, NULL) != CL_SUCCESS) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9809: "
				"VS MAD post to management class dispatcher "
				"has failed\n");
			osm_mad_pool_put(&p_osm->mad_pool, p_madw);
		}
		break;
	case IB_MAD_METHOD_TRAP:
		OSM_ASSERT(p_req_madw == NULL);
		if (cl_disp_post(p_vs->vs_trap_disp_h, OSM_MSG_MAD_VS_NOTICE,
				 p_madw, NULL, NULL) != CL_SUCCESS) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9824: "
				"VS MAD post to management class trap "
				"dispatcher has failed\n");
			osm_mad_pool_put(&p_osm->mad_pool, p_madw);
		}
		break;
	default:
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9825: "
			"Method not supported by vendor specific management class\n");
		osm_mad_pool_put(&p_osm->mad_pool, p_madw);
		if (p_req_madw)
			osm_mad_pool_put(&p_osm->mad_pool, p_req_madw);
	}
	OSM_LOG_EXIT(&p_osm->log);
}

static void vs_mad_send_err_callback(void * bind_context,
				     osm_madw_t * p_madw)
{
	osm_vendor_specific_t *p_vs = bind_context;
	osm_madw_context_t *p_madw_context = &p_madw->context;
	osm_opensm_t *p_osm = p_vs->p_osm;
	uint64_t node_guid = p_madw_context->vs_context.node_guid;
	uint64_t port_guid = p_madw_context->vs_context.port_guid;
	uint8_t port = p_madw_context->vs_context.port;
	osm_port_t *p_port;
	int log_flag = 1;

	OSM_LOG_ENTER(&p_osm->log);

	cl_plock_acquire(&p_osm->lock);

	if (OSM_LOG_IS_ACTIVE_V2(&p_osm->log, OSM_LOG_FRAMES))
		dump_vs_mad((ib_vs_mad_t *)p_madw->p_mad, BUF_SIZE,
			    &p_osm->log, p_madw->dest_guid);

	p_port = osm_get_port_by_guid(&p_osm->subn, port_guid);
	if (!p_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 980A: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(port_guid));
		cl_plock_release(&p_osm->lock);
		goto Exit;
	}

	/* If timed out before, don't bothering logging again
	 * we assume no VS support
	 */
	if (p_madw->status == IB_TIMEOUT && p_port->vs_timeout_count)
		log_flag = 0;

	if (log_flag)
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 980B: MAD Error (%s): "
			"attr id = 0x%x LID %u GUID 0x%016" PRIx64 " port %u "
			"TID 0x%" PRIx64 "\n",
			ib_get_err_str(p_madw->status),
			cl_ntoh16(p_madw->p_mad->attr_id),
			cl_ntoh16(p_madw->mad_addr.dest_lid),
			cl_ntoh64(node_guid),
			port,
			cl_ntoh64(p_madw->p_mad->trans_id));

	if (p_madw->status == IB_TIMEOUT) {
		/* Vendor specific class is currently sent only on the main port */
		p_osm->stats_counters[0][VS_MAD_TIMEOUTS_COUNT]++;
		p_port->vs_timeout_count++;
		if (p_port->vs_timeout_count > OSM_VS_TIMEOUT_COUNT_THRESHOLD
		    && !p_port->vs_unavailable_flag) {
			p_port->vs_unavailable_flag = TRUE;
			p_port->vs_timeout_count = 0;
		}
	} else {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 980C: Subnet initialization error\n");
		p_osm->subn.subnet_initialization_error = TRUE;
	}

	cl_plock_release(&p_osm->lock);

	osm_vendor_specific_decrement_outstanding_mads(p_vs);
Exit:
	osm_mad_pool_put(&p_osm->mad_pool, p_madw);
	OSM_LOG_EXIT(&p_osm->log);
}

ib_api_status_t osm_vendor_specific_bind(osm_vendor_specific_t * p_vs,
					 ib_net64_t port_guid)
{
	osm_opensm_t *p_osm = p_vs->p_osm;
	osm_bind_info_t bind_info;
	ib_api_status_t status = IB_SUCCESS;
	uint8_t port_index = p_osm->subn.sm_port_idx;

	OSM_LOG_ENTER(&p_osm->log);

	bind_info.port_guid = p_vs->port_guid = port_guid;
	bind_info.mad_class = IB_MCLASS_VS;
	bind_info.class_version = 1;
	bind_info.is_responder = FALSE;
	bind_info.is_report_processor = FALSE;
	bind_info.is_trap_processor = TRUE;
	bind_info.recv_q_size = OSM_SM_DEFAULT_QP1_RCV_SIZE;
	bind_info.send_q_size = OSM_SM_DEFAULT_QP1_SEND_SIZE;
	bind_info.timeout = p_osm->subn.opt.transaction_timeout;
	bind_info.retries = p_osm->subn.opt.transaction_retries;

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"Binding to port GUID 0x%" PRIx64 ", port_index: %u\n",
		cl_ntoh64(port_guid), port_index);

	if (p_vs->bind_handle[port_index] == OSM_BIND_INVALID_HANDLE)
		p_vs->bind_handle[port_index] =
			osm_vendor_bind(p_vs->vendor[port_index], &bind_info,
					&p_osm->mad_pool,
					vs_mad_recv_callback,
					vs_mad_send_err_callback, p_vs);
	else
		p_vs->bind_handle[port_index] =
			osm_vendor_rebind(p_vs->vendor[port_index], &bind_info,
					  &p_osm->mad_pool,
					  vs_mad_recv_callback,
					  vs_mad_send_err_callback,
					  p_vs, p_vs->bind_handle[port_index]);

	if (p_vs->bind_handle[port_index] == OSM_BIND_INVALID_HANDLE) {
		status = IB_ERROR;
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 980D: Vendor specific bind failed (%s)\n",
			ib_get_err_str(status));
		goto Exit;
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

void osm_vendor_specific_shutdown(osm_vendor_specific_t * p_vs)
{
	osm_opensm_t *p_osm = p_vs->p_osm;
	uint8_t port_index;

	OSM_LOG_ENTER(&p_osm->log);
	for (port_index = 0; port_index < OSM_MAX_BINDING_PORTS; port_index++) {
		if (p_vs->bind_handle[port_index] != CL_DISP_INVALID_HANDLE)
			osm_vendor_unbind(p_vs->bind_handle[port_index]);
	}
	cl_disp_unregister(p_vs->vs_disp_h);
	cl_event_signal(&p_vs->outstanding_mads_done_event);

	OSM_LOG_EXIT(&p_osm->log);
}

int osm_vendor_specific_wait_pending_transactions(struct osm_opensm * p_osm)
{
	osm_vendor_specific_t *p_vs = &p_osm->vs;

	while (1) {
		unsigned count = p_vs->outstanding_mads;
		if (!count || osm_exit_flag)
			break;
		cl_event_wait_on(&p_vs->outstanding_mads_done_event,
				 EVENT_NO_TIMEOUT,
				 TRUE);
	}

	return osm_exit_flag;
}

static ib_api_status_t vendor_specific_send_req_cpi(osm_sm_t * p_sm,
						    osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_sm->p_subn->p_osm;
	osm_madw_t *p_madw = NULL;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp;

	OSM_LOG_ENTER(&p_osm->log);
	if (p_port->disc_through != OSM_DISC_NORMAL)
		return IB_SUCCESS;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
		p_physp = osm_node_get_physp_ptr(p_node, 0);
	else
		p_physp = p_port->p_physp;

	p_madw = osm_mad_pool_get(&p_osm->mad_pool,
				  p_osm->vs.bind_handle[p_osm->subn.sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);

	if (p_madw == NULL) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9806: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	cl_plock_acquire(&p_osm->lock);
	vs_mad_post(&p_osm->vs, p_madw, p_physp, p_port,
		    IB_MAD_ATTR_CLASS_PORT_INFO, 0, IB_MAD_METHOD_GET);

	cl_plock_release(&p_osm->lock);

	OSM_LOG_EXIT(&p_osm->log);

	return IB_SUCCESS;
}

static void vendor_specific_send_cpi(osm_sm_t * p_sm, osm_port_t * p_port)
{
	ib_api_status_t status;
	boolean_t send_req_cpi = TRUE;

	if ((!p_port->vs_data.key_configured &&
	    (p_port->vs_data.resend_mask & OSM_VS_RESEND_VS_KEY_SUPPORT)) ||
	    p_sm->p_subn->coming_out_of_standby ||
	    p_sm->p_subn->first_time_master_sweep)
		goto req_cpi;

	if (p_port->is_new || p_port->p_physp->need_update ||
	    (p_port->p_node->sw && p_port->p_node->sw->need_update))
		goto req_cpi;

	/* continue key configuration */
	if (p_port->vs_data.resend_mask & OSM_VS_RESEND_VS_KEY_SUPPORT) {
		goto req_cpi;
	} else if (p_port->vs_data.resend_mask & OSM_VS_RESEND_TRAP_LID) {
		status = vendor_specific_send_set_cpi(p_sm->p_subn, p_port);
		if (status != IB_SUCCESS) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 9811: "
				"failed to update port (guid 0x%016" PRIx64 ") "
				"with VS classportinfo\n",
				cl_ntoh64(p_port->guid));
		}
	}

	send_req_cpi = FALSE;
req_cpi:
	if (send_req_cpi &&
	    !(p_port->vs_data.resend_mask & OSM_VS_RESEND_GET_KEY_INFO)) {
		p_port->vs_data.resend_mask |= OSM_VS_RESEND_VS_KEY_SUPPORT;
		vendor_specific_send_req_cpi(p_sm, p_port);
	}

}

static ib_api_status_t vendor_specific_send_set_keyinfo(osm_subn_t * p_subn,
							osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_subn->p_osm;
	osm_vendor_specific_t *p_vs = &p_subn->p_osm->vs;
	osm_madw_t *p_madw = NULL;
	ib_vs_mad_t *p_mad = NULL;
	ib_vend_key_info_t key_info_data, *p_mad_info_data = NULL;
	osm_subn_opt_t *p_opt = &p_subn->opt;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(&p_osm->log);

 	// bind_handler is different for mclass
	p_madw = osm_mad_pool_get(&p_osm->mad_pool, p_vs->bind_handle[p_osm->subn.sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);
	if (!p_madw) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9817: "
			"failed to allocate mad\n");
		status = IB_INSUFFICIENT_MEMORY;
		goto Exit;
	}

	p_mad = osm_madw_get_vs_mad_ptr(p_madw);
	p_mad_info_data = ib_vs_mad_get_data_ptr(p_mad);

	memset(&key_info_data, 0, sizeof(ib_vend_key_info_t));

	key_info_data.vend_key = osm_key_mgr_generate_key(p_osm, p_port, IB_MCLASS_VS,
							  OSM_KEY_TYPE_MANAGER);

	key_info_data.protect_bit = cl_hton16((uint16_t)p_opt->vs_key_ci_protect_bits << 15);
	key_info_data.lease_period = cl_hton16(p_opt->vs_key_lease_period);
	key_info_data.violations = 0;

	memcpy(p_mad_info_data, &key_info_data, sizeof(ib_vend_key_info_t));
	cl_plock_acquire(&p_osm->lock);
	vs_mad_post(p_vs, p_madw, p_port->p_physp, p_port,
		    IB_MAD_ATTR_VEND_KEY_INFO, 0, IB_MAD_METHOD_SET);
	cl_plock_release(&p_osm->lock);

	/* Updating the port structure after sending the keyInfo mad
	 * with previous key */
	p_port->vs_key = key_info_data.vend_key;
	osm_key_mgr_set_cached_key(p_osm, p_port, IB_MCLASS_VS, OSM_KEY_TYPE_MANAGER);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

static void vendor_specific_configure_keyinfo(osm_opensm_t * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;
	ib_api_status_t status;

        if (p_osm->subn.opt.vs_key_enable == OSM_CAPABILITY_IGNORED)
                return;

	/* Send KeyInfo to all ports which vskey is yet to be configured
 	 * and support the feature */
	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;

		if (!p_port->is_reachable || p_port->vs_unavailable_flag ||
		    (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA &&
		     osm_physp_get_port_state(p_port->p_physp) != IB_LINK_ACTIVE))
			continue;

		if (p_port->vs_data.resend_mask &
		    (OSM_VS_RESEND_VS_KEY_SUPPORT | OSM_VS_RESEND_TRAP_LID | OSM_VS_RESEND_GET_KEY_INFO))
			continue;

		if (p_port->vs_data.key_configured &&
		    !(p_osm->vs.recalculate_keys || p_osm->vs.vs_key_param_update)) {
			p_osm->vs.num_configured++;
			continue;
		}

		if (!(is_vskey_supported(p_port) &&
		    (p_port->p_physp->port_info.capability_mask &
		    IB_PORT_CAP_HAS_VEND_CLS)))
			continue;

		if ((p_port->vs_data.resend_mask & OSM_VS_RESEND_VS_KEY) ||
		    p_osm->vs.vs_key_param_update) {
			p_port->vs_data.resend_mask |= OSM_VS_RESEND_VS_KEY;
			status = vendor_specific_send_set_keyinfo(&p_osm->subn, p_port);
			if (status != IB_SUCCESS) {
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9822: "
				"failed to update port guid 0x%016" PRIx64
				" vskeyinfo\n",
				cl_ntoh64(p_port->guid));
			}
		}
	}
}

static void vendor_specific_configure_cpi(osm_opensm_t * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;

		if (!p_port->is_reachable ||
		    (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA &&
		     osm_physp_get_port_state(p_port->p_physp) != IB_LINK_ACTIVE))
			continue;

		if ((p_port->p_physp->port_info.capability_mask & IB_PORT_CAP_HAS_VEND_CLS) &&
		    !p_port->vs_unavailable_flag &&
		    (p_port->vs_data.resend_mask &
		     (OSM_VS_RESEND_VS_KEY_SUPPORT | OSM_VS_RESEND_TRAP_LID)))
			vendor_specific_send_cpi(&p_osm->sm, p_port);
	}
}

static ib_api_status_t vendor_specific_send_req_keyinfo(osm_sm_t * p_sm,
							osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_sm->p_subn->p_osm;
	osm_madw_t *p_madw = NULL;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (p_port->disc_through != OSM_DISC_NORMAL)
		return IB_SUCCESS;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
		p_physp = osm_node_get_physp_ptr(p_node, 0);
	else
		p_physp = p_port->p_physp;

	p_madw = osm_mad_pool_get(&p_osm->mad_pool,
				  p_osm->vs.bind_handle[p_osm->subn.sm_port_idx],
				  MAD_BLOCK_SIZE, NULL);

	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 9829: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	cl_plock_acquire(&p_osm->lock);
	vs_mad_post(&p_osm->vs, p_madw, p_physp, p_port,
		    IB_MAD_ATTR_VEND_KEY_INFO, 0, IB_MAD_METHOD_GET);

	cl_plock_release(&p_osm->lock);

	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static void vendor_specific_req_keys(struct osm_opensm * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;

	/* Send KeyInfo Get to all relevant ports in order to get current VS keys */
	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;

		if (!p_port->is_reachable ||
		    (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_CA &&
		     osm_physp_get_port_state(p_port->p_physp) != IB_LINK_ACTIVE) ||
		    !(p_port->vs_data.resend_mask & OSM_VS_RESEND_GET_KEY_INFO))
			continue;

		if (p_port->vs_data.key_configured &&
		    !(p_port->vs_data.resend_mask & OSM_VS_RESEND_GET_KEY_INFO))
			continue;

		if (!(p_port->p_physp->port_info.capability_mask &
		    IB_PORT_CAP_HAS_VEND_CLS))
			continue;

		if (vendor_specific_send_req_keyinfo(&p_osm->sm, p_port)) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9803: "
				"failed to update port guid 0x%016" PRIx64
				" vskeyinfo\n",
				cl_ntoh64(p_port->guid));
		}
	}
}

int osm_vendor_specific_configure_subnet_keys(struct osm_opensm * p_osm)
{
	int ret = 0;

	OSM_LOG_ENTER(&p_osm->log);

	p_osm->vs.num_configured = 0;

	vendor_specific_req_keys(p_osm);
	if (osm_vendor_specific_wait_pending_transactions(p_osm)) {
		ret = 1;
		goto Exit;
	}

	vendor_specific_configure_cpi(p_osm);
	if (osm_vendor_specific_wait_pending_transactions(p_osm)) {
		ret = 1;
		goto Exit;
	}

	vendor_specific_configure_keyinfo(p_osm);
	if (osm_vendor_specific_wait_pending_transactions(p_osm)) {
		ret = 1;
		goto Exit;
	}

	p_osm->vs.vs_key_param_update = FALSE;
	p_osm->vs.recalculate_keys = FALSE;

	if (p_osm->subn.opt.vs_key_enable == OSM_CAPABILITY_ENABLED)
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
			"%u out of %d ports were successfully configured for VS key\n",
			p_osm->vs.num_configured, cl_qmap_count(&p_osm->subn.port_guid_tbl));

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return ret;
}

static void vendor_specific_setup_key_configuration(struct osm_opensm * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;

	OSM_LOG_ENTER(&p_osm->log);

	if (p_osm->subn.opt.vs_key_enable == OSM_CAPABILITY_IGNORED)
		goto Exit;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;

		/*
		 * Set cached key at port key. Hence, key configuration MADs
		 * are send with cached key, if there is any.
		 */
		if (!(p_port->vs_data.resend_mask & OSM_VS_RESEND_GET_KEY_INFO))
			osm_key_mgr_get_cached_key(p_osm, p_port, IB_MCLASS_VS,
						   OSM_KEY_TYPE_MANAGER);

		osm_key_mgr_generate_key(p_osm, p_port, IB_MCLASS_VS,
					 OSM_KEY_TYPE_MANAGER);

		if (p_port->is_new || p_port->p_physp->need_update ||
		    (p_port->p_node->sw && p_port->p_node->sw->need_update) ||
		    p_osm->subn.first_time_master_sweep ||
		    p_osm->subn.force_reroute || p_osm->subn.coming_out_of_standby) {
			p_port->vs_data.resend_mask |= OSM_VS_RESEND_GET_KEY_INFO;
			p_port->vs_data.resend_mask |= OSM_VS_RESEND_VS_KEY_SUPPORT;
		}

		if (p_osm->vs.recalculate_keys)
			p_port->vs_data.resend_mask |= OSM_VS_RESEND_VS_KEY;
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

int osm_vendor_specific_setup(struct osm_opensm *p_osm) {
	int ret = 0;
	OSM_LOG_ENTER(&p_osm->log);

	if (p_osm->subn.opt.vs_key_enable) {
		vendor_specific_setup_key_configuration(p_osm);

		if (osm_vendor_specific_configure_subnet_keys(p_osm)) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9823: "
				"VS keys configuration has failed\n");
			ret = 1;
		}
	}

	OSM_LOG_EXIT(&p_osm->log);
	return ret;
}
