/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *  Copyright (C) 2010-2020, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */

/* osm_config.h needs to be before all other includes */
#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ARMGR_CPP

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <opensm/osm_ucast_mgr.h>
#include <complib/cl_qmap.h>
#include <opensm/osm_subnet.h>

#include <algorithm>

#include <iostream>
#include <sstream>

#include "opensm/osm_armgr.hpp"
#include "opensm/osm_max_flow_algorithm.h"
#include "opensm/osm_ar_algorithm.hpp"
#include "opensm/osm_ar_kdor_algorithm.hpp"
#include "opensm/osm_ar_dfp_algorithm.hpp"

#define GROUP_BLOCKS_NUMBER(calculated_groups_number , ar_info)  \
        (((calculated_groups_number*(ar_info.sub_grps_active + 1)) + \
          (AR_GROUP_TABLE_BLOCK_SIZE-1)) / AR_GROUP_TABLE_BLOCK_SIZE)

#define IS_GENERATE_ARN(master_db, sw_db_entry) \
        (master_db.m_arn_enable && \
	 (sw_db_entry.m_ar_info.enable_features & IB_AR_INFO_FEATURE_MASK_ARN_SUPPORTED) && \
         ib_ar_info_get_grp_tbl_cpy_sup(&sw_db_entry.m_ar_info))

#define IS_GENERATE_FRN(master_db, sw_db_entry) \
        (master_db.m_frn_enable && \
	 (sw_db_entry.m_ar_info.enable_features & IB_AR_INFO_FEATURE_MASK_FRN_SUPPORTED) && \
         ib_ar_info_get_grp_tbl_cpy_sup(&sw_db_entry.m_ar_info))


//Bit 0 = 2 sub groups (1,3,5,7)
//Bit 1 = 3 sub groups (2,3,6,7)
//Bit 2 = 4 sub groups (4,5,6,7)
//Bits 3 to 7 are reserved
                               //0,1,2,3,4,5,6,7
//u_int8_t DFSubGrpsRequired[7] = {1,2,3,1,3,1,2,1};

extern "C" {
	int wait_for_pending_transactions(osm_opensm_t * p_osm);
	cl_status_t osm_root_detection_init_dfp_potential_roots(osm_subn_t * p_subn,
								cl_u64_vector_t * p_dfp_potential_roots);
	void osm_minhop_ar_process_subnet(osm_subn_t *, void *, int, int);
	int osm_minhop_ar_process_switch(osm_subn_t *, osm_switch_t *);
}


using namespace OSM;


/****************************************************/
/**************** private methods *******************/
/****************************************************/
//============================================================================//
void OSMAdaptiveRoutingManager::WaitForPendingTransactions()
{

	osm_subn_t *p_subn = m_p_osm_subn;
	osm_sm_t *sm = &p_subn->p_osm->sm;

	/*
	 * If enable_queries_during_routing is disabled, this section is called
	 * under the SM lock while routing. Thus, need to release the lock in
	 * order to handle MADs
 	 */
	if (!p_subn->opt.enable_queries_during_routing)
		CL_PLOCK_RELEASE(sm->p_lock);

	CL_PLOCK_RELEASE(&p_subn->p_osm->routing_lock);

	(void) wait_for_pending_transactions(p_subn->p_osm);

	CL_PLOCK_EXCL_ACQUIRE(&p_subn->p_osm->routing_lock);

	if (!p_subn->opt.enable_queries_during_routing)
		CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
}

void OSMAdaptiveRoutingManager::CalcNumCas(ARSWDataBaseEntry &db_entry)
{
	int num_cas = 0;
	osm_switch_t *p_sw = db_entry.m_general_sw_info.m_p_osm_sw;
	osm_node_t *p_node = p_sw->p_node;

	for (uint8_t port_num = 1 ; port_num < p_sw->num_ports ; ++port_num) {
		osm_physp_t *p_physp = osm_node_get_physp_ptr(p_node, port_num);
		osm_node_t *p_remote_node = osm_node_get_remote_node(p_node, port_num, NULL);

		if (!p_remote_node || !p_physp || !osm_link_is_healthy(p_physp))
			continue;

		if ((p_remote_node->node_info.node_type == IB_NODE_TYPE_SWITCH) ||
		    (p_remote_node->node_info.node_type == IB_NODE_TYPE_ROUTER))
			continue;

		/* Don't count Aggregation Nodes as CAs */
		if (osm_phsyp_get_special_port_type(p_physp->p_remote_physp) == IB_SPECIAL_PORT_TYPE_AN)
			continue;

		num_cas++;
	}

	db_entry.m_num_cas = num_cas;
}

void OSMAdaptiveRoutingManager::CalcIsLeaf(ARSWDataBaseEntry &db_entry)
{
	bool is_leaf = true;
	osm_switch_t *p_sw = db_entry.m_general_sw_info.m_p_osm_sw;
	osm_node_t *p_node = p_sw->p_node;

	for (uint8_t port_num = 1 ; port_num < p_sw->num_ports ; ++port_num) {
		osm_physp_t *p_physp = osm_node_get_physp_ptr(p_node, port_num);
		osm_node_t *p_remote_node = osm_node_get_remote_node(p_node, port_num, NULL);

		if (!p_remote_node || !p_physp || !osm_link_is_healthy(p_physp))
			continue;

		if (p_remote_node->sw && (p_remote_node->sw->rank > p_sw->rank)) {
			is_leaf = false;
			break;
		}
	}

	db_entry.m_is_leaf = is_leaf;
}

void OSMAdaptiveRoutingManager::SetRequiredARInfo(ARSWDataBaseEntry &db_entry)
{
	CLEAR_STRUCT(db_entry.m_required_ar_info);

	if ((m_master_db.m_ar_algorithm == AR_ALGORITHM_DF_PLUS) ||
	    (m_master_db.m_ar_algorithm == AR_ALGORITHM_DFP2)) {
		u_int16_t dfp_en_vl_mask = m_master_db.m_dfp_en_vl_mask;
		if (m_master_db.m_en_sl_mask == OSM_ALL_SL_VL_AR_ENABLED) {
			//all sl/vl are AR enabled
			dfp_en_vl_mask = OSM_ALL_SL_VL_AR_ENABLED;
		}

		/*
		 * AR_ALGORITHM_DFP2 uses 3 pLFTs on Spine instead of disabling
		 * AR on odd VLs
		 */
		if ((db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_SPINE) &&
		    (m_master_db.m_ar_algorithm == AR_ALGORITHM_DF_PLUS)){
			db_entry.m_required_ar_info.enabled_and_cap_fields |=
					IB_AR_INFO_CAP_MASK_BY_SL_EN;
			db_entry.m_required_ar_info.enable_by_sl_mask =
				cl_ntoh16((u_int16_t)(dfp_en_vl_mask & AR_ENABLE_BY_VL_MASK));
		} else if (dfp_en_vl_mask != OSM_ALL_SL_VL_AR_ENABLED) {
			db_entry.m_required_ar_info.enabled_and_cap_fields |=
					IB_AR_INFO_CAP_MASK_BY_SL_EN;
			db_entry.m_required_ar_info.enable_by_sl_mask =
				cl_ntoh16(dfp_en_vl_mask);
		}
		//(when using DF, one can not disable AR on a specific switch )
		db_entry.m_required_ar_info.enable_features |= IB_AR_INFO_FEATURE_MASK_AR_ENABLED;
		//currently DF supports only 2 sub groups
		ib_ar_info_set_sub_grps_active(&db_entry.m_required_ar_info, 1);
	} else {
		if (m_master_db.m_en_sl_mask != OSM_ALL_SL_VL_AR_ENABLED) {
			if (!(db_entry.m_ar_info.enabled_and_cap_fields &
						IB_AR_INFO_CAP_MASK_BY_SL_CAP)) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8505: Switch GUID 0x%" PRIx64 ", LID %u, " \
					   "do not support by_sl_cap.\n",
					   db_entry.m_general_sw_info.m_guid,
					   db_entry.m_general_sw_info.m_lid);
				db_entry.m_option_on = false;
			} else {
				db_entry.m_required_ar_info.enabled_and_cap_fields |=
					IB_AR_INFO_CAP_MASK_BY_SL_EN;
				db_entry.m_required_ar_info.enable_by_sl_mask =
					cl_ntoh16(m_master_db.m_en_sl_mask);
			}
		}

		if (m_master_db.m_ar_algorithm == AR_ALGORITHM_TREE &&
		    (m_p_osm_subn->opt.ar_tree_asymmetric_flow ==
		     OSM_ASYMMETRIC_TREE_2_SUB_GROUPS_ENABLED)) {
			CalcIsLeaf(db_entry);
			CalcNumCas(db_entry);

			if (db_entry.m_is_leaf && db_entry.m_num_cas)
				ib_ar_info_set_sub_grps_active(&db_entry.m_required_ar_info, 1);
			else
				ib_ar_info_set_sub_grps_active(&db_entry.m_required_ar_info, 0);
		} else {
			ib_ar_info_set_sub_grps_active(&db_entry.m_required_ar_info, 0);
		}
	}

	if (db_entry.m_option_on)
		db_entry.m_required_ar_info.enable_features |= IB_AR_INFO_FEATURE_MASK_AR_ENABLED;
	else
		db_entry.m_required_ar_info.enable_features &= ~IB_AR_INFO_FEATURE_MASK_AR_ENABLED;

	if (m_master_db.m_en_tr_mask != AR_CFG_TRANSPORT_MASK_EN_ALL) {
		if (!(db_entry.m_ar_info.enabled_and_cap_fields &
					IB_AR_INFO_CAP_MASK_BY_TRANS_CAP)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8506: Switch GUID 0x%" PRIx64 ", LID %u, " \
				   "do not support by_transport_disable.\n",
				   db_entry.m_general_sw_info.m_guid,
				   db_entry.m_general_sw_info.m_lid);
			db_entry.m_option_on = false;
		} else {
			db_entry.m_required_ar_info.enabled_and_cap_fields |=
				IB_AR_INFO_CAP_MASK_BY_TRANS_CAP;
			/* the bits 4-7 of 'by transport disable' field are reserved */
			db_entry.m_required_ar_info.by_trns_disable =
				(~m_master_db.m_en_tr_mask) & 0xF;
		}
	}

	db_entry.m_required_ar_info.ageing_time_value =
		cl_hton32(db_entry.m_ageing_time_value);

	//RN
	if (IS_GENERATE_ARN(m_master_db, db_entry) ||
	    IS_GENERATE_FRN(m_master_db, db_entry))
		db_entry.m_required_ar_info.enable_features |=
			IB_AR_INFO_FEATURE_MASK_RN_XMIT_ENABLED;
	else if (m_master_db.m_arn_enable || m_master_db.m_frn_enable) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8507: Switch GUID 0x%" PRIx64 ", LID %u, " \
			   "do not support routing notification.\n",
			   db_entry.m_general_sw_info.m_guid,
			   db_entry.m_general_sw_info.m_lid);
	}

	if ((db_entry.m_ar_info.enable_features &
		IB_AR_INFO_FEATURE_MASK_FR_SUPPORTED) &&
	    m_master_db.m_fr_enable) {
		db_entry.m_required_ar_info.enable_features |=
			IB_AR_INFO_FEATURE_MASK_FR_ENABLED;
	}

	/*
	 * Don't support HBF for dfp. Can configure HBF data for DFP
	 * when the hbf sl mask is 0
	 */
	if (osm_ar_info_is_hbf_support(&db_entry.m_ar_info) &&
	    (!m_master_db.m_hbf_sl_mask || m_master_db.m_ar_algorithm != AR_ALGORITHM_DF_PLUS)) {
	 	/* Set HBF enable, when HBF SL mask is not 0. */
		if (m_master_db.m_hbf_sl_mask)
			db_entry.m_required_ar_info.enable_features1 |= IB_AR_INFO_HBF_EN_BY_SL_HBF_EN;
		else
			db_entry.m_required_ar_info.enable_features1 &= ~IB_AR_INFO_HBF_EN_BY_SL_HBF_EN;

		db_entry.m_required_ar_info.enable_by_sl_mask_hbf =
			cl_ntoh16(m_master_db.m_hbf_sl_mask);
	}
}

//============================================================================//
int OSMAdaptiveRoutingManager::Init()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	if (m_is_init)
		return 0;

	/*
	 * init m_ar_algorithm
	*/

	switch (m_master_db.m_ar_algorithm) {
		case AR_ALGORITHM_DFP2:
		{
			m_ar_algorithm = new ArDfpAlgorithm(
				m_p_osm_log, m_sw_db.m_sw_map, *this);

			if (m_ar_algorithm == NULL) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "Failed to create DFP algorithm\n");
				throw 1;
			}

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
				   "Set AR DFP algorithm\n");
			break;
		}
		case AR_ALGORITHM_KDOR_HC:
		{

			m_ar_algorithm = new ArHcAlgorithm(
				m_p_osm_log, m_sw_db.m_sw_map, *this);

			if (m_ar_algorithm == NULL) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "Failed to create HC algorithm\n");
				throw 1;
			}

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
				   "Set AR kDOR HC algorithm\n");
			break;
		}
		case AR_ALGORITHM_KDOR_GHC:
		{

			m_ar_algorithm = new ArGhcAlgorithm(
				m_p_osm_log, m_sw_db.m_sw_map, *this);

			if (m_ar_algorithm == NULL) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "Failed to create GHC algorithm\n");
				throw 1;
			}

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
				   "Set AR kDOR GHC algorithm\n");
			break;
		}
		default:
		{
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Not using AR algorithm object\n");
		}
	}

	/* If configuring DFP / DFP2 with root coloring algorithm, init roots vector */
	if ((m_master_db.m_ar_algorithm == AR_ALGORITHM_DFP2 ||
	     m_master_db.m_ar_algorithm == AR_ALGORITHM_DF_PLUS) &&
	    m_p_osm_subn->opt.dfp_find_roots_color_algorithm &&
	    cl_u64_vector_init(&m_dfp_potential_roots, 0, TRUE)) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8511: Failed to create roots array for root"
			   " detection algorithm\n");
		throw 1;
	}

	if (m_sw_db.Init()) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8510: ARSWDataBase init failed.\n");
		throw 1;
	}

	if (m_thread_pool.Init(m_p_osm_subn->opt.routing_threads_num))
		throw 1;

	m_port_groups_calculator.Init();

	m_is_init = true;
	OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

//============================================================================//

/***************/
void OSMAdaptiveRoutingManager::UpdateSW(const ARGeneralSWInfo& general_sw_info)
{
	osm_switch_t *p_sw;

	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	GuidToSWDataBaseEntryIter sw_node_it;
	// Case new switch was added
	if ((sw_node_it = this->m_sw_db.m_sw_map.find(general_sw_info.m_guid)) ==
	     this->m_sw_db.m_sw_map.end()) {
		ARSWDataBaseEntry new_sw_node(general_sw_info, &m_port_groups_calculator);
		GuidToSWDataBaseEntryInsertRes res = this->m_sw_db.m_sw_map.insert(
			pair<u_int64_t, ARSWDataBaseEntry>(general_sw_info.m_guid, new_sw_node));
		sw_node_it = res.first;
		sw_node_it->second.Init();
#ifdef _DEBUG_AR_UPDATE_DB
		printf("Switch GUID " U64H_FMT " was added to db\n",
		       sw_node_it->second.m_general_sw_info.m_guid);
#endif
	}

	// Save general info even if we know this switch already because
	// this general info might have been changed by osm
	sw_node_it->second.SetARGeneralSWInfo(general_sw_info);
	sw_node_it->second.in_temporary_error = false;

	p_sw = general_sw_info.m_p_osm_sw;
	/*
	 * No need to check update status of AR get MADS, since only ar_info get and
	 * plft_info get were sent
	 */
	if (p_sw->ar_data.ar_info_updated) {
		sw_node_it->second.m_ar_info = p_sw->ar_data.ar_info;
		MarkSWSupportAR(sw_node_it->second);
	} else {
		MarkSWNotSupportAR(sw_node_it->second, NO_ERR);
	}
	// Mark that update was done
	sw_node_it->second.m_app_data = ARMGR_SW_UPDATED;

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/***************/
void OSMAdaptiveRoutingManager::AddNewAndUpdateExistSwitches()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	osm_switch_t *p_osm_sw;
	// Go over only on switches in osm database to find new ones or changed ones
	for (p_osm_sw = (osm_switch_t *) cl_qmap_head(&this->m_p_osm_subn->sw_guid_tbl);
	     p_osm_sw != (osm_switch_t *) cl_qmap_end(&this->m_p_osm_subn->sw_guid_tbl);
	     p_osm_sw = (osm_switch_t *) cl_qmap_next(&p_osm_sw->map_item)) {

		u_int16_t device_id = cl_ntoh16(p_osm_sw->p_node->node_info.device_id);

		ARGeneralSWInfo general_sw_info(cl_ntoh64(osm_node_get_node_guid(p_osm_sw->p_node)),// GUID
				cl_ntoh16(osm_node_get_base_lid(p_osm_sw->p_node, 0)),              // Lid
				osm_node_get_num_physp(p_osm_sw->p_node) - 1,                       // Num ports
				p_osm_sw);                                                          // Switch ptr

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "AddNewAndUpdateExistSwitches GUID 0x%" PRIx64 " Lid: %u\n",
			   general_sw_info.m_guid, general_sw_info.m_lid);

		this->UpdateSW(general_sw_info);
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/***************/
void OSMAdaptiveRoutingManager::RemoveAbsentSwitches()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	map<u_int64_t, ARSWDataBaseEntry>::iterator aux_sw_node_it;
	while (sw_node_it !=  this->m_sw_db.m_sw_map.end()) {
		if (sw_node_it->second.m_app_data == ARMGR_SW_UPDATED) {
			++sw_node_it;
			continue;
		}

#ifdef _DEBUG_AR_UPDATE_DB
		printf("Switch GUID " U64H_FMT " was removed from db\n",
				sw_node_it->second.m_general_sw_info.m_guid);
#endif

		aux_sw_node_it = sw_node_it;
		++sw_node_it;

		this->m_sw_db.m_sw_map.erase(aux_sw_node_it);
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}


//============================================================================//
/***************/

void OSMAdaptiveRoutingManager::MarkSWNotSupportAR(ARSWDataBaseEntry &sw_db_entry,
		support_errs_t type)
{
	sw_db_entry.m_support[SUPPORT_AR] = NOT_SUPPORTED;
	sw_db_entry.m_error[SUPPORT_AR] = type;
}

bool OSMAdaptiveRoutingManager::IsARNotSupported(ARSWDataBaseEntry &sw_db_entry) {
	return (sw_db_entry.m_support[SUPPORT_AR] == NOT_SUPPORTED);
}

void OSMAdaptiveRoutingManager::MarkSWSupportAR(ARSWDataBaseEntry &sw_db_entry) {
	sw_db_entry.m_support[SUPPORT_AR] = SUPPORTED;
}

bool OSMAdaptiveRoutingManager::IsARActive(ARSWDataBaseEntry &sw_db_entry) {
	return (sw_db_entry.m_support[SUPPORT_AR] == SUPPORTED && sw_db_entry.m_option_on);
}

bool OSMAdaptiveRoutingManager::IsDFActive(ARSWDataBaseEntry &sw_db_entry) {
	return (IsARActive(sw_db_entry) &&
			sw_db_entry.m_support[SUPPORT_DF] == SUPPORTED);
}

int OSMAdaptiveRoutingManager::ARInfoGetGroupCapProcess()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	bool wait_for_responses = false;
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for ( ; sw_node_it != this->m_sw_db.m_sw_map.end() ; ++sw_node_it) {

		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
		int status;

		/*
		 * Step 1.1 -
		 * Skip if we know already that this switch not support AR
		 */
		if (this->IsARNotSupported(sw_node_it->second))
			continue;
		/*
		 * Step 1.2 -
		 * Determine required switch configuration.
		 */
		SetRequiredARInfo(sw_node_it->second);

		/*
		 * Step 1.3 -
		 * Query switch for it's capabilities according to required
		 * configuration.
		 */

		/* Check if the switch switch supports dynamic capability calculation */
		if (!(p_ar_data->ar_info.enabled_and_cap_fields &
		    IB_AR_INFO_CAP_MASK_DYN_CAP_CALC_SUP))
			continue;

		/*
		 * Check if capabilities according to required configuration
		 * are already known.
		 * If not, query the switch.
		 */
		if (!osm_ar_capabilities_search(p_ar_data, &sw_node_it->second.m_required_ar_info)) {
			status = osm_ar_query_ar_info(&m_p_osm->sm, p_sw,
						      &sw_node_it->second.m_required_ar_info);
			if (status) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8502: failed to query ARInfo on switch"
					   " 0x%" PRIx64 "\n",
					   cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
				continue;
			}
			wait_for_responses = true;
		}
	}

	/* If we sent queries, wait for responses */
	if (wait_for_responses)
		this->WaitForPendingTransactions();

	OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

int OSMAdaptiveRoutingManager::ValidateSwitchesCapabilities()
{
	// TODO: enhance capabilities search
	// for example, the search fail if the switch supports RN, but the
	// does not requires it.

	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for ( ; sw_node_it != this->m_sw_db.m_sw_map.end() ; ++sw_node_it) {

		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
		const ib_ar_info_t *p_req_ar_info;
		int status;

		if (this->IsARNotSupported(sw_node_it->second))
			continue;
		/*
		 * Check if the switch supports the required capabilities,
		 * if not, mark the switch not to use AR.
		 */
		p_req_ar_info = osm_ar_capabilities_search(p_ar_data,
							   &sw_node_it->second.m_required_ar_info);
		if (!p_req_ar_info) {
			sw_node_it->second.MarkSWNotSupport(SUPPORT_AR, AR_GROUP_CAP_ZERO);
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8503: switch 0x%" PRIx64 " does not support"
				   " required capabilities\n",
				   cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
		} else {
			osm_ar_set_required_ar_info(p_ar_data, p_req_ar_info);
			// HACK
			sw_node_it->second.m_required_ar_info = *p_req_ar_info;
			sw_node_it->second.m_ar_info = *p_req_ar_info;
		}
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

int OSMAdaptiveRoutingManager::ARInfoSetProcess()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	ib_ar_info_t smp_ar_info;
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for (; sw_node_it != this->m_sw_db.m_sw_map.end() ; ++sw_node_it) {

		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;

		if (this->IsARNotSupported(sw_node_it->second)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "AR not supported, ar information skipped.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			continue;
		}

		smp_ar_info = sw_node_it->second.m_required_ar_info;
		osm_ar_set_required_ar_info(p_ar_data, &smp_ar_info);

		/*
		 *  Set Group Top
		 */
		osm_ar_set_required_group_top(p_ar_data,
					      sw_node_it->second.m_group_top);
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

void OSMAdaptiveRoutingManager::HbfConfigProcess()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();

	for (; sw_node_it != this->m_sw_db.m_sw_map.end() ; ++sw_node_it) {

		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			"Switch GUID 0x%" PRIx64 ", LID %u: hbf is %ssupported, hbf is %senabled .\n",
			sw_node_it->second.m_general_sw_info.m_guid,
			sw_node_it->second.m_general_sw_info.m_lid,
			osm_ar_info_is_hbf_support(&sw_node_it->second.m_ar_info) ? "" : "not ",
			osm_ar_info_is_hbf_enabled(&sw_node_it->second.m_required_ar_info) ?
			"" : "not ");

		/* set hbf configuration when the bit is enabled */
		if (osm_ar_info_is_hbf_support(&sw_node_it->second.m_ar_info) &&
		    osm_ar_info_is_hbf_enabled(&sw_node_it->second.m_required_ar_info)) {
			/* configurable seed */
			p_sw->ar_data.req_hbf_config.hbf_seed_type =
				(m_p_osm_subn->opt.hbf_hash_type << 4) |
				(0x3 & m_p_osm_subn->opt.hbf_seed_type);

			/* take the lower 32 bits of the GUID */
			uint32_t seed = m_p_osm_subn->opt.hbf_seed == OSM_HBF_HASH_SEED_GUID ?
				sw_node_it->second.m_general_sw_info.m_guid & 0xFFFFFFFF :
				m_p_osm_subn->opt.hbf_seed;

			p_sw->ar_data.req_hbf_config.seed = cl_ntoh32(seed);

			/* currently only 32 LSB of 64 bits are defined */
			uint64_t hash_fields = 0xFFFFFFFF & cl_ntoh64(m_p_osm_subn->opt.hbf_hash_fields);

			/* disable specific hash fields bits if no capability */
			if (!osm_ar_info_is_hbf_bthdqp_hash_support(&sw_node_it->second.m_ar_info))
				hash_fields &= ~IB_HBF_CONFIG_HASH_FIELDS_BTH_DESTINATION_QP;
			if (!osm_ar_info_is_hbf_dceth_hash_support(&sw_node_it->second.m_ar_info)) {
				hash_fields &= ~IB_HBF_CONFIG_HASH_FIELDS_DCETH_V1_SRC_QP;
				hash_fields &= ~IB_HBF_CONFIG_HASH_FIELDS_DCETH_V1_ISID;
			}

			p_sw->ar_data.req_hbf_config.fields_enable = cl_hton64(hash_fields);
		}
	}
	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void OSMAdaptiveRoutingManager::TreeRoutingNotificationProcess()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	for (sw_node_it = this->m_sw_db.m_sw_map.begin();
	     sw_node_it != this->m_sw_db.m_sw_map.end();
	     ++sw_node_it) {
		if (this->IsARNotSupported(sw_node_it->second)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "AR not supported, ar information skipped.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			//we already suppose to have this switch in not supported list
			continue;
		}

		if (sw_node_it->second.in_temporary_error ||
		    !sw_node_it->second.m_config_rn)
			continue;

		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_physp_t *p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
		if (!p_physp)
			continue;

		osm_port_t *p_port = osm_get_port_by_guid(m_p_osm_subn, p_physp->port_guid);
		if (!p_port)
			continue;

		/*
		 * Step 1 -
		 *  RNGenStringTable
		 */

		uint16_t max_string = (uint16_t)
			((0x1 << ib_ar_info_get_string_width_cap(&sw_node_it->second.m_ar_info)) - 1);

		if (p_sw->rank > max_string) {
			if (p_port->disc_through == OSM_DISC_NORMAL) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8508: Switch GUID 0x%" PRIx64 ", LID %u, will not generate RN"
					    " because sw rank:%u is greater than max_string:%u.\n",
					    sw_node_it->second.m_general_sw_info.m_guid,
					    sw_node_it->second.m_general_sw_info.m_lid,
					    p_sw->rank, max_string);
			} else {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Switch GUID 0x%" PRIx64 ", LID %u, will not generate RN"
					    " because sw rank:%u is greater than max_string:%u and disc_through unhealthy.\n",
					    sw_node_it->second.m_general_sw_info.m_guid,
					    sw_node_it->second.m_general_sw_info.m_lid,
					    p_sw->rank, max_string);
			}
		} else {
			osm_ar_rn_direction_string_set(&p_sw->ar_data, 0, p_sw->rank);
		}

		/*
		 * Step 2 -
		 *  RNGenBySubGroupPriority
		 */

		uint8_t flags;
		flags = (IS_GENERATE_ARN(m_master_db, sw_node_it->second) ? 0x1 : 0) |
			(IS_GENERATE_FRN(m_master_db, sw_node_it->second) ? 0x2 : 0);

		osm_ar_subgroup_pri_rn_flags_set(&p_sw->ar_data, 0, flags);
		if ((m_p_osm_subn->opt.ar_tree_asymmetric_flow ==
		     OSM_ASYMMETRIC_TREE_2_SUB_GROUPS_ENABLED) &&
		    sw_node_it->second.m_is_leaf && sw_node_it->second.m_num_cas)
			osm_ar_subgroup_pri_rn_flags_set(&p_sw->ar_data, 1, flags);
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/***************/
void OSMAdaptiveRoutingManager::ARDumpSWSettings(ARSWDataBaseEntry &sw_db_entry)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/***************/

void OSMAdaptiveRoutingManager::SetPortGroups(ARSWDataBaseEntry &sw_db_entry,
					      GroupsData &groups_data)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	if (sw_db_entry.m_config_rn) {
		/*
		 * In case we use group_per_leaf_sw for RN
		 * we optimize ARGroup MAD sending by using copy
		 */
		SetPortGroupsForRn(sw_db_entry, groups_data);
		OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
	}

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "SetPortGroups for Switch GUID: 0x%" PRIx64 ", LID: %u, \n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid);
	/*
	 *  Update calculated_ar_group_table:
	 *  At this point we have all assigned groups in groups_data.m_assign_groups list
	 */
	GroupsListIter groups_iter;
	for (groups_iter = groups_data.m_assign_groups.begin();
	     groups_iter != groups_data.m_assign_groups.end();
	     ++groups_iter) {

		if ((*groups_iter)->m_group_size == 1)
			continue; /* this is Static port not a group */

		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "UpdateSWGroupElement group_number: %u\n",
				   (*groups_iter)->m_group_number);
			PrintGroupData("UpdateSWGroupElement: ",*(*groups_iter));
		}

		(*groups_iter)->UpdateSWGroupElement(
			m_p_osm_log,
			sw_db_entry,
			(*groups_iter)->m_group_number);
	}

	sw_db_entry.m_group_top = 0;
	if (!groups_data.m_assign_groups.empty()) {
		sw_db_entry.m_group_top =
			groups_data.m_next_group_number - 1;
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void OSMAdaptiveRoutingManager::SetPortGroupsForRn(ARSWDataBaseEntry &sw_db_entry,
						   GroupsData &groups_data)
{
	osm_ar_subgroup_t *p_subgroup;
	PortsBitset empty_group;
	osm_advanced_routing_data_t *p_ar_data =
		&sw_db_entry.m_general_sw_info.m_p_osm_sw->ar_data;

	/* Set group 0 to be empty group */
	for (int sub_group = 0; sub_group < p_ar_data->num_group_subgroups;
	     ++sub_group) {
		p_subgroup = osm_ar_subgroup_get(p_ar_data, 0, sub_group);
		empty_group.get_data(*p_subgroup);
	}

	pthread_mutex_lock(&m_sw_db.m_db_lock);
	ARUpdateSWGroupPerPLFT(sw_db_entry, groups_data.m_groups);
	pthread_mutex_unlock(&m_sw_db.m_db_lock);
}

void OSMAdaptiveRoutingManager::ARUpdateSWGroupPerPLFT(ARSWDataBaseEntry &sw_db_entry,
						       GroupsPtrVec &groups)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	uint16_t group_number = 0;
	uint16_t group_top = 0;
	osm_ar_subgroup_t calculated_element;
	osm_ar_subgroup_t *p_current_element;
	int element_cmp;
	uint16_t group_block_idx;
	/* When using group per island groups in m_group_numbers may be non unique */
	vector <bool> used_groups;
	uint8_t plft;
	uint8_t unique_plfts_num = UCAST_MGR_MAX_PLFT;
	used_groups.reserve(sw_db_entry.GetGroupCap());

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "ARUpdateSWGroupPerPLFT "   \
		   "Switch GUID 0x%" PRIx64 ", LID %u groups numbrt: %u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   (uint16_t)groups.size());

	/*
	 * For pFRN configuration, PLFT 0 and PLFT 2 are using the same groups per LID.
	 * Hence, no need to set the group twice.
	 */
	if (IsGeneratePFRN())
		unique_plfts_num = UCAST_MGR_MAX_DFP2_UNIQUE_PLFT;

	GroupsPtrVec::iterator groups_iter = groups.begin();
	for ( ;groups_iter != groups.end(); ++groups_iter) {
		for (plft = 0; plft < unique_plfts_num; plft++) {
			uint16_t max_group_number = 0;

			if ((*groups_iter)->m_group_numbers[plft].size() == 0)
				continue;

			if (m_p_osm_log->level & OSM_LOG_DEBUG) {
				stringstream sstr;
				(*groups_iter)->BitsetToOstream(sstr);
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "group_numbers size: %u bitmask: %s plft: %u\n",
					   (uint16_t)(*groups_iter)->m_group_numbers[plft].size(),
					   sstr.str().c_str(), plft);
			}

			vector<uint16_t>::iterator iter = (*groups_iter)->m_group_numbers[plft].begin();
			for (; iter != (*groups_iter)->m_group_numbers[plft].end(); ++iter) {
				max_group_number = max(max_group_number, *iter);
			}

			used_groups.clear();
			used_groups.insert(used_groups.begin(), max_group_number + 1, false);

			iter = (*groups_iter)->m_group_numbers[plft].begin();
			for (; iter != (*groups_iter)->m_group_numbers[plft].end(); ++iter) {
				if (used_groups[*iter]) {
					continue;
				}

				used_groups[*iter] = true;
				(*groups_iter)->UpdateSWGroupElement(m_p_osm_log, sw_db_entry, *iter);
			}

			group_top = max(group_top, max_group_number);
		}
	}

	sw_db_entry.m_group_top = group_top;

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void OSMAdaptiveRoutingManager::ARUpdateSWGroupTableTop(
		ARSWDataBaseEntry &sw_db_entry,
		uint16_t calculated_groups_number)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	if (calculated_groups_number == 0) {
		sw_db_entry.m_group_top = 0;
		return;
	}

	sw_db_entry.m_group_top = calculated_groups_number - 1;
	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//will consume strings
// string == sw_rank - 1
// sw_rank < string < max_consume_rank
//if fabric uses parallel links max_consume_rank == sw_rank + 1
// will consume only string == sw_rank - 1 (like pure fat tree)
void OSMAdaptiveRoutingManager::UpdateRNRcvString(ARSWDataBaseEntry &sw_db_entry,
						  uint8_t max_rank,
						  uint8_t sw_rank,
						  uint8_t max_consume_rank)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	osm_switch_t *p_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_ar_rn_string_handling_clear(p_ar_data);

	uint8_t consume_decision =
		(IS_GENERATE_FRN(m_master_db, sw_db_entry) ?
		 RN_RCV_STRING_CONSUME_ALL : RN_RCV_STRING_CONSUME_ARN);

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "UpdateRNRcvString Switch GUID 0x%" PRIx64 ", LID %u "
		   " max_rank:%u sw_rank:%u"
		   " max_consume_rank:%u consume_decision:%u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   max_rank, sw_rank, max_consume_rank, consume_decision);
	uint8_t max_rn_string_cap = p_ar_data->num_strings;
	if (sw_rank >= max_rn_string_cap || max_consume_rank > max_rn_string_cap || max_rank > max_rn_string_cap)
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
			   "Switch GUID 0x%" PRIx64 ", LID %u routing notification"
			   " can't be set. The routing notification capability is %u."
			   " (max_rank: %u sw_rank: %u max_consume_rank: %u)\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   max_rn_string_cap, max_rank, sw_rank, max_consume_rank);

	if (sw_rank != 0 && sw_rank < max_rn_string_cap) {
		// consume sw_rank - 1;
		osm_ar_rn_string_handling_set(p_ar_data, sw_rank - 1,
					      consume_decision, 0, 0);

		// pass on sw_rank
		osm_ar_rn_string_handling_set(p_ar_data, sw_rank,
					      RN_RCV_STRING_PASS_ON, sw_rank, 0);
	}

	uint8_t rank = sw_rank + 1;

	for ( ;rank < max_consume_rank && rank < max_rn_string_cap; ++rank)
		osm_ar_rn_string_handling_set(p_ar_data, rank,
					      consume_decision, 0, 0);

	for ( ;rank < max_rank && rank < max_rn_string_cap; ++rank)
		osm_ar_rn_string_handling_set(p_ar_data, rank,
					      RN_RCV_STRING_PASS_ON, rank, 0);

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void OSMAdaptiveRoutingManager::UpdateRNXmitPortMask(ARSWDataBaseEntry &sw_db_entry,
						     PortsBitset &ca_ports,
						     PortsBitset &sw_ports,
						     bool is_down_sw)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	osm_switch_t *p_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;

	if (is_down_sw) {
		u_int8_t calc_mask = 0;

		if(IS_GENERATE_ARN(m_master_db, sw_db_entry))
			calc_mask |= RN_XMIT_PORT_MASK_GENERATE_ARN;

		if(IS_GENERATE_FRN(m_master_db, sw_db_entry))
			calc_mask |= RN_XMIT_PORT_MASK_GENERATE_FRN;

		calc_mask |= RN_XMIT_PORT_MASK_PASS_ON_RN;

		for (u_int8_t port_num = 1;
		     port_num <= sw_db_entry.m_general_sw_info.m_num_ports;
		     ++port_num) {

			if (ca_ports.test(port_num))
				/* Do not send RN to CA ports */
				osm_ar_rn_port_xmit_mask_add_port(p_ar_data, port_num, 0);
			else if (sw_ports.test(port_num))
				osm_ar_rn_port_xmit_mask_add_port(p_ar_data, port_num, calc_mask);
		}

		if (sw_db_entry.m_to_set_rn_xmit_port_mask) {
			PortsBitset disconnected_ports = ca_ports | sw_ports;
			disconnected_ports.flip();

			for (u_int8_t port_num = 1;
			     port_num <= sw_db_entry.m_general_sw_info.m_num_ports;
			     ++port_num) {
				if (disconnected_ports.test(port_num))
					osm_ar_rn_port_xmit_mask_add_port(p_ar_data, port_num, 0);
			}
		}

	} else
		//clear all xmit_port_mask
		osm_ar_rn_port_xmit_mask_clear(p_ar_data);

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/***************/
void OSMAdaptiveRoutingManager::ARCalculatePortGroupsParallelLinks()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	// some variables for calculation
	PortsBitset ports_per_remote_node[IB_LID_UCAST_END_HO + 1];
	u_int8_t port_2_group[MAX_SUPPORTED_PORTS];

	unsigned calculated_max_pg_id;
	unsigned calculated_max_lid;

	PortsBitset group_bitset;

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "Calculate Port Groups Parallel Links. (switches number:%u)\n",
		   (int)this->m_sw_db.m_sw_map.size());

	/*
	 * Go through all the switches and create their port groups.
	 * In this algorithm ports groups are created out of "parallel"
	 * links - links that are connecting the same pair of nodes.
	 * So here's what we're doing here:
	 *   - for each switch
	 *       - for each port of the switch (port_num)
	 *           - get the remote node
	 *           - if the remote node is CA or RTR
	 *               - no need to create a group, as there can be
	 *                 only one link to its port - skip this port
	 *           - if remote node is switch
	 *               - get remote switch's LID (remote_lid)
	 *               - set bit num. port_num in the bitmask that holds
	 *                 bit for every port that has this LID as a remote
	 *                 side of the link (ports_per_remote_node[remote_lid])
	 *       - at this point we have all the ports organized in groups
	 *         by the remote node (ports_per_remote_node[])
	 *       - set the groups table according to ports_per_remote_node[]
	 *       - create port_2_group map
	 *       - set the lft table according to port_2_group and ucast_lft_port used before
	 */
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
		/*
		 * Step 1.1 -
		 *  Ensure that this switch supports AR, and that AR is enabled.
		 */
		if (!this->IsARActive(sw_node_it->second)) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID 0x%" PRIx64", LID %u: " \
				   "AR/FR not supported or not enabled, calculate of port groups skipped.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			//we already suppose to have this switch in not supported list
			goto next_switch;
		}
		/*
		 * Step 1.2 -
		 *  Set default values for calculations
		 */
		for (unsigned lid_num = 0; lid_num <= m_p_osm_subn->max_lid; ++lid_num)
			ports_per_remote_node[lid_num].reset();
		memset(port_2_group, OSM_NO_PATH, sizeof(port_2_group));// default is no group for this port

		calculated_max_pg_id = m_master_db.m_fr_enable ? 1 : 0;
		calculated_max_lid = 0;


		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculate LAG groups for Switch GUID 0x%" PRIx64 ", LID %u \n",
			   sw_node_it->second.m_general_sw_info.m_guid,
			   sw_node_it->second.m_general_sw_info.m_lid);

		/*
		 * Step 2 -
		 *  Iterate through all the switch ports and fill ports_per_remote_node
		 *  We create bitmask for each remote node that is connected to this switch.
		 *  Each bit that is set represent a port number that connected to the remote node that
		 *  this bit mask belongs to.
		 */
		for (u_int8_t port_num = 1;
		     port_num <= sw_node_it->second.m_general_sw_info.m_num_ports;
		     ++port_num) {
			osm_node_t *p_node = sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;

			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_node, port_num);
			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;
			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;

			osm_node_t *p_remote_node = osm_physp_get_node_ptr(p_remote_physp);
			// Ignore any loopback connection on switch
			if (p_node == p_remote_node) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Ignoring loopback on Switch GUID 0x%" PRIx64
					   ", LID %u, during calculate of port groups.\n",
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid);
				continue;
			}

			u_int64_t remote_node_guid = cl_ntoh64(osm_node_get_node_guid(p_remote_node));
			u_int8_t remote_node_type = osm_node_get_type(p_remote_node);
			uint16_t remote_node_lid;
			if (remote_node_type == IB_NODE_TYPE_ROUTER || remote_node_type == IB_NODE_TYPE_CA) {
				remote_node_lid = cl_ntoh16(osm_physp_get_base_lid(p_remote_physp));
			} else if (remote_node_type == IB_NODE_TYPE_SWITCH) {
				remote_node_lid = cl_ntoh16(osm_node_get_base_lid(p_remote_node, 0));
			} else {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 8509: Switch GUID 0x%" PRIx64 " - Unknown node type: "  \
					   "%u, calculate of port groups skipped.\n",
					   remote_node_guid, remote_node_type);
				throw 1;
			}

			// Mark that port number port_num routes to lid remote_node_lid.
			// Note that bit number represent port number.
			ports_per_remote_node[remote_node_lid].set(port_num);

			if (this->m_p_osm_log->level & OSM_LOG_DEBUG)
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Update ports_per_remote_node on "  \
					   "Switch GUID 0x%" PRIx64 ", LID: %u : port_num %u, " \
					   "ports_per_remote_node[%u]=%s.\n",
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid,
					   port_num,
					   remote_node_lid,
					   ports_per_remote_node[remote_node_lid].to_string().c_str());

		}   /* End of loop switch ports */

		/*
		 * Step 3 -
		 *  Create all port groups in the following way:
		 *  At this point we have all the ports marked their bits in the ports_per_remote_node array
		 *  Bits that are marked in the same array entry effectively form port group.
		 *
		 *  Number of port groups <= number of switch ports.
		 *
		 *  Check all the entries of the ports_per_remote_node array.
		 *  If an entry contain more than one port and the entry is not all 0, create port group.
		 */
		for (unsigned lid_num = 1; lid_num <= m_p_osm_subn->max_lid; ++lid_num) {
			if (!ports_per_remote_node[lid_num].any())
				continue;

			// count number of ports that are connected directly to the same LID (remote node) */
			u_int8_t ports_counter = 0;
			for (unsigned i = 1; i <= sw_node_it->second.m_general_sw_info.m_num_ports; ++i)
				if (ports_per_remote_node[lid_num].test(i))
					++ports_counter;

			// If there is only one port in such group, then the AR mode would be Static, and there is no
			// need to create port group for it - the GroupNumber in the LFT entry for Static LID
			// State is actually a port number. So, if we see a port that is mapped to group
			// id OSM_NO_PATH, it means that the port is alone - no port group was created for it.
			if (ports_counter == 1)
				continue;

			osm_ar_subgroup_t *p_subgroup = osm_ar_subgroup_get(p_ar_data, calculated_max_pg_id, 0);
			ports_per_remote_node[lid_num].get_data(*p_subgroup);

			++calculated_max_pg_id;
		}      /* End of loop all lids */
		if (calculated_max_pg_id)
			--calculated_max_pg_id;

		/*
		 * Step 4.1 -
		 *  Init port-2-group mapping.
		 *  Value OSM_NO_PATH (255) in port-2-group mapping means that the port is mapped to itself,
		 *  which is valid in case of Static LID State.
		 */
		for (unsigned pg_id = 0; pg_id <= calculated_max_pg_id; ++pg_id) {
			osm_ar_subgroup_t *p_subgroup = osm_ar_subgroup_get(p_ar_data, pg_id, 0);
			group_bitset.set(*p_subgroup);
			if (!group_bitset.any())
				// no ports in this group - skip it
				continue;

			for (unsigned i = 0;
			     i <= sw_node_it->second.m_general_sw_info.m_num_ports;
			     ++i)
				if (group_bitset.test(i))
					port_2_group[i] = pg_id;
		}      /* End of loop valid port groups */
		//TODO enter assert regarding port zero

		/*
		 * Step 4.2 -
		 *  Now let's configure AR LFTs.
		 */
		for (unsigned dest_lid = 1; dest_lid <= m_p_osm_subn->max_lid; ++dest_lid) {
			// Find out which port was used to route to lid_num in the usual unicast routing
			u_int8_t ucast_lft_port =
				OsmSwitchGetPortByLidUseNewLft(
						sw_node_it->second.m_general_sw_info.m_p_osm_sw,
						dest_lid);
			if (ucast_lft_port == OSM_NO_PATH)  //no route to this lid_num from this switch
				continue;
			calculated_max_lid = dest_lid;


			// Check port_2_group mapping of this port.
			// If there is no port group for this port (the mapped port group ID is OSM_NO_PATH),
			// then it means that the port is alone. In that case, LID state would be Static,
			// which means that the GroupNumber in the LFT entry is actually a port number.
			//
			// Note: if (lid_num == my LID), then (ucast_lft_port == 0), and its port_2_group
			// mapping will be OSM_NO_PATH. This is loopback scenario.
			unsigned pg_id = port_2_group[ucast_lft_port];

			osm_advanced_routing_data_t *p_ar_data =
				&sw_node_it->second.m_general_sw_info.m_p_osm_sw->ar_data;
			if (pg_id != OSM_NO_PATH) {
				osm_ar_lft_set(p_ar_data, dest_lid, 0, ucast_lft_port,
					       (this->m_master_db.m_ar_enable ?
					       this->m_master_db.m_ar_mode :
					       OSM_AR_IB_LID_STATE_STATIC),
					       pg_id);
			} else
				osm_ar_lft_set(p_ar_data, dest_lid, 0, ucast_lft_port,
						OSM_AR_IB_LID_STATE_STATIC, 0);
		}      /* End of loop destination lids */

		/*
		 * Step 5 -
		 *  Now we will decide if the calculated settings are different from old,
		 *  or if we need to ignore old settings and use the new ones
		 */
		this->ARUpdateSWGroupTableTop(sw_node_it->second,
					      calculated_max_pg_id + 1);

next_switch:
		++sw_node_it;
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void OSMAdaptiveRoutingManager::ARCalculatePortGroupsMinhop()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
	osm_subn_t *p_subn = this->m_p_osm_subn;
	int ar_enable = this->m_master_db.m_ar_enable;
	int ar_mode   = this->m_master_db.m_ar_mode;
	uint16_t max_pg_id;

	/*
	 * Run the AR code for the subnet in parallel mode.  The second parameter is
	 * actually a pointer to the structure switch internal minhop_t.
	 */
	osm_minhop_ar_process_subnet(p_subn, p_sw->re_priv, ar_enable, ar_mode);

	/*
	 * Get the max_pg_id for each switch.
	 */
	for (sw_node_it = this->m_sw_db.m_sw_map.begin();
	     sw_node_it != this->m_sw_db.m_sw_map.end();
	     sw_node_it++) {
		osm_switch_t *p_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
		max_pg_id = osm_minhop_ar_process_switch(p_subn, p_sw);
		this->ARUpdateSWGroupTableTop(sw_node_it->second, max_pg_id);
	}

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

uint16_t OSMAdaptiveRoutingManager::GetGroupNumberRN(uint16_t sw_lid_num,
						     ARSWDataBaseEntry &sw_db_entry,
						     uint8_t plft_id,
						     AnalyzeDFSetupData &setup_data)
{
	if (setup_data.sw_lid_to_df_data_mapping[sw_lid_num]->
	    m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF)
		return AllocateSwArGroup(sw_lid_num,
					 sw_db_entry.GetGroupCap(),
					 plft_id);
	return AR_GROUP_UNASSIGNED;
}

int OSMAdaptiveRoutingManager::AssignGroupIDPerPlftLeafSwitch(ARSWDataBaseEntry &sw_db_entry,
							      GroupsData &groups_data,
							      uint8_t plft_id)
{
	int rc = 0;

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "AssignGroupIDPerPlftLeafSwitch for Switch GUID: 0x%" PRIx64
		   ", LID: %u, groups number: %u, plft_id %u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   (int) groups_data.m_groups.size(), plft_id);

	uint16_t group_number = 0;
	uint16_t lid_num;
	uint8_t plft;
	bool is_group_cap_exceeded = false;

	/*
	 * Divide GroupCap by m_groups_per_leaf_ since each allocated
	 * group_base is used to define m_groups_per_leaf_ separate groups
	 */
	uint16_t group_cap =
		 sw_db_entry.GetGroupCap() / groups_data.m_groups_per_leaf;

	GroupsPtrVec::iterator iter = groups_data.m_groups.begin();
	for (; iter != groups_data.m_groups.end(); ++iter) {
		(*iter)->m_group_number = AR_GROUP_UNASSIGNED;
		(*iter)->m_group_table_number = 0;

		for (plft = 0; plft < UCAST_MGR_MAX_PLFT; plft++) {
			vector<uint16_t>::const_iterator lid_iter =
				(*iter)->m_lid_per_plft[plft].begin();
			for (; lid_iter != (*iter)->m_lid_per_plft[plft].end() ; ++lid_iter) {
				lid_num = (*lid_iter);
				group_number = AllocateSwArGroup(lid_num, group_cap, plft);

				if ((*iter)->m_group_number == AR_GROUP_UNASSIGNED)
					(*iter)->m_group_number = group_number;

				if (group_number == AR_GROUP_UNASSIGNED) {
					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
						   "Switch GUID 0x%" PRIx64 ", "
						   "Can not assign group (0x%p) for lid: %u "
						   "group_cap: %u groups_per_leaf: %u plft: %u\n",
						   sw_db_entry.m_general_sw_info.m_guid,
						   *iter, lid_num,
						   group_cap,
						   groups_data.m_groups_per_leaf, plft);
						   is_group_cap_exceeded = true;
					continue;
				}

				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Assign group %u for leaf switch lid:%u plft %u\n",
					   group_number, lid_num, plft);

				(*iter)->m_group_numbers[plft].push_back(group_number);
			}

			(*iter)->ClearVector((*iter)->m_lid_per_plft);
		}
	}

	if (is_group_cap_exceeded)
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 851A: Switch GUID 0x%" PRIx64
			   ", Can not assign separate groups number"
			   " for all leaf switches. group_cap: %u\n",
			   sw_db_entry.m_general_sw_info.m_guid, group_cap);

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}


int OSMAdaptiveRoutingManager::AssignPortGroupsDFP(ARSWDataBaseEntry &sw_db_entry,
						       GroupsData &groups_data,
						       uint8_t plft_id)
{
	int rc = 0;
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	if (sw_db_entry.m_config_rn) {
		rc = AssignGroupIDPerPlftLeafSwitch(sw_db_entry, groups_data, plft_id);
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	rc = AssignPortGroups(sw_db_entry, groups_data);
	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}


u_int16_t OSMAdaptiveRoutingManager::AllocateSwArGroup(uint16_t sw_lid,
						       uint16_t group_cap,
						       uint8_t plft_id)
{
	uint8_t adjusted_plft_id;
	//After we configure the fabric once, all leaf switches should be assigned
	//with AR groups so we optimize by not locking m_sw_db.m_db_lock on read

	/*
	 * Used for pFRN configuration by DFP2.
	 * Although Group IDs are per LID and PLFT, AR Groups of PLFT0 and PLFT2 shall be identical
	 * (same group IDs and exit ports for each destination).
	 * Hence, when requesting for PLFT 2, should be handled as if PLFT 0 was requested.
	 * Note: PLFT2 is required due to lack of some entries comparing to PLFT0.
	 */
	adjusted_plft_id = ((plft_id == PLFT_ID_FROM_DOWN_DIRECTION_VL1) && IsGeneratePFRN()) ?
	    PLFT_ID_FROM_UP_DIRECTION : plft_id;

	pthread_mutex_lock(&m_sw_db.m_db_lock);
	uint16_t old_group_id = osm_ucast_mgr_get_groupid(&this->m_p_osm->sm.ucast_mgr,
							  sw_lid, adjusted_plft_id);
	/*
	 * If we use adjusted PLFT and its group ID is different than original PLFT group id,
	 * use this parameter to indicate that original  PLFT group ID should be updated with
	 * group ID of adjusted PLFT.
	 */
	boolean_t update_origin_plft_id = (adjusted_plft_id != plft_id) &&
	    (old_group_id != osm_ucast_mgr_get_groupid(&this->m_p_osm->sm.ucast_mgr,
						       sw_lid, plft_id));

	if (old_group_id != AR_GROUP_UNASSIGNED) {
		if (update_origin_plft_id)
			osm_ucast_mgr_set_groupid(&this->m_p_osm->sm.ucast_mgr,
						  sw_lid, old_group_id, plft_id);

		pthread_mutex_unlock(&m_sw_db.m_db_lock);
		if (old_group_id < group_cap)
			return old_group_id;
		else
			return AR_GROUP_UNASSIGNED;
	}

	uint16_t new_group_id = AR_GROUP_UNASSIGNED;

	new_group_id = osm_ucast_mgr_find_free_groupid(&this->m_p_osm->sm.ucast_mgr);

	if (new_group_id != AR_GROUP_UNASSIGNED) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "AllocateSwArGroup Sw_Lid:%u plft: %u Group:%u.\n",
			   sw_lid, plft_id, new_group_id);
		osm_ucast_mgr_set_groupid(&this->m_p_osm->sm.ucast_mgr,
					  sw_lid, new_group_id, adjusted_plft_id);
		if (update_origin_plft_id) {
			osm_ucast_mgr_set_groupid(&this->m_p_osm->sm.ucast_mgr,
						  sw_lid, new_group_id, plft_id);
		}
	} else {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
			   "Failed to AllocateSwArGroup Sw_Lid:%u.\n",
			   sw_lid);
	}
 
	pthread_mutex_unlock(&m_sw_db.m_db_lock);

	return new_group_id < group_cap ? new_group_id : AR_GROUP_UNASSIGNED;
}

int OSMAdaptiveRoutingManager::SetHcaToSwLidMapping(
	osm_physp_t *p_hca_physp,
	osm_node_t *p_remote_sw_node,
	u_int16_t hca_to_sw_lid_mapping[IB_MAX_NUM_OF_PLANES_XDR + 1][IB_LID_UCAST_END_HO + 1])
{
	u_int16_t node_lid = cl_ntoh16(osm_physp_get_base_lid(p_hca_physp));
	if (p_remote_sw_node->node_info.node_type != IB_NODE_TYPE_SWITCH ||
	    node_lid == 0 )
		return -1;

	uint8_t lmc = osm_physp_get_lmc(p_hca_physp);
	uint8_t physp_plane = osm_hierarchy_is_planarized_physp(p_hca_physp) ?
			          p_hca_physp->hierarchy_info.plane_number : 0;
	uint8_t physp_plane_mask = physp_plane ? (1 << physp_plane) : 0;

	u_int16_t remote_node_lid = cl_ntoh16(osm_node_get_base_lid(p_remote_sw_node, 0));

	/*
	 * The mapping is between HCA lid to sw lid.
	 * In planarized network, all plane ports of planarized HCA have the same lid.
	 * Don't map plane port that has different plane than remote switch.
	 * This check is irrelevant for legacy (non-planarized) HCA port.
	 * This is extra check for sanity.
	 * Planarized HCA has to have a common plane with remote switch.
	 */
	if (physp_plane_mask && !(physp_plane_mask & p_remote_sw_node->sw->plane_mask)) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR, "ERR 851B: "
			   "Failed to map (HCA/Router) LID %u (plane: %u) to SW LID %u "
			   "(plane mask: %u).\n",
			   node_lid, physp_plane, remote_node_lid,
			   p_remote_sw_node->sw->plane_mask);
		return -1;
	}

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "Map (HCA/Router) LID %u lmc:%u (plane: %u) to SW LID %u.\n",
		   node_lid, lmc, physp_plane, remote_node_lid);

	u_int16_t numLidsPerPort = (u_int16_t)((u_int16_t)1 << lmc);
	u_int16_t lid_limit = node_lid+numLidsPerPort;

	for (u_int16_t l = node_lid; l < lid_limit; l++) {
		hca_to_sw_lid_mapping[physp_plane][l] = remote_node_lid;
	}

	return 0;
}

void OSMAdaptiveRoutingManager::ARCalculatePortGroupsTree()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    void **adr_vertices;

    TreeAlgorithmContext *context = new TreeAlgorithmContext();
    if (!context) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "Failed to allocate TreeAlgorithmContext\n");
        throw 1;
    }

    if (m_p_osm_subn->opt.ar_tree_asymmetric_flow)
        context->p_graph = osm_flow_build_graph(m_p_osm_subn);

    //Minimal hops array contains information for switches only
    //find for each hca the connected switch and fill in hca_to_sw_lid_mapping

    osm_node_t *p_osm_node;
    osm_port_t *p_port;
    osm_physp_t *p_physp;
    // Go over all nodes in osm database
    for (p_osm_node =  (osm_node_t *) cl_qmap_head(&m_p_osm_subn->node_guid_tbl);
         p_osm_node != (osm_node_t *) cl_qmap_end(&m_p_osm_subn->node_guid_tbl);
         p_osm_node =  (osm_node_t *) cl_qmap_next(&p_osm_node->map_item)) {

        if (p_osm_node->node_info.node_type == IB_NODE_TYPE_SWITCH) {
            u_int16_t lid = cl_ntoh16(osm_node_get_base_lid(p_osm_node, 0));

            p_physp = osm_node_get_physp_ptr(p_osm_node, 0);
            if (!p_physp)
                continue;

            p_port = osm_get_port_by_guid(m_p_osm_subn, p_physp->port_guid);
            if (!p_port)
                continue;

            context->m_sw_lids.push_back(lid);

            //set context->sw_lid_to_rank_mapping
            if (m_master_db.m_frn_enable || m_master_db.m_arn_enable) {
		if (p_port->disc_through != OSM_DISC_NORMAL)
                    continue;

                if (p_osm_node->sw->rank == OSM_SW_NO_RANK) {
                    m_master_db.m_frn_enable = false;
                    m_master_db.m_arn_enable = false;

                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                            "ERR 8517: Switch GUID 0x%" PRIx64  ", LID %u: "
                            "has no rank - can not support ARN/FRN.\n",
                            cl_ntoh64(osm_node_get_node_guid(p_osm_node)),
                            lid);

                } else if (p_osm_node->sw->rank > MAX_SUPPORTED_RANK) {
                    m_master_db.m_frn_enable = false;
                    m_master_db.m_arn_enable = false;

                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
                            "Switch GUID 0x%" PRIx64  ", LID %u: "
                            "has rank:%u which exceeds maximum.\n",
                            cl_ntoh64(osm_node_get_node_guid(p_osm_node)),
                            lid,
                            p_osm_node->sw->rank);

                }

                context->m_sw_lid_to_rank_mapping[lid] = p_osm_node->sw->rank;
                context->m_max_rank = max(context->m_max_rank, p_osm_node->sw->rank);
            }

            for (uint8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; ++port_num) {
                uint8_t remote_port;
                osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
                osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, &remote_port);

                if (!p_remote_node || !p_physp || !osm_link_is_healthy(p_physp))
                    continue;

                if (p_remote_node->node_info.node_type == IB_NODE_TYPE_SWITCH)
                    continue;

                osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
                uint16_t remote_lid = cl_ntoh16(osm_node_get_base_lid(p_remote_node, remote_port));
                uint16_t max_remote_lid = remote_lid + (1 << osm_physp_get_lmc(p_remote_physp));

                /* don't add Aggregation Nodes or Routers as valid destination for AR/FR */
		ib_special_port_type_t special_type =
			osm_phsyp_get_special_port_type(p_physp->p_remote_physp);
		if (special_type == IB_SPECIAL_PORT_TYPE_AN ||
		    special_type == IB_SPECIAL_PORT_TYPE_ROUTER)
			continue;

                for (; remote_lid < max_remote_lid; ++remote_lid) {
		    osm_switch_t *p_sw = p_osm_node->sw;
                    uint16_t group_cap = ib_ar_info_get_group_cap_ho(&p_sw->ar_data.ar_info);
		    /* Only PLFT 0 is used by tree topologies */
		    AllocateSwArGroup(lid, group_cap, 0);
                    context->m_sw_to_lids[lid].push_back(remote_lid);
                }

                SetHcaToSwLidMapping(p_remote_physp, p_osm_node,
                                     context->m_hca_to_sw_lid_mapping);
            }

            continue;
        }
    }

    if (m_p_osm_subn->opt.ar_tree_asymmetric_flow)
        osm_flow_calculate_parallel_flow(context->p_graph);

    /*
     * In this algorithm ports groups are created out of minimal hops
     * links - each lid will receive a group containing all the ports
     * with minimal hops to this lid.
     * So here's what we're doing here:
     *   - for each switch
     *       - for each port of the switch (port_num)
     *           - calculate group bitmask for this lid
     *           - asisign groups according to device limitations
     */
    m_port_groups_calculator.CalculatePortGroupsTree(context);
    if (m_p_osm_subn->opt.ar_tree_asymmetric_flow)
        osm_flow_destroy_graph(&context->p_graph);
    delete context;


    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void
OSMAdaptiveRoutingManager::ARSetUcastPort(ARSWDataBaseEntry &sw_db_entry, PSGroupData &group_data,
                                          uint16_t lid, int first_port_in_group)
{
    osm_switch_t *p_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
    uint8_t ucast_lft_port =
        OsmSwitchGetPortByLidUseNewLft(p_sw, lid);

    if (!group_data.m_ps_group_bitmask.test_pri(ucast_lft_port)) {
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Switch GUID 0x%" PRIx64 " : for lid_dest: %u, "
		   "ucast port not in group: %u change to port: %u\n",
		   sw_db_entry.m_general_sw_info.m_guid, lid,
		   ucast_lft_port, first_port_in_group);
        p_sw->new_lft[lid] = first_port_in_group;
    }
}

void
OSMAdaptiveRoutingManager::ARCalculateSwitchPortGroupsTree(
    ARSWDataBaseEntry &sw_db_entry,
    TreeAlgorithmContext *context)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    // some variables for calculation
    osm_switch_t *p_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw, *p_dest_sw;
    osm_port_t *p_port;
    osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
    unsigned calculated_max_pg_id;
    unsigned calculated_max_lid;
    int rc = 0;

    PSGroupsData algorithm_data;
    algorithm_data.m_groups.reserve(sw_db_entry.GetGroupCap());

    uint16_t max_lid = 0;
    uint8_t **hops = NULL;
    PSGroupData group_data;
    PSGroupData empty_group_data;

    //for ARN/FRN consume string calculation
    PortsBitset down_sw_ports;
    PortsBitset ca_ports;
    PortsBitset sw_ports;

    uint8_t sw_rank = OSM_SW_NO_RANK;
    uint8_t max_consume_rank = 0;

    /* Group is always assigned per leaf switch in Tree topologies */
    bool is_group_per_leaf_sw = true;
    uint8_t sw_plane = 0;

    /*
     * Step 1.1 -
     *  Ensure that this switch supports AR, and that AR is enabled.
     */
    if (!this->IsARActive(sw_db_entry)) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                "Skip Switch GUID 0x%" PRIx64 ", LID %u: " \
                   "AR not supported or not enabled.\n",
                sw_db_entry.m_general_sw_info.m_guid,
                sw_db_entry.m_general_sw_info.m_lid);
        //we already suppose to have this switch in not supported list
        OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
    }

    /*
     * Step 1.2 -
     *  Set default values for calculations
     */
    calculated_max_pg_id = 0;
    calculated_max_lid = 0;

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Calculate group bitmask for GUID 0x%" PRIx64 ", LID %u: \n",
            sw_db_entry.m_general_sw_info.m_guid,
            sw_db_entry.m_general_sw_info.m_lid);

    /*
     * Step 1.3 -
     *  Build down switch ports and ca ports bitset
     */
    sw_db_entry.m_config_rn =
        (IS_GENERATE_ARN(m_master_db, sw_db_entry) ||
         IS_GENERATE_FRN(m_master_db, sw_db_entry));

    if (sw_db_entry.m_config_rn) {

        down_sw_ports.reset();
        ca_ports.reset();
        sw_ports.reset();

        sw_rank = sw_db_entry.m_general_sw_info.m_p_osm_sw->rank;

        osm_node_t *p_osm_node =
            sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;

        for( u_int8_t port_num = 1 ;
             port_num <= p_osm_node->node_info.num_ports ; port_num++) {

            osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
            if (!p_physp || !osm_link_is_healthy(p_physp))
                continue;

            osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
            if (!p_remote_physp || !p_remote_physp->p_node)
                continue;


            if(p_remote_physp->p_node->sw){
                sw_ports.set(port_num);
                if (p_remote_physp->p_node->sw->rank > sw_rank)
                    down_sw_ports.set(port_num);
            } else
                ca_ports.set(port_num);
        }

    }


    flow_graph_t *p_ctx_graph;
    flow_vertex_t *p_ctx_src_vertex;
    unsigned int *sw_lid_to_vertex_mapping;
    flow_vertex_t *p_src_vertex;
    bool calculate_flow = false;
    bool calc_2_sub_groups = (m_p_osm_subn->opt.ar_tree_asymmetric_flow ==
                              OSM_ASYMMETRIC_TREE_2_SUB_GROUPS_ENABLED);
    int i;
    int start_idx = 0;

    if (m_p_osm_subn->opt.ar_tree_asymmetric_flow) {
        p_ctx_graph = context->p_graph;
        p_ctx_src_vertex = (flow_vertex_t *)p_sw->priv;
        calculate_flow = p_ctx_src_vertex->calculate_flow;
    }

    if (calculate_flow) {
        sw_lid_to_vertex_mapping = p_ctx_graph->sw_lid_to_vertex_mapping;
        p_src_vertex = &p_ctx_graph->vertices[p_ctx_src_vertex->index];
    }

    /*
     * Step 2.1 -
     *  Iterate through all lids and calculate group bitmask for each lid.
     */
    max_lid = sw_db_entry.m_general_sw_info.m_p_osm_sw->max_lid_ho;
    hops = sw_db_entry.m_general_sw_info.m_p_osm_sw->hops;
    //u_int8_t last_port_num;

    uint16_t sw_lid_num;
    vector<uint16_t>::iterator sw_lid_iter = context->m_sw_lids.begin();
    for (; sw_lid_iter != context->m_sw_lids.end(); ++sw_lid_iter) {
        bool is_flow_subgroup_calculated = FALSE;
        int first_port_in_group = 0;

        sw_lid_num = *sw_lid_iter;
        p_port = osm_get_port_by_lid(&m_p_osm->subn, cl_hton16(sw_lid_num));
        if (!p_port || !p_port->p_node || !p_port->p_node->sw) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "can't find p_port for sw_lid_num %u (p_port %p), skip\n",
                       sw_lid_num, p_port);
            continue;
        }
        p_dest_sw = p_port->p_node->sw;

        if (!hops[sw_lid_num] || hops[sw_lid_num][0] < 1 || hops[sw_lid_num][0] == OSM_NO_PATH) {
            if(!hops[sw_lid_num]) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "hops[lid:%u] == NULL \n", sw_lid_num);
            } else {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "hops[lid:%u][0] == %u \n", sw_lid_num, hops[sw_lid_num][0]);
            }
            continue;
        }

        group_data.Clear();

        if (calculate_flow) {
            unsigned int dst_vertex_index = sw_lid_to_vertex_mapping[sw_lid_num];
            flow_vertex_t *p_dst_vertex = &p_ctx_graph->vertices[dst_vertex_index];
            flow_edge_t *p_edge;
            flow_port_group_t *port_group_representative;
            flow_port_groups_t *src_port_groups = p_ctx_graph->src_port_groups;
            flow_port_groups_t *src_port_group = &src_port_groups[p_src_vertex->index];
            int e, port_num, vertex_index_of_representative;
            uint8_t *port_map_of_representative;
            uint8_t *ports;

            if (p_dst_vertex->calculate_flow) {
                if (src_port_group->vertex_index_of_representative != p_dst_vertex->index) {
                    vertex_index_of_representative = src_port_group->vertex_index_of_representative;
                    port_map_of_representative = src_port_group->representative_port_map;
                } else {
                    vertex_index_of_representative = src_port_group->vertex_index_of_second_representative;
                    port_map_of_representative = src_port_group->second_representative_port_map;
                }

                port_group_representative = &src_port_groups[vertex_index_of_representative].port_groups[p_dst_vertex->index];
                ports = port_group_representative->ports;
                if (port_group_representative->num_ports == 0) {
                    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                               "No path found between src_vertex: %d, "
                               "dst_vertex: %d, vertex represent: %d\n",
                               p_src_vertex->index, p_dst_vertex->index,
                               vertex_index_of_representative);
                }

                if (port_group_representative->num_ports)
                    first_port_in_group = port_map_of_representative[ports[0]];

                for (i = 0; i < port_group_representative->num_ports; i++) {
                    group_data.m_ps_group_bitmask.set_pri(port_map_of_representative[ports[i]]);
                    if (!calc_2_sub_groups)
                        group_data.m_group_size++;

                }

                is_flow_subgroup_calculated = TRUE;
            }
        }

        /*
         *  Calculate group bitmask.
         */
        for (u_int8_t port_num = 1; port_num <= sw_db_entry.m_general_sw_info.m_num_ports; ++port_num) {
            if (hops[sw_lid_num][0] == hops[sw_lid_num][port_num]) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "hops[%u][%u]:%u \n", sw_lid_num, port_num , hops[sw_lid_num][port_num]);
                if (!is_flow_subgroup_calculated) {
                    if (!first_port_in_group)
			first_port_in_group = port_num;
                    group_data.m_ps_group_bitmask.set_pri(port_num);
                    group_data.m_group_size++;
                } else if (calc_2_sub_groups) {
                    group_data.m_ps_group_bitmask.set_sec(port_num);
                    group_data.m_group_size++;
                }
            }
        }

        if (!is_flow_subgroup_calculated && calculate_flow && calc_2_sub_groups) {
            group_data.m_ps_group_bitmask.get_secondary() =
                group_data.m_ps_group_bitmask.get_primary();
        }

        if ((calculate_flow && calc_2_sub_groups) &&
            (group_data.m_ps_group_bitmask.get_primary() !=
                 group_data.m_ps_group_bitmask.get_secondary())) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "Different primary and secondary group bitmask's for "
                       "GUID 0x%" PRIx64 ", LID %u: %s\n",
                       sw_db_entry.m_general_sw_info.m_guid,
                       sw_db_entry.m_general_sw_info.m_lid,
                       group_data.m_ps_group_bitmask.to_string().c_str());
        }

        if (group_data.m_group_size >= 1 || sw_db_entry.m_config_rn) {
            PSGroupBitmaskToDataMapInsertRes res =
                algorithm_data.m_ps_groups_map.insert(
                    make_pair(group_data.m_ps_group_bitmask, group_data));

	    if (res.second) { /* If new group */
		    algorithm_data.m_groups.push_back(&res.first->second);
	    }

	    this->AddLidToARGroup(context->m_sw_to_lids[sw_lid_num], sw_lid_num,
                                  &res.first->second, algorithm_data,
                                  is_group_per_leaf_sw);

            vector<uint16_t> &lids = context->m_sw_to_lids[sw_lid_num];
            vector<uint16_t>::iterator lid_iter = lids.begin();
            if (group_data.m_group_size >= 1) {
                ARSetUcastPort(sw_db_entry, group_data, sw_lid_num, first_port_in_group);

                /*
                 * Don't change in AR LFT for destination HCA static port to be part of ar group
                 * if the configured switch and destination switch have no common plane.
                 * All plane ports of the planarized HCA aport have the same lid,
                 * m_sw_to_lids[sw_lid_num] has HCA lids that are connected to this
                 * destination switch (sw_lid_num). If the destination switch is a single plane sw
                 * (ex Black Mamba), check/fix static port of HCA plane port, that has the same
                 * plane as the destination switch.
                 * Example is Black Mamba connected directly to planarized HCA. Each single plane
                 * asic points in static port to the plane port that is connected directly to it,
                 * don't fix it to this system other asic destination switch ar group
                 */
		if ((p_sw->plane_mask || p_dest_sw->plane_mask) &&
                    !(p_sw->plane_mask & p_dest_sw->plane_mask)) {
                    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                               "sw 0x%" PRIx64 " (0x%x) and sw_lid_num %u 0x%" PRIx64
                               " (0x%x) have different planes, skip hcas\n",
                               sw_db_entry.m_general_sw_info.m_guid, p_sw->plane_mask,
                               sw_lid_num, cl_ntoh64(osm_node_get_node_guid(p_port->p_node)),
                               p_dest_sw->plane_mask);
		    continue;
		}

                for (; lid_iter != lids.end(); ++lid_iter) {
                    ARSetUcastPort(sw_db_entry, group_data, *lid_iter, first_port_in_group);
                }
            }
        }
    }

    /*
     * Step 2.2 -
     *  try to assign group number and group table number(for IS4) for all groups.
     */
    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Step 2.2 AssignPortGroups \n");

    if (!algorithm_data.m_lid_to_group_map.empty()) {       
        rc = this->AssignPerLeafSwitchGroups(sw_db_entry, algorithm_data);
        if (rc) 
            OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
    }

    /*
     * Step 2.3 -
     *  Calculate min string to consume.
     */

    max_consume_rank = sw_rank;
    if (sw_db_entry.m_config_rn &&
        !algorithm_data.m_groups.empty()) {

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "Step 2.3 Calculate min string to consume."
                   " sw_rank:%u\n",
                   sw_rank);

        GroupsPtrVec::iterator groups_iter = algorithm_data.m_groups.begin();
        uint8_t grp_min_rank = sw_rank;

        for ( ;
              groups_iter != algorithm_data.m_groups.end();
              ++groups_iter) {

	    PSGroupData *p_group_data = (PSGroupData *)(*groups_iter);

            this->PrintGroupData("Calculate consume_rank for group: ",
                           *p_group_data);

            PortsBitset &primmary_bitset = p_group_data->m_ps_group_bitmask.get_primary();
            PortsBitset &secondary_bitset = p_group_data->m_ps_group_bitmask.get_secondary();
            if (p_group_data->m_group_size > 1 &&
                ((primmary_bitset | secondary_bitset) & down_sw_ports).any() ) {

                    grp_min_rank = context->m_max_rank;

                    for (vector<uint16_t>::iterator lid_iter = (*groups_iter)->m_sw_lids_list.begin();
                         lid_iter != (*groups_iter)->m_sw_lids_list.end();
                         ++lid_iter)
                        grp_min_rank = min(grp_min_rank,
                                           context->m_sw_lid_to_rank_mapping[*lid_iter]);

                    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                               "grp_min_rank:%u \n", grp_min_rank);

                    //ignore parallel links (grp_min_rank == sw_rank+1)
                    if (grp_min_rank < sw_rank+2 ||
                        max_consume_rank == grp_min_rank)
                        continue;

                    if (max_consume_rank == sw_rank)
                        max_consume_rank = grp_min_rank;
                    else {
			    stringstream sstr;
			    if (m_p_osm_log->level & OSM_LOG_DEBUG) {
				    p_group_data->BitsetToOstream(sstr);
			    }

			OSM_AR_LOG(
                            m_p_osm_log, OSM_LOG_VERBOSE,
                            "Group minimal rank (%u) "
                            "is different from current rank (%u) "
                            "for GUID 0x%" PRIx64 ", LID %u: bitmask:%s\n",
                            grp_min_rank, max_consume_rank,
                            sw_db_entry.m_general_sw_info.m_guid,
                            sw_db_entry.m_general_sw_info.m_lid,
			    sstr.str().c_str());

                        //set maximal rank
                        if (max_consume_rank > grp_min_rank)
                            max_consume_rank = grp_min_rank;
                    }

                    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                               "max_consume_rank:%u \n", max_consume_rank);


            }


        }
    }

    /*
     * Step 3
     *  Now let's configure AR LFTs.
     */
    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Step 3 configure AR LFTs \n");

    for (unsigned dest_lid = 1; dest_lid <= m_p_osm_subn->max_lid; ++dest_lid) {
	/* Default Static lid state and group 0 - Empty on RN otherwise reserved */
	uint16_t group_number = 0;
        enum OSM_AR_LID_STATE ar_mode = OSM_AR_IB_LID_STATE_STATIC;
        // Find out which port was used to route to lid_num in the usual unicast routing
        u_int8_t ucast_lft_port =
                OsmSwitchGetPortByLidUseNewLft(sw_db_entry.m_general_sw_info.m_p_osm_sw,
                        dest_lid);
        if (ucast_lft_port == OSM_NO_PATH)  //no route to this lid_num from this switch
            continue;
        calculated_max_lid = dest_lid;

	BaseGroupData *iter = algorithm_data.m_lid_to_group_map.get(dest_lid);

        if (iter == NULL) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "LID %u dest_lid:%u - Set lid sate Static Group:%u\n",
                       sw_db_entry.m_general_sw_info.m_lid, 
                       dest_lid, 
                       ucast_lft_port);
            // We already mark static in lid state by default
        } else if (iter->m_group_table_number == ARMGR_MAX_GROUP_TABLES) {
		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			stringstream sstr;
			iter->BitsetToOstream(sstr);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "LID %u dest_lid:%u - was not mapped to group because group bitmask:%s was not mapped.\n",
				   sw_db_entry.m_general_sw_info.m_lid,
				   dest_lid,
				   sstr.str().c_str());
		}
        } else {
            group_number = iter->m_group_number;
            if (is_group_per_leaf_sw) {

                if (osm_get_type_by_lid_ho(&m_p_osm->subn, dest_lid) != IB_NODE_TYPE_SWITCH) {
                    //this is not sw lid. get the sw lid connected to this lid
                    
		    /* for prism or legacy switch, sw_plane would be 0 */
                    (void)osm_switch_is_prism_switch(p_sw, &sw_plane);
                    sw_lid_num = context->m_hca_to_sw_lid_mapping[sw_plane][dest_lid];
                }
                else
                    sw_lid_num = dest_lid;

                group_number = algorithm_data.m_sw_lid_to_group_number[sw_lid_num];
            }

            ar_mode = this->m_master_db.m_ar_mode;

		if ((group_number == AR_GROUP_UNASSIGNED) ||
		    !this->m_master_db.m_ar_enable) { /* only FR/FRN enabled */
			ar_mode = OSM_AR_IB_LID_STATE_STATIC;
			if (group_number == AR_GROUP_UNASSIGNED) {
				/*
				 * In case FR enabled group 0 is empty, else
				 * group number is reserved for ar_mode == STATIC
				 */
				group_number = 0;
			}
		}

            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "LID %u dest_lid:%u - Set lid state %u Group:%u Table:%u (p_group:%p)\n",
                       sw_db_entry.m_general_sw_info.m_lid, dest_lid, 
                       ar_mode,
                       group_number,
                       iter->m_group_table_number,
                       iter);
        } 
        osm_ar_lft_set(p_ar_data, dest_lid, 0, ucast_lft_port, ar_mode,
                       group_number);

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "LID %u dest_lid:%u - GroupNumber:%u TableNumber:%u LidState:%u ucast_lft_port %u\n",
                   sw_db_entry.m_general_sw_info.m_lid, dest_lid,
                   group_number, 0, ar_mode, ucast_lft_port);
    }      /* End of loop destination lids */

    /*
     * Step 4 -
     * Update calculated_ar_group_table.
     * we optimize ARGroup MAD sending by using copy
     */
    SetPortGroupsForRn(sw_db_entry, algorithm_data);

	/*
	* Step 5 -
	* Update RN tables if needed
	*/
    if (sw_db_entry.m_config_rn){
        this->UpdateRNRcvString(sw_db_entry, context->m_max_rank, sw_rank,
                                max_consume_rank);
        this->UpdateRNXmitPortMask(sw_db_entry, ca_ports, sw_ports,
                                   down_sw_ports.any());
    }
}

void OSMAdaptiveRoutingManager::AddLidToARGroup(vector<uint16_t> &lids, uint16_t sw_lid_num,
						BaseGroupData *p_group_data,
						GroupsData &groups_data,
						bool is_group_per_leaf_sw)
{
	p_group_data->m_lids_list.push_back(sw_lid_num);
	p_group_data->m_sw_lids_list.push_back(sw_lid_num);
	groups_data.m_lid_to_group_map.set(sw_lid_num, p_group_data);

	vector<uint16_t>::iterator lid_iter = lids.begin();
	for (; lid_iter != lids.end(); ++lid_iter) {
		p_group_data->m_lids_list.push_back(*lid_iter);
		groups_data.m_lid_to_group_map.set(*lid_iter, p_group_data);
	}

	if (is_group_per_leaf_sw && lids.size()) {
		/* if switch has CAs connected to it, add switch to p_group_data->m_leaf_switches */
		p_group_data->m_leaf_switches.push_back(sw_lid_num);
	}

	PrintGroupData("Add Group Data: ", *p_group_data);
}

void OSMAdaptiveRoutingManager::PrintGroupData(const char *str,
					       BaseGroupData &group_data)
{

	if (this->m_p_osm_log->level & OSM_LOG_DEBUG) {

		stringstream sstr;
		vector<uint16_t>::const_iterator iter =  group_data.m_lids_list.begin();
		int lid_count = 1;
		//print first 100 lids
		for (; iter != group_data.m_lids_list.end() ; ++iter) {
			sstr << " " << *iter;
			lid_count++;
			if (lid_count > 100) {
				sstr << " ...";
				break;
			}
		}

		sstr << " bitset:";
		group_data.BitsetToOstream(sstr);

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				"%s Group:%p size:%d lids_number:%u lids:%s\n",
				str,
				&group_data,
				group_data.m_group_size,
				(uint16_t)group_data.m_lids_list.size(),
				sstr.str().c_str());
	}
}

int OSMAdaptiveRoutingManager::AssignPortGroups(ARSWDataBaseEntry &sw_db_entry,
						GroupsData &groups_data)
{
	int rc = 0;
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	uint16_t group_cap = sw_db_entry.GetGroupCap();

	if (sw_db_entry.m_config_rn) {
		rc = AssignPerLeafSwitchGroups(sw_db_entry, groups_data);
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	GroupsList group_list;

	rc = GetOrderedGroupList(groups_data, group_list);
	if (rc) OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "AssignPortGroups for "
		   "Switch GUID: 0x%" PRIx64 ", LID: %u "
		   "groups number: %u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   (int)group_list.size());


	//reserve group 0 as empry group
	unsigned &next_group_number = groups_data.m_next_group_number;
	if (groups_data.m_reset_next_group_number) {
		next_group_number = (m_master_db.m_fr_enable ? 1 : 0);
		groups_data.m_reset_next_group_number = false;
	}

	GroupsListIter list_iter = group_list.begin();
	for ( ; list_iter != group_list.end() && next_group_number < group_cap;
	     ++list_iter) {
		(*list_iter)->m_group_number = next_group_number;
		(*list_iter)->m_group_table_number = 0;

		if ((*list_iter)->m_group_size == 1){
			continue;
		}
		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Assign AR port group number: %u "
				   "for (p_group: %p):\n",
				   next_group_number, *list_iter);
			PrintGroupData("AR port group: ",*(*list_iter));
		}
		next_group_number++;

		groups_data.m_assign_groups.push_front(*list_iter);
	}

	if ((next_group_number == group_cap) &&
	    (m_p_osm_log->level & OSM_LOG_DEBUG)) {

		for ( ;list_iter != group_list.end(); ++list_iter) {
			stringstream sstr;
			(*list_iter)->BitsetToOstream(sstr);

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID 0x%" PRIx64
				   ", Can not assign group %s \n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sstr.str().c_str());
		}
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int OSMAdaptiveRoutingManager::AssignPerLeafSwitchGroups(ARSWDataBaseEntry &sw_db_entry,
							 GroupsData &groups_data)
{
	int rc = 0;

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "AssignPerLeafSwitchGroups for Switch GUID: 0x%" PRIx64
		   ", LID: %u, groups number: %u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   (int) groups_data.m_groups.size());

	uint16_t group_number = 0;
	uint16_t group_base;
	bool is_group_cap_exceeded = false;

	/*
	 * Divide GroupCap by m_groups_per_leaf_ since each allocated
	 * group_base is used to define m_groups_per_leaf_ separate groups
	 */
	uint16_t group_cap =
		sw_db_entry.GetGroupCap() / groups_data.m_groups_per_leaf;

	GroupsPtrVec::iterator iter = groups_data.m_groups.begin();
	for (; iter != groups_data.m_groups.end(); ++iter) {
		// will not perform AR for switches that are not leaf switches
		(*iter)->m_group_number = AR_GROUP_UNASSIGNED;
		(*iter)->m_group_table_number = 0;

		// Go over all LeafSwitches
		vector<uint16_t>::iterator leafs_sw_iter = (*iter)->m_leaf_switches.begin();
		for (; leafs_sw_iter != (*iter)->m_leaf_switches.end(); ++leafs_sw_iter) {

			/* Only PLFT 0 is used by tree topologies */
			group_base = AllocateSwArGroup(*leafs_sw_iter, group_cap, 0);

			if (group_base == AR_GROUP_UNASSIGNED) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Switch GUID 0x%" PRIx64 ", "
					   "Can not assign group (0x%p) for lid: %u "
					   "group_cap: %u groups_per_leaf: %u\n",
					   sw_db_entry.m_general_sw_info.m_guid,
					   *iter, *leafs_sw_iter,
					   group_cap,
					   groups_data.m_groups_per_leaf);
				is_group_cap_exceeded = true;
				continue;
			}
			/*
			 * One group_base number is allocated per switch
			 * The group nember allocated for m_per_sw_group_index equals
			 * (group_base - 1) * m_groups_per_leaf + m_per_sw_group_index
			 */
			group_number =
				(group_base - 1) * groups_data.m_groups_per_leaf +
				groups_data.m_per_sw_group_index;

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "Assign group %u for leaf switch lid:%u "
				   "per_sw_group_index: %u\n",
				   group_number, *leafs_sw_iter,
				   groups_data.m_per_sw_group_index);

			// Update both GroupData and GroupsData objects
			/* Only PLFT 0 is used by tree topologies */
			(*iter)->m_group_numbers[0].push_back(group_number);
			groups_data.m_sw_lid_to_group_number[*leafs_sw_iter] = group_number;
		}
		/*
		 * Clear m_leaf_switches to prepare calculating groups for the
		 * next  groups_data.m_per_sw_group_index
		 */
		(*iter)->m_leaf_switches.clear();
	}

	if (is_group_cap_exceeded)
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8519: SX Switch GUID 0x%" PRIx64
			   ", Can not assign separate groups number"
			   " for all leaf switches. group_cap: %u\n",
			   sw_db_entry.m_general_sw_info.m_guid, group_cap);

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

bool compare_groups_by_members(BaseGroupData* p_group1, BaseGroupData* p_group2)
{
	//compare by the following ordered by priority:
	//1) group members number (descending ordered)
	//2) group ports number   (descending ordered) - if there are 2 sub groups sort by the first
	//3) first lid number     (ascending ordered) - for stable sorting
	if (p_group1->m_lids_list.size() != p_group2->m_lids_list.size())
		return p_group1->m_lids_list.size() > p_group2->m_lids_list.size();
	if (p_group1->m_group_size != p_group2->m_group_size)
		return p_group1->m_group_size > p_group2->m_group_size;

	if (p_group1->m_lids_list.size() != 0) {
		return *p_group1->m_lids_list.begin() < *p_group2->m_lids_list.begin();
	}

	return *p_group1 < *p_group2;
}

int OSMAdaptiveRoutingManager::GetOrderedGroupList(GroupsData &groups_data,
						   GroupsList& group_list)
{
	int rc = 0;
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "GetOrderedGroupList: number of groups: %u\n",
		   (uint16_t) groups_data.m_groups.size());

	GroupsPtrVec::iterator iter = groups_data.m_groups.begin();
	for (; iter != groups_data.m_groups.end() ; ++iter) {
		if ((*iter)->m_group_number != AR_GROUP_UNASSIGNED) {
			continue; /* Skip - Already allocated group number */
		}
		group_list.push_back(*iter);
	}

	group_list.sort(compare_groups_by_members);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "GetOrderedGroupList: number of unassigned groups: %u\n",
		   (uint16_t) group_list.size());

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

/*
 * Preprocess the subnet and determine whether the routing algorithm can run
 * on the subnet or not.
 * Return 0 - Indicates the routing algorithm can route the subnet.
 * O/W - Indicates the routing algorithm can not route the subnet.
 */
int OSMAdaptiveRoutingManager::Preprocess()
{
	//TODO: Rename ARInfoGetGroupCapProcess to GetSwitchesCapabilities
	//TODO: Write appropriate error message if validation fails.

	int result = 0;

	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	/* Update inventory with new/removed switches */
	this->UpdateFabricSwitches();

	this->UpdateUserOptions();

	if ((m_master_db.m_ar_algorithm != AR_ALGORITHM_DF_PLUS) ||
	    !this->m_master_db.m_enable) {
		// Mark all switches as not updated in this cycle
		map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
			this->m_sw_db.m_sw_map.begin();
		while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
			sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;
			++sw_node_it;
		}
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, result);
}

/***************/
int OSMAdaptiveRoutingManager::ARCycle()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	int rc = 0;
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it;
	uint8_t configured_max_op_vl = m_p_osm_subn->opt.max_op_vls_sw ?
		m_p_osm_subn->opt.max_op_vls_sw :
		m_p_osm_subn->opt.max_op_vls;

	if (m_master_db.m_ar_algorithm == AR_ALGORITHM_DF_PLUS) {

		if (m_p_osm->subn.opt.qos != TRUE) {
    			OSM_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			"ERR 8513: DF_PLUS Algorithm not supported together with disabled qos.\n");
			this->m_master_db.m_enable = false;
		}

		/* It's important to have enough VLs on switches */
		if ((m_master_db.m_en_sl_mask != OSM_ALL_SL_VL_AR_ENABLED) &&
		    (configured_max_op_vl <= DF_MIN_VL_CAP)) {
			OSM_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				"ERR 8514: DF_PLUS Algorithm do not support disabling AR on sl "
				" when op_vls equals %u (max_op_vls %u max_op_vls_sw %u)\n",
				configured_max_op_vl,
				m_p_osm_subn->opt.max_op_vls, m_p_osm_subn->opt.max_op_vls_sw);
			this->m_master_db.m_enable = false;
		}

		if (!m_master_db.m_ar_enable) {
			OSM_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				"ERR 8515: DF_PLUS Algorithm does not support ar_mode %u\n",
				m_p_osm->subn.opt.ar_mode);
			this->m_master_db.m_enable = false;
		}

		if (!this->m_master_db.m_enable) {
			ClearAllDragonflyConfiguration();
			rc = -1;
		}
		else
			rc = ARDragonFlyCycle();

		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	if (!this->m_master_db.m_enable) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
			   "AR and Shield are disabled. enable %d\n",
			   this->m_master_db.m_enable);
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	/* Query for required switch capabilities */
	this->ARInfoGetGroupCapProcess();

	/* Validate switch capabilities */
	rc = this->ValidateSwitchesCapabilities();
	if (rc) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			"ERR 8501: failed to validate the required ar capabilities "
			"for advanced routing configuration\n");
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	rc = osm_ar_prepare_tables(this->m_p_osm_subn);
	if (rc) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			"ERR 8504: failed to prepare switches for advanced"
			" routing configuration\n");
		OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
	}

	switch( m_master_db.m_ar_algorithm )
	{
		case AR_ALGORITHM_TREE:
			this->ARCalculatePortGroupsTree();
			break;
		case AR_ALGORITHM_LAG:
			this->ARCalculatePortGroupsParallelLinks();
			break;
		case AR_ALGORITHM_MINHOP:
			this->ARCalculatePortGroupsMinhop();
			break;
		default:
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 8516: Unknown algorithm: %d.\n",
				   m_master_db.m_ar_algorithm);
			OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
	};

	// Set ARInfo after group top calculation
	this->ARInfoSetProcess();

	// HBF config
	this->HbfConfigProcess();

	// Currently RoutingNotification is supported only on Tree algorithm
	if ((m_master_db.m_ar_algorithm == AR_ALGORITHM_TREE) &&
			(m_master_db.m_frn_enable || m_master_db.m_arn_enable))
		this->TreeRoutingNotificationProcess();

	//mark the switches as updated
	sw_node_it = this->m_sw_db.m_sw_map.begin();
	for (; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {
	if (this->IsARActive(sw_node_it->second))
		// Mark that update was done for this switch in this cycle
		sw_node_it->second.m_app_data = ARMGR_SW_UPDATED;
	}

	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int OSMAdaptiveRoutingManager::Route()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	int rc = 0;

	m_is_temporary_error = false;
	m_is_permanent_error = false;

	static unsigned cycle_num = 0;
	++cycle_num;

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "Started configuration cycle\n");
	try {
		this->Preprocess();

		if (m_ar_algorithm)
			rc = m_ar_algorithm->RunCycle();
		else
			rc = ARCycle();

	} catch (int error) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 8500: Caught exception = %u, exiting AR Manager...\n",
			   error);
		rc = -1;
	}

	this->AROSMIntegrationProcess();
	if (rc) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "Configuration cycle (number %u) failed\n", cycle_num);
		fprintf(stdout, "AR Manager - Configuration cycle (number %u) failed\n", cycle_num);
		/*
		 * Route() should return negative value on error,
		 * in order to fail the current routing engine and to use fallback routing engine instead
		 */
		rc = -1;
	} else {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "Configuration cycle (number %u) completed successfully\n", cycle_num);
		fprintf(stdout, "AR Manager - Configuration cycle (number %u) completed successfully\n", cycle_num);
	}
	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

//============================================================================//
/***************/
void OSMAdaptiveRoutingManager::AROSMIntegrationProcess()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	unsigned int number_of_switches = this->m_sw_db.m_sw_map.size();
	unsigned int ar_configured_num = 0;
	unsigned int df_configured_num = 0;

	/*
	 * Step 1 -
	 *  Decide for each switch the status
	 */
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
		osm_switch_t *p_osm_sw = sw_node_it->second.m_general_sw_info.m_p_osm_sw;

		if ((sw_node_it->second.m_app_data == ARMGR_SW_UPDATED) &&
		    (this->IsARActive(sw_node_it->second))) {
			/*
			 * Step 1.1 -
			 *  we configured Adaptive Routing for this switch, need to mark it
			 */
			p_osm_sw->ar_configured = OSM_SW_AR_CONF_CONFIGURED;
			ar_configured_num++;
			/* it's impossible that switch would be df configured without ar */
			if (this->IsDFActive(sw_node_it->second))
				df_configured_num++;
		} else {
			// All AR configuration will be deleted from the switch
			// in case we had temp error (timeout) we will need to reconfigure.
			if (ib_ar_info_get_group_cap_ho(&sw_node_it->second.m_ar_info) != 0)
				sw_node_it->second.ClearARData();
			/*
			 * Step 1.2 -
			 *  we didn't configured ar for this switch, need to mark it
			 *  and also abort ar if it was enabled in previous cycle
			 */
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
				   "AR NOT configured on Switch GUID 0x%" PRIx64
				   ", LID %u\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);

			if (p_osm_sw->ar_configured) {
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "Clear old lft for Switch GUID 0x%" PRIx64
					   ", LID %u (because was configured in previous cycle)\n",
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid);
				// mark to SM to perform LFT set
				p_osm_sw->ar_configured = OSM_SW_AR_CONF_DISABLED;
			}
		}

		++sw_node_it;
	}
	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
		   "AR configured on %u/%u switches, DF+ routing configured on %u/%u switches\n",
		   ar_configured_num, number_of_switches,
		   df_configured_num, number_of_switches);

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

/****************************************************/
/***************** public methods *******************/
/****************************************************/
//============================================================================//
OSMAdaptiveRoutingManager::OSMAdaptiveRoutingManager(osm_opensm_t *p_osm,
						     ar_algorithm_t ar_algorithm):
	m_p_osm(p_osm),
	m_p_osm_subn(p_osm->sm.ucast_mgr.p_subn),
	m_p_osm_log(&p_osm->log),
	m_thread_pool(&p_osm->log),
	m_port_groups_calculator(this, &m_thread_pool, &p_osm->log, &m_sw_db),
	m_p_error_window_arr(NULL),
	m_is_permanent_error(false),
	m_is_temporary_error(false),
	m_df_configured(false),
	m_ar_algorithm(NULL),
	m_sw_info_configured(false),
	m_max_df_group_number(0),
	m_group_discovered(false),
	m_options_file_crc(0),
	m_is_init(false)
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	m_master_db.m_ar_algorithm = ar_algorithm;
	cl_u64_vector_construct(&m_dfp_potential_roots);
	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//============================================================================//
OSMAdaptiveRoutingManager::~OSMAdaptiveRoutingManager()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "Exiting...\n");
	if (m_p_error_window_arr != NULL)
		delete[] m_p_error_window_arr;
	delete m_ar_algorithm;
	m_root_switches.clear();
	cl_u64_vector_destroy(&m_dfp_potential_roots);

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//============================================================================//
void OSMAdaptiveRoutingManager::UpdateFabricSwitches()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);

	// Mark all switches as not updated
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
		sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;
		++sw_node_it;
	}

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "UpdateFabricSwitches\n");

	// Add new switches from osm database and update details on exist ones
	this->AddNewAndUpdateExistSwitches();

	// Remove non exists switches (in osm database) from our database
	this->RemoveAbsentSwitches();

	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//============================================================================//
void
OSMAdaptiveRoutingManager::UpdateUserOptions()
{
	ar_configuration_params_t *p_ar_conf = &m_p_osm_subn->ar_conf_params;

	this->SetDefaultConfParams();

	switch (p_ar_conf->ar_mode) {
	case OSM_ADV_ROUTING_MODE_ENABLED:
		m_master_db.m_ar_enable = true;
		m_master_db.m_arn_enable = false;
		break;
	case OSM_ADV_ROUTING_MODE_RN:
		m_master_db.m_ar_enable = true;
		m_master_db.m_arn_enable = true;
		break;
	default:
		m_master_db.m_ar_enable = false;
		m_master_db.m_arn_enable = false;
	}

	switch (p_ar_conf->shield_mode) {
	case OSM_ADV_ROUTING_MODE_ENABLED:
		m_master_db.m_flfr_enable = true;
		m_master_db.m_flfr_remote_disable = true;
		break;
	case OSM_ADV_ROUTING_MODE_RN:
		m_master_db.m_flfr_enable = true;
		m_master_db.m_flfr_remote_disable = false;
		break;
	default:
		m_master_db.m_flfr_enable = false;
		m_master_db.m_flfr_remote_disable = true;
	}

	if (p_ar_conf->pfrn_mode)
		m_master_db.m_pfrn_enable = true;

	m_master_db.m_enable =     m_master_db.m_enable && (m_master_db.m_ar_enable || m_master_db.m_flfr_enable);
	m_master_db.m_ar_enable =  m_master_db.m_enable && m_master_db.m_ar_enable;
	m_master_db.m_arn_enable = m_master_db.m_enable && m_master_db.m_arn_enable;
	m_master_db.m_fr_enable =  m_master_db.m_enable && m_master_db.m_flfr_enable;
	m_master_db.m_frn_enable = m_master_db.m_fr_enable && !m_master_db.m_flfr_remote_disable;
	m_master_db.m_pfrn_enable = m_master_db.m_enable && m_master_db.m_pfrn_enable;
	m_master_db.m_en_sl_mask = m_p_osm_subn->opt.ar_sl_mask;
	m_master_db.m_en_tr_mask = m_p_osm_subn->opt.ar_transport_mask;
	m_master_db.m_max_cas_on_spine = m_p_osm_subn->opt.dfp_max_cas_on_spine;
	m_master_db.m_hbf_sl_mask = m_p_osm_subn->opt.hbf_sl_mask;

	/*
	 * Currently only TREE supports FRN/ARN
	 * in addition LAG supports FR
	 */
	if (this->m_master_db.m_ar_algorithm != AR_ALGORITHM_TREE) {
		if (this->m_master_db.m_arn_enable || this->m_master_db.m_frn_enable) {
			this->m_master_db.m_arn_enable = false;
			this->m_master_db.m_frn_enable = false;

			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
				   "Currently only tree ar algorithm supports "
				   "ar routing notifications, "
				   "disabling routing notifications of ARN/FRN\n");
		}
		if ((this->m_master_db.m_ar_algorithm != AR_ALGORITHM_LAG) &&
		    (this->m_master_db.m_ar_algorithm != AR_ALGORITHM_DFP2)) {
			this->m_master_db.m_enable = this->m_master_db.m_ar_enable;
			if (this->m_master_db.m_fr_enable) {
				this->m_master_db.m_fr_enable = false;
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
					   "Currently only tree, minhop and lag ar algorithms " \
					   "support shield mode, disabling shield mode\n");
			}
		}
	}

	if ((m_master_db.m_ar_algorithm == AR_ALGORITHM_DF_PLUS) ||
	    (m_master_db.m_ar_algorithm == AR_ALGORITHM_DFP2))
		/* spine-to-spine non-DragonFly routes allowed */
		m_master_db.m_op_mode = 1;

	OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
		   "m_enable 0x%x, m_ar_enable 0x%x, m_arn_enable 0x%x, m_fr_enable 0x%x, m_frn_enable 0x%x, "
		   "m_pfrn_enable 0x%x, m_en_sl_mask 0x%x, m_en_tr_mask 0x%x, m_max_cas_on_spine %u, "
		   "m_hbf_sl_mask 0x%x\n",
		   m_master_db.m_enable, m_master_db.m_ar_enable, m_master_db.m_arn_enable,
		   m_master_db.m_fr_enable, m_master_db.m_frn_enable, m_master_db.m_pfrn_enable,
		   m_master_db.m_en_sl_mask, m_master_db.m_en_tr_mask,
		   m_master_db.m_max_cas_on_spine, m_master_db.m_hbf_sl_mask);
}

int OSMAdaptiveRoutingManager::Construct()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	int rc = 0;

	this->UpdateFabricSwitches();

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
		sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;
		++sw_node_it;
	}

	// Initialize - if already been initialized than do nothing
	this->Init();
	OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

void OSMAdaptiveRoutingManager::SetDefaultConfParams()
{
	OSM_AR_LOG_ENTER(this->m_p_osm_log);
	/* Set Master Database */
	m_master_db.m_enable                 = AR_CFG_DEFAULT_ENABLE;
	m_master_db.m_ar_enable              = AR_CFG_DEFAULT_AR_ENABLE;
	m_master_db.m_arn_enable             = AR_CFG_DEFAULT_ARN_ENABLE;
	m_master_db.m_flfr_enable            = AR_CFG_DEFAULT_FLFR_ENABLE;
	m_master_db.m_flfr_remote_disable    = AR_CFG_DEFAULT_FLFR_REMOTE_DISABLE;
	m_master_db.m_pfrn_enable	     = AR_CFG_DEFAULT_PFRN_ENABLE;
	m_master_db.m_ar_mode                = AR_CFG_DEFAULT_SW_AR_MODE;
	m_master_db.m_en_sl_mask             = AR_CFG_DEFAULT_EN_SL_MASK;
	m_master_db.m_dfp_en_vl_mask         = AR_CFG_DEFAULT_DFP_EN_VL_MASK;
	m_master_db.m_en_tr_mask             = AR_CFG_DEFAULT_TRANSPORT_MASK;
	m_master_db.m_op_mode                = 0x0;

	/* Set SW Database */
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
	while (sw_node_it != this->m_sw_db.m_sw_map.end()) {
		sw_node_it->second.m_ageing_time_value = AR_CFG_DEFAULT_SW_AGEING_TIME;
		sw_node_it->second.m_option_on = AR_CFG_DEFAULT_SW_ENABLE;
		++sw_node_it;
	}
	OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

int OSMAdaptiveRoutingManager::GetSL2VL(osm_physp_t *osm_phys_port,
					uint8_t iport_num, uint8_t oport_num,
					uint8_t (*sl2vl)[IB_NUMBER_OF_SLS])
{
	if (m_ar_algorithm)
		return m_ar_algorithm->GetSL2VL(osm_phys_port, iport_num, oport_num, sl2vl);

	if (m_df_configured)
		return GetSL2VLDF(osm_phys_port, iport_num, oport_num, sl2vl);

	return -1;
}

void SingleSubGroupPerGroupData::UpdateSWGroupElement(osm_log_t *p_osm_log,
						      ARSWDataBaseEntry &sw_db_entry,
						      uint16_t group_number)
{
	OSM_AR_LOG_ENTER(p_osm_log);

	OSM_AR_LOG(p_osm_log, OSM_LOG_DEBUG,
		   "UpdateSWGroupElement Switch GUID 0x%" PRIx64 ", LID %u " \
		   "group_number:%u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   group_number);

	osm_ar_subgroup_t *p_subgroup;

	p_subgroup = osm_ar_subgroup_get(&sw_db_entry.m_general_sw_info.m_p_osm_sw->ar_data,
					 group_number, 0);
	m_group_bitmask.get_data(*p_subgroup);
	OSM_AR_LOG(p_osm_log, OSM_LOG_DEBUG,
		   "Set Group:%u (ports mask head: 0x%" PRIx64 ")\n",
		   group_number,
		   cl_hton64(*((ib_net64_t*)&p_subgroup->mask[24])));

	OSM_AR_LOG_RETURN_VOID(p_osm_log);
}

void PSGroupData::UpdateSWGroupElement(osm_log_t *p_osm_log,
				       ARSWDataBaseEntry &sw_db_entry,
				       uint16_t group_number)
{
	OSM_AR_LOG_ENTER(p_osm_log);

	OSM_AR_LOG(p_osm_log, OSM_LOG_DEBUG,
		   "UpdateSWGroupElement Switch GUID 0x%" PRIx64 ", LID %u " \
		   "group_number:%u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   group_number);

	osm_ar_subgroup_t *p_subgroup;
	osm_advanced_routing_data_t *p_ar_data;

	p_ar_data = &sw_db_entry.m_general_sw_info.m_p_osm_sw->ar_data;

	p_subgroup = osm_ar_subgroup_get(p_ar_data, group_number, 0);
	PortsBitset &primary_bitset = m_ps_group_bitmask.get_primary();
	PortsBitset &secondary_bitset = m_ps_group_bitmask.get_secondary();

	primary_bitset.get_data(*p_subgroup);

	OSM_AR_LOG(p_osm_log, OSM_LOG_DEBUG,
		   "Set Group:%u (primary ports mask head:0x%" PRIx64 ")\n",
		   group_number,
		   cl_hton64(*((ib_net64_t*)&p_subgroup->mask[24])));

	if (secondary_bitset.any()) {
		p_subgroup = osm_ar_subgroup_get(
				p_ar_data,
				group_number, 1);

		secondary_bitset.get_data(*p_subgroup);
		OSM_AR_LOG(p_osm_log, OSM_LOG_DEBUG,
			   "Set Group:%u (secondary ports mask head:0x%" PRIx64 ")\n",
			   group_number,
			   cl_hton64(*((ib_net64_t*)&p_subgroup->mask[24])));
	}

	OSM_AR_LOG_RETURN_VOID(p_osm_log);
}
